#define STRICT
#define WINVER 0x400
#define _WIN32_WINNT 0x400
#define _WIN32_IE 0x400

#include "edit.h"
#include "pcbe.h"
#include "drawelem.h"
#include "element.h"
#include "defs.h"
#include "board.h"
#include "dialog.h"
#include "vectfont.h"
#include "getinterval.h"
#include "scrclip.h"
#include "clipelem.h"

#define TEMP_MEM_SIZE	0x200000
#define TEMP_MEM_OVER	64
#define ETAR			1.0e-4
#define DBL_HI_LIM		1.0e100
#define DBL_MINI_LIM	1.0e-100
#define PAINT_OVER		0.95
#define TH_LAND_AD		0.05

#define DOUBLE_ADJUST	63

#define CHECK_NUM		10000

typedef enum { ED_MOVE, ED_EXPAND, ED_GROUP, ED_LINE2, ED_ARC_R, ED_ARC_ANG } edit_edit_t;

extern BYTE gTempMem[];

#pragma pack(push, 1)

typedef struct edit_pos {
	struct edit_elem {
		edit_elem *next;
		container *pc;
	};
	enum { f_arc_r = 1, f_arc_r2 = 2, top_ofs = 8 };
	static void *top;
	struct edit_pos *next;
	edit_elem *pe;
	pos p;
	static void Init();
	static edit_pos *GetPosR2Top()
	{
		return (edit_pos *)((char *)gTempMem + top_ofs);
	}
	static edit_pos *GetPosTop()
	{
		return &GetPosR2Top()[1];
	}
	static edit_pos *GetArcP1Top()
	{
		return GetPosTop();
	}
	static void SetArcR()
	{
		*gTempMem = f_arc_r;
	}
	static int IsArcR()
	{
		return *gTempMem & f_arc_r;
	}
	static int IsArcR2()
	{
		return *gTempMem & f_arc_r2;
	}
	static void SetArcR2()
	{
		if (top == GetPosTop()) {
			top = &((edit_pos *)top)[1];
		}
		*gTempMem = (f_arc_r | f_arc_r2);
	}
	static void ChengArcR2()
	{
		*gTempMem ^= f_arc_r2;
	}
	static edit_pos *GetArcP2Top()
	{
		return GetPosR2Top();
	}
	static int __fastcall Add(container *, const pos * = NULL);
} *pedit_pos;

typedef struct {
	enum { T_NON, T_LINE, T_ARC };
	enum { B_V = 0x01, B_H = 0x02, B_ST = 0x04 };
	int mode;
	int f;
	posXY pn;
	union {
		struct {
			posXY px;
			double dx, dy, a;
		} li;
		struct {
			posXY pb;
			double r, dr;
			double st, end;
		} ar;
	} da;
} para_t;

typedef struct {
	pos px;
	element *el;
	double l;
} conns_t;

class paint_t: public node
{
public:
	enum { B_LINE = 1, B_ARC = 2, B_LINEV = 4 };
	int flag;
	int dummy;
	pos pl, ph;
	union {
		struct {
			double dx, dy, a;
		} li;
		struct {
			posXY pb;
			double r, st, en;
		} ar;
	};
	paint_t *Ref()
	{
		return (paint_t *)node::Ref();
	}
	paint_t *Cut()
	{
		return (paint_t *)node::Cut();
	}
	paint_t *Prev()
	{
		return (paint_t *)node::Prev();
	}
};

typedef struct {
	node el;
	node pa;
} netBase_t;

typedef struct thHole_t {
	int num;
	thHole_t *next;
	element *pe[];
} thHole_t;

#pragma pack(pop)

static void EditDummy();	// pcbe.cpp ŎQ
static void FnEditUndone();
static void FnLandMv();
static void FnLandLu();
static void FnLandValu();
static void FnLandKey();
static void FnLandEsc();
#define FnFlshMv FnLandMv
static void FnFlshLu();
static void FnFlshValu();
static void FnFlshKey();
static void FnFlshEsc();
#define FnLineMv1 FnLandMv
static void FnLineLu1();
static void FnLineValu1();
static void FnLineKey1();
static void FnLineEsc();
static void FnLineLu2();
static void FnLineMv2();
static void FnLineValu2();
static void FnLineKey2();
static void FnLineUnd2();
static void FnLineEsc2();
static void FnArcLu();
static void FnArcValu();
static void FnArcKey();
static void FnArcEsc();
static void FnArcLu1();
static void FnArcValu1();
static void FnArcKey1();
static void FnArcUnd1();
static void FnArcEsc1();
static void FnArcMv2();
static void FnArcLu2();
static void FnArcValu2();
#define FnArcKey2 FnArcKey1
static void FnArcUnd2();
static void FnArcEsc2();
static void FnPadLu();
static void FnPadValu();
static void FnPadKey();
static void FnPadEsc();
static void FnTextLu();
static void FnTextValu();
static void FnTextKey();
static void FnTextEsc();
static void FnEditLu();
static void FnEditKey();
static void FnEditEsc();
static void FnEditMv1();
static void FnEditLu1();
static void FnEditValu1();
static void FnEditKey1();
static void FnEditUnd1();
#define FnEditEsc1 FnEditUnd1
static void FnEraseLu();
static void FnEraseLMv();
static void FnEraseKey();
static void FnEraseEsc();
static void FnEraseLu1();
static void FnEraseEsc1();
static void FnEraseLu2();
static void FnEraseKey2();
static void FnEraseEsc2();
#define FnEraseUnd2 FnEraseEsc2
static void FnParaLu();
static void FnParaKey();
static void FnParaEsc();
static void FnParaLu1();
static void FnParaUnd1();
#define FnParaEsc1 FnParaUnd1
static void FnParaLu2();
#define FnParaUnd2 FnParaUnd1
#define FnParaEsc2 FnParaUnd1
static void FnParaLu3();
#define FnParaUnd3 FnParaUnd1
#define FnParaEsc3 FnParaUnd1
static void FnParaLu4();
static void FnParaUnd4();
static void FnParaEsc4();
static void FnConnLu();
static void FnConnLMv();
static void FnConnKey();
static void FnConnEsc();
static void FnConnLu1();
static void FnConnUnd1();
#define FnConnEsc1 FnConnUnd1
static void FnConnLu2();
static void FnConnUnd2();
#define FnConnEsc2 FnConnUnd2
static void FnConnLMv3();
static void FnConnUnd3();
static void FnConnEsc3();
static void FnConnLu4();
static void FnConnEsc4();
static void FnConnLu5();
#define FnConnUnd5 FnConnUnd1
#define FnConnEsc5 FnConnUnd1
static void FnChengeLu();
static void FnChengeKey();
static void FnChengeEsc();
static void FnPaintLu();
static void FnPaintKey();
static void FnPaintEsc();
static void FnPaintLu1();
static void FnPaintUnd1();
static void FnPaintEsc1();
static void FnPaintLu2();
static void FnPaintKey2();
static void FnPaintUnd2();
#define FnPaintEsc2 FnPaintEsc1
static void FnPaintLu3();
static void FnPaintUnd3();
static void FnPaintEsc3();
static void FnGroupLu();
static void FnGroupLMv();
static void FnGroupKey();
static void FnGroupEsc();
static void FnGroupLu1();
static void FnGroupLMv1();
static void FnGroupKey1();
static void FnGroupUnd1();
static void FnGroupEsc1();
static void FnGroupLu2();
static void FnGroupEsc2();
static void FnGroupLu3();
static void FnGroupEsc3();
static void FnGroupLu4();
static void FnGroupValue4();
static void FnGroupKey4();
static void FnGroupUnd4();
#define FnGroupEsc4 FnGroupUnd4
static void FnGroupUnd5();
#define FnGroupEsc5 FnGroupUnd5
static void FnGroupMv5();
static void FnGroupLu5();
static void FnGroupValue5();
static void FnGroupKey5();
static void FnCrsInLu();
static void FnCrsInUnd();
#define FnCrsInEsc FnCrsInUnd
static void FnCrsInLu1();
static void FnCrsInUnd1();
static void FnCrsInEsc1();
static void FnZoomInLUp1();
static void FnZoomInUnd1();
#define FnZoomInEsc1 FnZoomInUnd1
static void FnZoomInLUp2();
#define FnZoomInUnd2 FnZoomInUnd1
static void FnZoomInEsc2();
static void FnOrigLu();
static void FnOrigValu();
static void FnOrigKey();
static void FnOrigUnd();
#define FnOrigEsc FnOrigUnd
static void FnCoorLu();
static void FnCoorValu();
static void FnCoorKey();
static void FnCoorUnd();
#define FnCoorEsc FnCoorUnd
static void FnDistLu();
static void FnDistUnd();
#define FnDistEsc FnDistUnd
static void FnDistLu1();
static void FnDistUnd1();
static void FnDistEsc1();
static void FnDistLu2();
#define FnDistUnd2 FnDistUnd1
#define FnDistEsc2 FnDistEsc1
static void FnDispNetLu();
static void FnDispNetEsc();
#define FnDispNetUnd FnDispNetEsc
static void FnDispPinLu();
void FnDispPinEsc();
#define FnDispPinUnd FnDispPinEsc

const edit_func gEditFnLand[][MAX_EDFNC] = {
	{ FnLandMv, FnLandLu, EditDummy, FnLandValu, FnLandKey, FnEditUndone, FnLandEsc }
};
const edit_func gEditFnSq[][MAX_EDFNC] = {
	{ FnFlshMv, FnFlshLu, EditDummy, FnFlshValu, FnFlshKey, FnEditUndone, FnFlshEsc }
};
const edit_func gEditFnLine[][MAX_EDFNC] = {
	{ FnLineMv1, FnLineLu1, EditDummy, FnLineValu1, FnLineKey1, FnEditUndone, FnLineEsc },
	{ FnLineMv2, FnLineLu2, EditDummy, FnLineValu2, FnLineKey2, FnLineUnd2, FnLineEsc2 }
};
const edit_func gEditFnArc[][MAX_EDFNC] = {
	{ EditDummy, FnArcLu, EditDummy, FnArcValu, FnArcKey, FnEditUndone, FnArcEsc },
	{ EditDummy, FnArcLu1, EditDummy, FnArcValu1, FnArcKey1, FnArcUnd1, FnArcEsc1 },
	{ FnArcMv2, FnArcLu2, EditDummy, FnArcValu2, FnArcKey2, FnArcUnd2, FnArcEsc2 }
};
const edit_func gEditFnPad[][MAX_EDFNC] = {
	{ EditDummy, FnPadLu, EditDummy, FnPadValu, FnPadKey, FnEditUndone, FnPadEsc }
};
const edit_func gEditFnText[][MAX_EDFNC] = {
	{ EditDummy, FnTextLu, EditDummy, FnTextValu, FnTextKey, FnEditUndone, FnTextEsc }
};
const edit_func gEditFnEdit[][MAX_EDFNC] = {
	{ EditDummy, FnEditLu, EditDummy, EditDummy, FnEditKey, FnEditUndone, FnEditEsc },
	{ FnEditMv1, FnEditLu1, EditDummy, FnEditValu1, FnEditKey1, FnEditUnd1, FnEditEsc1 }
};
const edit_func gEditFnErase[][MAX_EDFNC] = {
	{ EditDummy, FnEraseLu, FnEraseLMv, EditDummy, FnEraseKey, FnEditUndone, FnEraseEsc },
	{ EditDummy, FnEraseLu1, EditDummy, EditDummy, EditDummy, EditDummy, FnEraseEsc1 },
	{ EditDummy, FnEraseLu2, EditDummy, EditDummy, FnEraseKey2, FnEraseUnd2, FnEraseEsc2 }
};
const edit_func gEditFnPara[][MAX_EDFNC] = {
	{ EditDummy, FnParaLu, EditDummy, EditDummy, FnParaKey, FnEditUndone, FnParaEsc },
	{ EditDummy, FnParaLu1, EditDummy, EditDummy, EditDummy, FnParaUnd1, FnParaEsc1 },
	{ EditDummy, FnParaLu2, EditDummy, EditDummy, EditDummy, FnParaUnd2, FnParaEsc2 },
	{ EditDummy, FnParaLu3, EditDummy, EditDummy, EditDummy, FnParaUnd3, FnParaEsc3 },
	{ EditDummy, FnParaLu4, EditDummy, EditDummy, EditDummy, FnParaUnd4, FnParaEsc4 }
};
const edit_func gEditFnConn[][MAX_EDFNC] = {
	{ EditDummy, FnConnLu, FnConnLMv, EditDummy, FnConnKey, FnEditUndone, FnConnEsc },
	{ EditDummy, FnConnLu1, EditDummy, EditDummy, EditDummy, FnConnUnd1, FnConnEsc1 },
	{ EditDummy, FnConnLu2, EditDummy, EditDummy, EditDummy, FnConnUnd2, FnConnEsc2 },
	{ EditDummy, EditDummy, FnConnLMv3, EditDummy, EditDummy, FnConnUnd3, FnConnEsc3 },
	{ EditDummy, FnConnLu4, EditDummy, EditDummy, EditDummy, EditDummy, FnConnEsc4 },
	{ EditDummy, FnConnLu5, EditDummy, EditDummy, EditDummy, FnConnUnd5, FnConnEsc5 }
};
const edit_func gEditFnCeng[][MAX_EDFNC] = {
	{ EditDummy, FnChengeLu, EditDummy, EditDummy, FnChengeKey, FnEditUndone, FnChengeEsc }
};
const edit_func gEditFnPaint[][MAX_EDFNC] = {
	{ EditDummy, FnPaintLu, EditDummy, EditDummy, FnPaintKey, FnEditUndone, FnPaintEsc },
	{ EditDummy, FnPaintLu1, EditDummy, EditDummy, EditDummy, FnPaintUnd1, FnPaintEsc1 },
	{ EditDummy, FnPaintLu2, EditDummy, EditDummy, FnPaintKey2, FnPaintUnd2, FnPaintEsc2 },
	{ EditDummy, FnPaintLu3, EditDummy, EditDummy, EditDummy, FnPaintUnd3, FnPaintEsc3 }
};
const edit_func gEditFnGroup[][MAX_EDFNC] = {
	{ EditDummy, FnGroupLu, FnGroupLMv, EditDummy, FnGroupKey, FnEditUndone, FnGroupEsc },
	{ EditDummy, FnGroupLu1, FnGroupLMv1, EditDummy, FnGroupKey1, FnGroupUnd1, FnGroupEsc1 },
	{ EditDummy, FnGroupLu2, EditDummy, EditDummy, EditDummy, EditDummy, FnGroupEsc2 },
	{ EditDummy, FnGroupLu3, EditDummy, EditDummy, EditDummy, EditDummy, FnGroupEsc3 },
	{ EditDummy, FnGroupLu4, EditDummy, FnGroupValue4, FnGroupKey4, FnGroupUnd4, FnGroupEsc4 },
	{ FnGroupMv5, FnGroupLu5, EditDummy, FnGroupValue5, FnGroupKey5, FnGroupUnd5, FnGroupEsc5 }
};
const edit_func gEditFnXs[][MAX_EDFNC] = {
	{ EditDummy, FnCrsInLu, EditDummy, EditDummy, EditDummy, FnCrsInUnd, FnCrsInEsc },
	{ EditDummy, FnCrsInLu1, EditDummy, EditDummy, EditDummy, FnCrsInUnd1, FnCrsInEsc1 }
};
const edit_func gEditFnZmUp[][MAX_EDFNC] = {
	{ EditDummy, FnZoomInLUp1, EditDummy, EditDummy, EditDummy, FnZoomInUnd1, FnZoomInEsc1 },
	{ EditDummy, FnZoomInLUp2, EditDummy, EditDummy, EditDummy, FnZoomInUnd2, FnZoomInEsc2 }
};
const edit_func gEditFnOrig[][MAX_EDFNC] = {
	{ EditDummy, FnOrigLu, EditDummy, FnOrigValu, FnOrigKey, FnOrigUnd, FnOrigEsc }
};
const edit_func gEditFnDist[][MAX_EDFNC] = {
	{ EditDummy, FnDistLu, EditDummy, EditDummy, EditDummy, FnDistUnd, FnDistEsc },
	{ EditDummy, FnDistLu1, EditDummy, EditDummy, EditDummy, FnDistUnd1, FnDistEsc1 },
	{ EditDummy, FnDistLu2, EditDummy, EditDummy, EditDummy, FnDistUnd2, FnDistEsc2 }
};
const edit_func gEditFnCoor[][MAX_EDFNC] = {
	{ EditDummy, FnCoorLu, EditDummy, FnCoorValu, FnCoorKey, FnCoorUnd, FnCoorEsc }
};
const edit_func gEditFnDispNet[][MAX_EDFNC] = {
	{ EditDummy, FnDispNetLu, EditDummy, EditDummy, EditDummy, FnDispNetUnd, FnDispNetEsc }
};
const edit_func gEditFnDispPin[][MAX_EDFNC] = {
	{ EditDummy, FnDispPinLu, EditDummy, EditDummy, EditDummy, FnDispPinUnd, FnDispPinEsc }
};
const edit_func *gEditFunc = gEditFnLand[0];
int gEditLayer = DEF_LAYER;
int gEditLand = DEF_AP;
int gEditFlsh = DEF_AP;
int gEditLineCir = DEF_AP;
int gEditLineSq = DEF_AP;
int gEditPad = 0;
int gEditText = DEF_TEXT_AP;
long gEditFlags = 0;
long gEditFlags1 = 0;
int gEditMode = EDMOD_LINE;
double gRectX = 1.0;
double gRectY = 1.0;
double gRectAng = 0.0;
int gRectAp = 0;
const char *gTextIndex = GetNamePtr(" TEXT");
const char *gPadIndex = GetNamePtr(" PAD");
const char *gPaintedIndex = GetNamePtr(" PAINTED");
const char *gPaintedLoop = GetNamePtr(" LOOP");
const char *gPaintedLines = GetNamePtr(" LINES");
const char *gPaintedMado = GetNamePtr(" WINDOW");
const char *gDrcIndex = GetNamePtr(" DRC");
const char *gDottedLineIndex = GetNamePtr(" DOTTEDLINE");

const char *gNoDrcAttr = GetNamePtr("NO_DRC");
const char gTmpPartsNameBase[] = "tmp";

textin_t gTEXT_IN = { 0.0, DEF_TEXT_HI, DEF_TEXT_WI, DEF_TEXT_AP, 0, "" };
chenge_t gCHENGE = { 0, 0, 0, 0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0,
	{ 1.0, 1.5 }, { 0.0, DEF_TEXT_HI, DEF_TEXT_WI, DEF_TEXT_AP, 0, "" }, { 0 }
} ;
static double gArcSize[3] = { 5.0, 0.0, 360.0 };
static int gUndoneMax = DEF_UNDONE_MAX;
static int gEditKey;
static double gValueX, gValueY;
static pos gClicBase;
static pos gEditBase;
static pos gSeleP1, gSeleP2;
static pos gDistClic;
static node gEditSeles(0);
static node gUndone(0);
static node gReUndone(0);
static node gTmpUndone(0);
static node gCrsPadSave(0);
static node gZoomPadSave(0);
static node gOrgPadSave(0);
static node gCoorPadSave(0);
static node gDistPadSave(0);
static node gDispNetPadSave(0);
static node gDispPinPadSave(0);
static node gHoleBase;
static element *gEditCross = NULL;
static double gConnR = 1.0;
#pragma pack(push, 8)
BYTE gTempMem[TEMP_MEM_SIZE];
#pragma pack(pop)
const void *gTempMemMax = gTempMem + TEMP_MEM_SIZE - TEMP_MEM_OVER;
static BYTE *gTempMemPtr;
static edit_edit_t gEditEditMode;
void *edit_pos::top;

int IsEditSeles()
{
	return gEditCross || !gEditSeles.IsNull();
}

static inline void *ResetMem()
{
	return gTempMemPtr = gTempMem;
}

/*
static void* __fastcall GetMem(size_t size)
{
	BYTE *ptr;
	if ((ptr = gTempMemPtr + size) >= gTempMemMax)
		return NULL;
	return gTempMemPtr = ptr;
}
*/

static inline void *NextMem(size_t size)
{
	return gTempMemPtr += size;
}

static inline void SetMemCount(int n)
{
	*(int *)gTempMem = n;
}

static inline void SetMemCount2(int n)
{
	((int *)gTempMem)[1] = n;
}

static inline int GetMemCount()
{
	return *(int *)gTempMem;
}

static inline int GetMemCount2()
{
	return ((int *)gTempMem)[1];
}

static inline int IsTempMemOver(void *ptr)
{
	return (BYTE *)ptr > gTempMemMax;
}

static inline double GetHitInPoint(element *pe)
{
	return (pe->IsRect() ? ((rect *)pe)->GetDy() / 2.0 : GetAperture(pe)) * 1.5 + GetErrDot();
}

const double &GetClickX()
{
	return gClicBase.x;
}

const double &GetClickY()
{
	return gClicBase.y;
}

int __fastcall GetEditAp(double &ap, double &y)
{
	switch (gEditMode) {
	case EDMOD_LAND:
		ap = GetCirDiam(gEditLand);
		return gEditLand;
	case EDMOD_FLSH:
		if (IsFFlashRect()) {
			ap = gRectX;
			y = gRectY;
			return -1;
		}
		ap = GetSqDiam(gEditFlsh);
		return gEditFlsh;
	case EDMOD_ARC:
	case EDMOD_PARA:
	case EDMOD_NON:
		ap = GetCirDiam(gEditLineCir);
		return gEditLineCir;
	case EDMOD_PAD:
		return gEditPad | 0x80000000;
	case EDMOD_TEXT:
		ap = GetCirDiam(gEditText);
		return gEditText;
	}
	// EDMOD_LINE
	ap = ((element *)gDrawLand.Ref())->IsSq() ? GetSqDiam(gEditLineSq) : GetCirDiam(gEditLineCir);
	return ((element *)gDrawLand.Ref())->IsSq() ? gEditLineSq : gEditLineCir;
}

static void __fastcall DrawEditSeles(HDC hDC, const node &bas)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsElem()) {
			if (gPcb->IsFDisp((element *)pc)) {
				Draw(hDC, (element *)pc);
			}
		} else if (pc->IsGroup()) {
			DrawEditSeles(hDC, ((group *)pc)->GetCont());
		}
	}
}

static void __fastcall DrawEditSelesNorm(HDC hDC, const node &bas)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			DrawEditSelesNorm(hDC, ((group *)pc)->GetCont());
		} else if (pc->IsElem() && gPcb->IsFDisp((element *)pc)) {
			SelectObject(hDC, gPcb->GetPen((element *)pc));
			Draw(hDC, (element *)pc);
		}
	}
}

static void __fastcall DrawEditCross(HDC hDC, const container *pc)
{
	if (pc->IsGroup()) {
		DrawEditSeles(hDC, ((group *)pc)->GetCont());
	} else if (pc->IsElem() && gPcb->IsFDisp((element *)pc)) {
		Draw(hDC, ((element *)pc));
	}
}

static void __fastcall DrawEditCrossNorm(HDC hDC, const container *pc)
{
	if (pc->IsGroup()) {
		DrawEditSelesNorm(hDC, ((group *)pc)->GetCont());
	} else if (pc->IsElem() && gPcb->IsFDisp((element *)pc)) {
		SelectObject(hDC, gPcb->GetPen((element *)pc));
		Draw(hDC, ((element *)pc));
	}
}

static void __fastcall DrawOwnUndone(undone *und)
{
	if (!und->IsNull() && und->IsData()) {
		HDC hDC = GetScrHandle()->GetScrDC();
		SetROP2(hDC, R2_COPYPEN);
		const node &bas = und->GetDataBase();
		for (undone::data *pd = (undone::data *)bas.Ref(); pd != &bas; pd = pd->Ref()) {
			DrawEditCrossNorm(hDC, pd->GetAdd());
		}
		FlushScr();
	}
}

static void __fastcall DrawEditSelesCrossNorm(HDC hDC)
{
	DrawEditCrossNorm(hDC, gEditCross);
	DrawEditSelesNorm(hDC, gEditSeles);
}

static void __fastcall DrawEditSeleCrosSub(HDC hDC, int rop)
{
	SelectObject(hDC, gSelePen);
	SetROP2(hDC, rop);
	if (! gEditSeles.IsNull()) {
		DrawEditSeles(hDC, gEditSeles);
	}
	if (gEditCross) {
		DrawEditCross(hDC, gEditCross);
	}
	if (gEditFlags & F_CONDSP) {
		double x, y;
		Pos2NoSnap(gSeleP1, x, y);
		MoveToScr(hDC, int(x), int(y));
		Pos2NoSnap(gSeleP2, x, y);
		LineToScr(hDC, int(x), int(y));
	}
}

static inline void GetMiniMaxScrCopySize(int sz[], Screen *scr)
{
	pos pl(1e300, 1e300), ph(-1e300, -1e300);
	if (! gEditSeles.IsNull()) {
		AppendElemsMiniMax(gEditSeles, pl, ph);
	}
	if (gEditCross) {
		AppendElemsMiniMax(gEditCross, pl, ph);
	}
	double xl, yl, xh, yh;
	if (IsMirror()) {
		Pos2NoSnap(pl, xh, yh);
		Pos2NoSnap(ph, xl, yl);
	} else {
		Pos2NoSnap(pl, xl, yh);
		Pos2NoSnap(ph, xh, yl);
	}
	sz[0] = xl < 0.0 ? 0 : int(xl);
	sz[1] = yl < 0.0 ? 0 : int(yl);
	if ((sz[2] = int(xh) + 1) > scr->GetScrX()) {
		sz[2] = scr->GetScrX();
	}
	if ((sz[3] = int(yh) + 1) > scr->GetScrY()) {
		sz[3] = scr->GetScrY();
	}
}

void __fastcall DrawEditSeleCros(HDC hDC, int rop)
{
#if 1
	if (rop == R2_XORPEN) {
		HDC hdcTmp;
		PatBlt(hdcTmp = GetScrHandle()->GetTempDC(), 0, 0,
			   GetScrHandle()->GetScrX(), GetScrHandle()->GetScrY(), BLACKNESS);
		DrawEditSeleCrosSub(hdcTmp, R2_MERGEPEN);
		int sz[4];
		if (! gEditSeles.IsNull() || gEditCross) {
			GetMiniMaxScrCopySize(sz, GetScrHandle());
			BitBlt(hDC, sz[0], sz[1], sz[2], sz[3], hdcTmp, sz[0], sz[1], SRCINVERT);
			//			BitBlt(hDC, 0, 0, GetScrHandle()->GetScrX(), GetScrHandle()->GetScrY(), hdcTmp, 0, 0, SRCINVERT);
		}
	} else {
		DrawEditSeleCrosSub(hDC, rop);
	}
#else
	DrawEditSeleCrosSub(hDC, rop);
#endif
}

void __fastcall SetUndoneMax(int n)
{
	if (n > 0) {
		gUndoneMax = n;
	}
}

int GetUndoneMax()
{
	return gUndoneMax;
}

static void ResUndoneSize()
{
	undone *pu = (undone *)gUndone.Prev();
	for (int i = 0; i < gUndoneMax; ++i) {
		if (pu == &gUndone) {
			return;
		}
		pu = pu->Prev();
	}
	while (pu != &gUndone) {
		undone *del = pu;
		pu = pu->Prev();
		del->Del();
	}
}

int IsUndoneNull()
{
	return gUndone.IsNull();
}

int IsReUndone()
{
	return ! gReUndone.IsNull();
}

int IsReUndoneNull()
{
	return gReUndone.IsNull();
}

int IsUndoneOn()
{
	return gEditFunc[EDFN_UND] != EditDummy && ! gTmpUndone.IsNull();
}

void ReUndone()
{
	if (! gReUndone.IsNull()) {
		gUndone.Push(((undone *)gReUndone.Prev())->ReUndone());
		GetScrHandle()->Invalidate();
		SetFFileChenge();
	}
}

static void __fastcall DrawGrp(HDC hDC, const node &bas, int f = 0)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			DrawGrp(hDC, ((group *)pc)->GetCont());
		} else if (pc->IsElem() && gPcb->IsFDisp((element *)pc)) {
			if (f == 0) {
				SelectObject(hDC, gPcb->GetPen((element *)pc));
			}
			Draw(hDC, (element *)pc);
		}
	}
}

static void __fastcall DrawEditSeles(int f = 0)
{
	HDC hDC = GetScrHandle()->GetScrDC();
	SetROP2(hDC, R2_BLACK);
	DrawGrp(hDC, gEditSeles, f);
	FlushScr();
}

static void __fastcall DrawUndone(HDC hDC, const node &bas)
{
	for (undone::data *pd = (undone::data *)bas.Ref(); pd != &bas; pd = pd->Ref()) {
		if (pd->GetOrg()) {
			if (pd->GetAdd()->IsGroup()) {
				DrawGrp(hDC, ((group *)pd->GetAdd())->GetCont());
			} else if (pd->GetAdd()->IsElem() && gPcb->IsFDisp((element *)pd->GetAdd())) {
				SelectObject(hDC, gPcb->GetPen((element *)pd->GetAdd()));
				Draw(hDC, (element *)pd->GetAdd());
			}
		}
	}
}

static void __fastcall DrawUndoneErase(HDC hDC, const node &bas)
{
	for (undone::data *pd = (undone::data *)bas.Ref(); pd != &bas; pd = pd->Ref()) {
		if (!pd->GetOrg()) {
			if (pd->GetAdd()->IsGroup()) {
				DrawGrp(hDC, ((group *)pd->GetAdd())->GetCont(), 1);
			} else if (pd->GetAdd()->IsElem() && gPcb->IsFDisp((element *)pd->GetAdd())) {
				Draw(hDC, (element *)pd->GetAdd());
			}
		}
	}
}

static void DrawTmpUndoneElem()
{
	HDC hDC = GetScrHandle()->GetScrDC();
	DrawEditSeleCros(hDC, R2_BLACK);
	SetROP2(hDC, R2_BLACK);
	for (undone *und = (undone *)gTmpUndone.Ref(); und != &gTmpUndone; und = und->Ref()) {
		DrawUndoneErase(hDC, und->GetDataBase());
	}
	SetROP2(hDC, R2_MERGEPEN);
	for (undone *und = (undone *)gTmpUndone.Ref(); und != &gTmpUndone; und = und->Ref()) {
		DrawUndone(hDC, und->GetDataBase());
	}
	FlushScr();
}

static void UndoneOwn()
{
	if (!gUndone.IsNull()) {
		gReUndone.Push(((undone *)gUndone.Prev()->Cut())->Undone());
		GetScrHandle()->Invalidate();
		SetFFileChenge();
	}
}

void __fastcall AddOwnUndone(undone *und)
{
	if (und->IsData()) {
		while (! gReUndone.IsNull()) {
			((undone *)gReUndone.Prev())->Del();
		}
		gUndone.Push(und);
		ResUndoneSize();
	} else {
		und->Del();
	}
}

static inline void SetOwnTmpUndone(undone *und)
{
	if (und->IsData()) {
		gTmpUndone.Push(und);
	} else {
		und->Del();
	}
}

static int __fastcall IsInElem(element *el, group *pg)
{
	const node &bas = pg->GetCont();
	for (element *pe = (element *)bas.Ref(); pe != &bas; pe = pe->Ref()) {
		if (pe->IsGroup()) {
			return IsInElem(el, (group *)pe);
		}
		if (pe == el) {
			return 1;
		}
	}
	return 0;
}

void ClearUndone()
{
	while (!gTmpUndone.IsNull()) {
		((undone *)gTmpUndone.Prev())->Del();
	}
	while (!gUndone.IsNull()) {
		((undone *)gUndone.Prev())->Del();
	}
	while (!gReUndone.IsNull()) {
		((undone *)gReUndone.Prev())->Del();
	}
}

void __fastcall RunEditKey(int key)
{
	gEditKey = key;
	GetScrHandle()->CallEditFunc(gEditFunc[EDFN_KEY]);
}

static void __fastcall RunEditValue(const double &x, const double &y)
{
	if (x > -1e16 && x < 1e16 && y > -1e16 && y < 1e16) {
		gValueX = x;
		gValueY = y;
		gEditFunc[EDFN_VALIN]();
		FlushScr();
	} else {
		MessageBeep(MB_OK);
	}
}

static int __fastcall GetCrossLineLinePos(line *pl1, line *pl2, pos &p)
{
	double dx1 = pl1->GetPe().x - pl1->GetPos().x;
	double dy1 = pl1->GetPe().y - pl1->GetPos().y;
	double dx2 = pl2->GetPe().x - pl2->GetPos().x;
	double dy2 = pl2->GetPe().y - pl2->GetPos().y;
	if (fabs(dx1) < ERR_MINI) {
		if (fabs(dx2) < ERR_MINI) {
			if (pl1->GetAp() == pl2->GetAp() && pl1->GetLy() == pl2->GetLy()
					&& fabs(pl1->GetPos().x - pl2->GetPos().x) < ERR_MINI) {
				return 2;
			}
			return 1;
		}
		if (fabs(dy2) < ERR_MINI) {
			p.Set(pl1->GetPos().x, pl2->GetPos().y);
		} else {
			p.Set(pl1->GetPos().x, pl2->GetPos().y + (pl1->GetPos().x - pl2->GetPos().x) * dy2 / dx2);
		}
	} else if (fabs(dx2) < ERR_MINI) {
		if (fabs(dy1) < ERR_MINI) {
			p.Set(pl2->GetPos().x, pl1->GetPos().y);
		} else {
			p.Set(pl2->GetPos().x, pl1->GetPos().y + (pl2->GetPos().x - pl1->GetPos().x) * dy1 / dx1);
		}
	} else {
		double a1 = dy1 / dx1, a2 = dy2 / dx2;
		if (fabs(a1 - a2) < ERR_MINI) {
			if (pl1->GetAp() != pl2->GetAp() || pl1->GetLy() != pl2->GetLy()) {
				return 1;
			}
			if (fabs(dy1) < ERR_MINI) {
				if (fabs(dy2) < ERR_MINI && fabs(pl1->GetPos().y - pl2->GetPos().y) < ERR_MINI) {
					return 2;
				}
			} else if (fabs(dy2) >= ERR_MINI && IsEq(pl1->GetPos().y - pl1->GetPos().x * a1,
					   pl2->GetPos().y - pl2->GetPos().x * a2)) {
				return 2;
			}
			return 1;
		}
		double x = (pl2->GetPos().y - pl1->GetPos().y - a2 * (pl2->GetPos().x - pl1->GetPos().x)) / (a1 - a2);
		p.Set(pl1->GetPos().x + x, pl1->GetPos().y + a1 * x);
	}
	return 0;
}

static int __fastcall GetCrossLineArcPos(line *pl, arc *pa, pos p[])
{
	double dx = pl->GetPe().x - pl->GetPos().x, dy = pl->GetPe().y - pl->GetPos().y;
	if (fabs(dx) < ERR_MINI) {
		dx = pl->GetPos().x - pa->GetPos().x;
		dy = pa->GetR() - fabs(dx);
		if (dy < -ERR_MINI) {
			return 1;
		}
		if (dy < ERR_MINI) {
			dy = 0.0;
		} else {
			dy = sqrt(pa->GetR() * pa->GetR() - dx * dx);
		}
		p[0].Set(pl->GetPos().x, pa->GetPos().y + dy);
		p[1].Set(pl->GetPos().x, pa->GetPos().y - dy);
	} else if (fabs(dy) < ERR_MINI) {
		dy = pl->GetPos().y - pa->GetPos().y;
		dx = pa->GetR() - fabs(dy);
		if (dx < -ERR_MINI) {
			return 1;
		}
		if (dx < ERR_MINI) {
			dx = 0.0;
		} else {
			dx = sqrt(pa->GetR() * pa->GetR() - dy * dy);
		}
		p[0].Set(pa->GetPos().x + dx, pl->GetPos().y);
		p[1].Set(pa->GetPos().x - dx, pl->GetPos().y);
	} else {
		double a = dy / dx;
		double b = (pl->GetPos().y - pa->GetPos().y) - a * (pl->GetPos().x - pa->GetPos().x);
		double a2 = a * a;
		double b2 = b * b;
		double a1_2 = a2 + 1;
		double tmp = a2 * b2 - a1_2 * (b2 - pa->GetR() * pa->GetR());
		if (tmp < -ERR_MINI) {
			return 1;
		} else if (tmp < ERR_MINI) {
			tmp = 0.0;
		} else {
			tmp = sqrt(tmp);
		}
		double x = (-a * b + tmp) / a1_2;
		p[0].Set(pa->GetPos().x + x, pa->GetPos().y + a * x + b);
		x = (-a * b - tmp) / a1_2;
		p[1].Set(pa->GetPos().x + x, pa->GetPos().y + a * x + b);
	}
	return 0;
}

static int __fastcall GetCrossArcArcPos(arc *pa1, arc *pa2, pos p[])
{
	if (pa1->GetPos().IsEq(pa2->GetPos())) {
		if (fabs(pa1->GetR() - pa2->GetR()) < ERR_MINI && pa1->GetLy() == pa2->GetLy() && pa1->GetAp() == pa2->GetAp()) {
			return 2;
		}
		return 1;
	}
	double l = pa1->GetPos().Distance(pa2->GetPos());
	if (l > pa1->GetR() + pa2->GetR() || fabs(pa1->GetR() - pa2->GetR()) > l) {
		return 1;
	} else {
		double lx = (l * l + pa1->GetR() * pa1->GetR() - pa2->GetR() * pa2->GetR()) / (2.0 * l);
		double ang1 = acos(lx / pa1->GetR());
		double ang0 = atan2(pa2->GetPos().y - pa1->GetPos().y, pa2->GetPos().x - pa1->GetPos().x);
		double ang = ang0 + ang1;
		p[0].Set(pa1->GetPos().x + pa1->GetR() * cos(ang), pa1->GetPos().y + pa1->GetR() * sin(ang));
		ang = ang0 - ang1;
		p[1].Set(pa1->GetPos().x + pa1->GetR() * cos(ang), pa1->GetPos().y + pa1->GetR() * sin(ang));
	}
	return 0;
}

static inline void GetMouseSnapPos(pos &p)
{
	Scr2Pos(GetScrHandle()->GetMouseX(), GetScrHandle()->GetMouseY(), p);
}

static inline void GetMouseHitPos(pos &p)
{
	Scr2NoSnap(GetScrHandle()->GetMouseX(), GetScrHandle()->GetMouseY(), p);
}

static int __fastcall IsInBoxGroup(const pos &pl, const pos &ph, const node &bas)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			if (IsInBoxGroup(pl, ph, ((group *)pc)->GetCont())) {
				return 1;
			}
		} else if (pc->IsElem() && gPcb->IsFEdit((element *)pc) && ((element *)pc)->IsIn(pl, ph)) {
			return 1;
		}
	}
	return 0;
}

static undone *__fastcall GetInBoxElems(pos pl, pos ph, node &in, node &bas)
{
	if (pl.x > ph.x) {
		Swap(pl.x, ph.x);
	}
	if (pl.y > ph.y) {
		Swap(pl.y, ph.y);
	}
	undone *und = undone::New();
	for (container *p1 = (container *)bas.Ref(); p1 != &bas;) {
		container *pc = p1;
		p1 = p1->Ref();
		if (pc->IsGroup()) {
			if (IsInBoxGroup(pl, ph, ((group *)pc)->GetCont())) {
				und->AddData(pc, pc->Ref());
				in.Add(pc->Cut());
			}
		} else if (pc->IsElem() && gPcb->IsFEdit((element *)pc) && ((element *)pc)->IsIn(pl, ph)) {
			und->AddData(pc, pc->Ref());
			in.Add(pc->Cut());
		}
	}
	if (! und->GetDataBase().IsNull()) {
		return und;
	}
	und->Del();
	return NULL;
}

static void EditDummy()	// pcbe.cpp ŎQ
{
}

static void FnEditUndone()
{
	UndoneOwn();
}

static void FnLandMv()
{
	pos p;
	Scr2Pos(GetScrHandle()->GetMouseX(), GetScrHandle()->GetMouseY(), p);
	element *li = (element *)gDrawLand.Ref();
	li->MoveAbs(p);
}

int __fastcall IsOutSize(const node &bas)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			if (IsOutSize(((group *)pc)->GetCont())) {
				return 1;
			}
		} else if (pc->IsElem()
				   && (((element *)pc)->GetLo().x < 0.0 || ((element *)pc)->GetLo().y < 0.0
					   || ((element *)pc)->GetHi().x > GetSizeX() || ((element *)pc)->GetHi().y > GetSizeY())) {
			return 1;
		}
	}
	return 0;
}

int __fastcall IsOutSize(element *pe)
{
	if (pe->IsGroup()) {
		return IsOutSize(((group *)pe)->GetCont());
	}
	return pe->GetLo().x < 0.0 || pe->GetLo().y < 0.0 || pe->GetHi().x > GetSizeX() || pe->GetHi().y > GetSizeY();
}

static void __fastcall AddPcbOwnElem(element *pe)
{
	gPcb->Add(pe);
	AddOwnUndone(undone::New(pe));
	SetFFileChenge();
	HDC hDC = GetScrHandle()->GetScrDC();
	SelectObject(hDC, gPcb->GetPen(pe));
	SetROP2(hDC, R2_MERGEPEN);
	Draw(hDC, pe);
}

static void FnLandLu()
{
	pos p;
	GetMouseSnapPos(p);
	land *pl;
	if (IsOutSize(pl = land::New(gEditLayer, gEditLand, 0, p))) {
		pl->Del();
		MessageBeep(MB_OK);
	} else {
		AddPcbOwnElem(pl);
	}
}

static void FnLandValu()
{
	land *pl;
	if (IsOutSize(pl = land::New(gEditLayer, gEditLand, 0, pos(gValueX, gValueY)))) {
		pl->Del();
		MessageBeep(MB_OK);
	} else {
		AddPcbOwnElem(pl);
	}
}

static void __fastcall EdKeyValSub(int f)
{
	static valuein_t val = { 0.0, 0.0, 1 };
	if (val.f != f) {
		val.f = f;
		val.x = val.y = 0.0;
	}
	switch (CallDialogBoxParam(gApp->Handle(), IDD_VALUIN, ValuInDlg, (LPARAM)&val)) {
	case 0:
		RunEditValue(val.x, val.y);
		break;
	case 1:
		RunEditValue(GetOrigX() + val.x, GetOrigY() + val.y);
		break;
	case 2:
		RunEditValue(gClicBase.x + val.x, gClicBase.y + val.y);
		break;
	}
}

static void FnLandKey()
{
	apsele_t ap;
	switch (gEditKey) {
	case EDKEY_LY:
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 1)) {
			GetScrHandle()->Invalidate();
		}
		((land *)gDrawLand.Ref())->ChengeLy(gEditLayer);
		SetLyApMenu();
		break;
	case EDKEY_SEL:
		ap.sq = 0;
		ap.pap = &gEditLand;
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELELAND, ApSeleDlg, (LPARAM)&ap)) {
			((element *)gDrawLand.Ref())->ChengeAp(gEditLand);
			SetLyApMenu();
		}
		break;
	case EDKEY_VAL:
		EdKeyValSub(1);
		break;
	}
}

static void FnSetEsc()
{
	SetButtonNormal();
	ClearDraws();
	GetScrHandle()->OffLdMove();
}

static void __fastcall FnSetEsc(pos &p)
{
	SetButtonNormal();
	ClearDraws();
	GetMouseSnapPos(p);
	GetScrHandle()->OffLdMove();
}

static void FnLandEsc()
{
	pos p;
	FnSetEsc(p);
	ResetStack(MOD_LAND, IDST_LAND);
	gApp->PushButton(IDB_CIRLAND, TRUE);
	gApp->UpdateButton();
	gDrawLand.Add(land::New(gEditLayer, gEditLand, 0, p));
	gEditMode = EDMOD_LAND;
	SetLyApMenu();
}

static void FnFlshLu()
{
	pos p;
	GetMouseSnapPos(p);
	element *el;
	if (IsFFlashRect()) {
		el = rect::New(gEditLayer, gRectAp, p, gRectX, gRectY, gRectAng);
	} else {
		el = land::New(gEditLayer, gEditFlsh, 1, p);
	}
	if (IsOutSize(el)) {
		el->Del();
		MessageBeep(MB_OK);
	} else {
		AddPcbOwnElem(el);
	}
}

static void FnFlshValu()
{
	element *el;
	if (IsFFlashRect()) {
		el = rect::New(gEditLayer, gRectAp, pos(gValueX, gValueY), gRectX, gRectY, gRectAng);
	} else {
		el = land::New(gEditLayer, gEditFlsh, 1, pos(gValueX, gValueY));
	}
	if (IsOutSize(el)) {
		el->Del();
		MessageBeep(MB_OK);
	} else {
		AddPcbOwnElem(el);
	}
}

static void FnFlshKey()
{
	apsele_t ap;
	int rct;
	switch (gEditKey) {
	case EDKEY_LY:
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 1)) {
			GetScrHandle()->Invalidate();
		}
		((land *)gDrawLand.Ref())->ChengeLy(gEditLayer);
		SetLyApMenu();
		break;
	case EDKEY_SEL:
		ap.sq = 1;
		ap.pap = &gEditFlsh;
		rct = IsFFlashRect();
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELEFLASH, FlashSeleDlg, (LPARAM)&ap)) {
			if (rct != IsFFlashRect()) {
				if (!gDrawLand.IsNull()) {
					((element *)gDrawLand.Ref())->Del();
				}
				pos p;
				Scr2Pos(GetScrHandle()->GetMouseX(), GetScrHandle()->GetMouseY(), p);
				if (IsFFlashRect()) {
					gDrawLand.Add(rect::New(gEditLayer, gRectAp, p, gRectX, gRectY, gRectAng));
				} else {
					gDrawLand.Add(land::New(gEditLayer, gEditFlsh, 1, p));
				}
			} else {
				if (IsFFlashRect()) {
					((rect *)gDrawLand.Ref())->ChengeSize(gRectX, gRectY, gRectAng);
				} else {
					((element *)gDrawLand.Ref())->ChengeAp(gEditFlsh);
				}
			}
		}
		SetLyApMenu();
		break;
	case EDKEY_VAL:
		EdKeyValSub(1);
		break;
	}
}

static void FnFlshEsc()
{
	pos p;
	FnSetEsc(p);
	ResetStack(MOD_SQ, IDST_LAND);
	gApp->PushButton(IDB_SQLAND, TRUE);
	gApp->UpdateButton();
	if (IsFFlashRect()) {
		gDrawLand.Add(rect::New(gEditLayer, gRectAp, p, gRectX, gRectY, gRectAng));
	} else {
		gDrawLand.Add(land::New(gEditLayer, gEditFlsh, 1, p));
	}
	gEditMode = EDMOD_FLSH;
	SetLyApMenu();
}

static int __fastcall GetLineHitPos(pos &p, const node &bas, int ly)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsElem() && ((element *)pc)->GetLy() == ly) {
			if (gPcb->IsFEdit((element *)pc) && ((element *)pc)->IsHit(p)) {
				double err = GetHitInPoint((element *)pc);
				if (pc->IsLand()) {
					p = ((land *)pc)->GetPos();
					return 0;
				} else if (pc->IsLine()) {
					if (p.Distance2(((line *)pc)->GetPos()) < p.Distance2(((line *)pc)->GetPe())) {
						if (p.Distance(((line *)pc)->GetPos()) < err) {
							p = ((line *)pc)->GetPos();
							return 0;
						}
					} else {
						if (p.Distance(((line *)pc)->GetPe()) < err) {
							p = ((line *)pc)->GetPe();
							return 0;
						}
					}
				} else if (pc->IsRect()) {
					if (p.Distance2(((rect *)pc)->GetP1()) < p.Distance2(((rect *)pc)->GetP2())) {
						if (p.Distance2(((rect *)pc)->GetP1()) < p.Distance2(((rect *)pc)->GetPos())) {
							if (p.Distance(((rect *)pc)->GetP1()) < err) {
								p = ((rect *)pc)->GetP1();
								return 0;
							}
						} else {
							if (p.Distance(((rect *)pc)->GetPos()) < err) {
								p = ((rect *)pc)->GetPos();
								return 0;
							}
						}
					} else {
						if (p.Distance2(((rect *)pc)->GetP2()) < p.Distance2(((rect *)pc)->GetPos())) {
							if (p.Distance(((rect *)pc)->GetP2()) < err) {
								p = ((rect *)pc)->GetP2();
								return 0;
							}
						} else {
							if (p.Distance(((rect *)pc)->GetPos()) < err) {
								p = ((rect *)pc)->GetPos();
								return 0;
							}
						}
					}
				} else if (pc->IsArc()) {
					if (p.Distance2(((arc *)pc)->GetStPos()) < p.Distance2(((arc *)pc)->GetEndPos())) {
						if (p.Distance(((arc *)pc)->GetStPos()) < err) {
							p = ((arc *)pc)->GetStPos();
							return 0;
						}
					} else {
						if (p.Distance(((arc *)pc)->GetEndPos()) < err) {
							p = ((arc *)pc)->GetEndPos();
							return 0;
						}
					}
				}
			}
		} else if (pc->IsGroup() && GetLineHitPos(p, ((group *)pc)->GetCont(), ly) == 0) {
			return 0;
		}
	}
	return 1;
}

static inline pos GetLineHitPos(const double &x, const double &y)
{
	pos p;
	Scr2NoSnap(x, y, p);
	if ((GetScrHandle()->GetMouseKey() & MK_CONTROL) || GetLineHitPos(p, gPcb->GetElems(), gEditLayer)) {
		Scr2Pos(x, y, p);
	}
	return p;
}

static void FnLineLu1()
{
	if (IsFLineHitHit()) {
		GetMouseSnapPos(gClicBase);
	} else {
		gClicBase = GetLineHitPos(GetScrHandle()->GetMouseX(), GetScrHandle()->GetMouseY());
	}
	((line *)gDrawLand.Ref())->MoveAbs(gClicBase);
	if (IsOutSize((element *)gDrawLand.Ref())) {
		MessageBeep(MB_OK);
	} else {
		gDrawLine.Add(gDrawLand.Ref()->Cut());
		if (IsFRPopup() || IsFRBs()) {
			PushStack(1, IDST_LINE2);
		} else {
			PushStack(1, IDST_LINE3);
		}
		SetButtonOnState();
	}
}

static void FnLineValu1()
{
	((element *)gDrawLand.Ref())->MoveAbs(pos(gValueX, gValueY));
	if (IsOutSize((element *)gDrawLand.Ref())) {
		((element *)gDrawLand.Ref())->MoveAbs(pos(GetScrHandle()->GetMouseX(), GetScrHandle()->GetMouseY()));
		MessageBeep(MB_OK);
	} else {
		gDrawLine.Add(gDrawLand.Ref()->Cut());
		PushStack(1, IDST_LINE2);
		SetButtonOnState();
		gClicBase.Set(gValueX, gValueY);
		FnLineMv2();
	}
}

static void FnLineKey1()
{
	switch (gEditKey) {
	case EDKEY_LY:
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 1)) {
			GetScrHandle()->Invalidate();
		}
		((line *)gDrawLand.Ref())->ChengeLy(gEditLayer);
		SetLyApMenu();
		break;
	case EDKEY_SEL:
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELELINE, LineSelDlg, 0)) {
			((element *)gDrawLand.Ref())->Del();
			pos p;
			GetMouseSnapPos(p);
			gDrawLand.Add(line::New(gEditLayer, IsFLineSq() ? gEditLineSq : gEditLineCir, IsFLineSq() ? 1 : 0, p, p));
			SetLyApMenu();
		}
		break;
	case EDKEY_VAL:
		EdKeyValSub(1);
		break;
	}
}

static void FnLineEsc()
{
	pos p;
	FnSetEsc(p);
	ResetStack(MOD_LINE, IDST_LINE1);
	gApp->PushButton(IDB_LINE, TRUE);
	gApp->UpdateButton();
	gDrawLand.Add(line::New(gEditLayer, IsFLineSq() ? gEditLineSq : gEditLineCir, IsFLineSq() ? 1 : 0, p, p));
	gEditMode = EDMOD_LINE;
	SetLyApMenu();
}

static inline int IsEditAng8AndKey()
{
	return !(GetScrHandle()->GetMouseKey() & MK_CONTROL)
			&& (IsFEditAng8() || ((GetScrHandle()->GetMouseKey() & MK_SHIFT)));
}

static void __fastcall GetLineMousePos8(pos &p, pos pb)
{
	double dx = p.x - pb.x;
	double dy = p.y - pb.y;
	double ddx = fabs(dx), ddy = fabs(dy);
	double l0 = pb.Distance(p);
	if (ddx < ddy) {
		if (ddy * sqrt(1.0 + 0.5 * 0.5) < l0) {
			if (dy < 0.0) {
				p.y = pb.y - ddx;
			} else {
				p.y = pb.y + ddx;
			}
		} else {
			p.x = pb.x;
		}
	} else {
		if (ddx * sqrt(1.0 + 0.5 * 0.5) < l0) {
			if (dx < 0.0) {
				p.x = pb.x - ddy;
			} else {
				p.x = pb.x + ddy;
			}
		} else {
			p.y = pb.y;
		}
	}
}

static void __fastcall GetLineMousePos(pos &p)
{
	GetMouseSnapPos(p);
	if (IsEditAng8AndKey()) {
		GetLineMousePos8(p, ((line*)gDrawLine.Ref())->GetPos());
	}
}

static void __fastcall GetLineMousePos1(pos &p)
{
	GetMouseSnapPos(p);
	if (IsFEditAng8() || ((GetScrHandle()->GetMouseKey() & MK_SHIFT))) {
		GetLineMousePos8(p, ((line*)gDrawLine.Ref())->GetPos());
	}
}

static double __fastcall GetApHitLen(element *pe)
{
	if (pe->IsRect()) {
		return ((rect*)pe)->GetDy() * 1.5 / 2.0;
	}
	return pe->IsSq() ? GetSqAp(pe->GetAp()) : GetCirAp(pe->GetAp()) * 1.5;
}

static pos __fastcall GetAng8(const pos &pb, pos p)
{
	if (!(GetScrHandle()->GetMouseKey() & MK_CONTROL)) {
		pos pm;
		GetMouseHitPos(pm);
		if (!IsFLineHitHit()) {
			element *pe = gPcb->IsHit(pm);
			if (pe && ((line*)gDrawLine.Ref())->GetLy() == pe->GetLy()
					&& pm.Distance(pe->GetNearPos(pm)) < GetApHitLen(pe)) {
				return pe->GetNearPos(pm);
			}
		}
		if (IsEditAng8AndKey()) {
			GetLineMousePos8(p, pb);
		}
	}
	return p;
}

static void FnLineLu2()
{
	pos p0;
	if (IsFLineHitHit()) {
		GetMouseSnapPos(p0);
	} else {
		p0 = GetLineHitPos(GetScrHandle()->GetMouseX(), GetScrHandle()->GetMouseY());
	}
	pos p = GetAng8(((line *)gDrawLine.Ref())->GetPos(), p0);
	((line *)gDrawLine.Ref())->Expand(((line *)gDrawLine.Ref())->GetPe(), p);
	if (((line *)gDrawLine.Ref())->GetPos().IsEq(((line *)gDrawLine.Ref())->GetPe())
			|| IsOutSize((element *)gDrawLine.Ref())) {
		MessageBeep(MB_OK);
	} else {
		line *li1;
		gPcb->Add(li1 = ((line *)gDrawLine.Ref())->Cut());
		SetOwnTmpUndone(undone::New(li1));
		HDC hDC = GetScrHandle()->GetScrDC();
		SelectObject(hDC, gPcb->GetPen(li1));
		SetROP2(hDC, R2_MERGEPEN);
		Draw(hDC, li1);
		GetMouseSnapPos(gClicBase);
		gDrawLine.Add(line::New(gEditLayer,
								IsFLineSq() ? gEditLineSq : gEditLineCir, IsFLineSq() ? 1 : 0, li1->GetPe(), gClicBase));
	}
}

static void FnLineMv2()
{
	pos p;
	GetLineMousePos(p);
	((line *)gDrawLine.Ref())->Expand(((line *)gDrawLine.Ref())->GetPe(), p);
}

static undone *__fastcall MoveTmpUndoToUndone(int m = 0)
{
	undone *pu = undone::New();
	while (! gTmpUndone.IsNull()) {
		pu->GetAllData((undone *)gTmpUndone.Ref()->Cut())->Del();
	}
	if (pu->GetDataBase().IsNull()) {
		pu->Del();
		pu = NULL;
	} else if (m == 0) {
		AddOwnUndone(pu);
	}
	return pu;
}

static void FnLineRtn2()
{
	MoveTmpUndoToUndone();
	line *li;
	gDrawLand.Add(li = (line *)gDrawLine.Ref()->Cut());
	li->Expand(li->GetPos(), li->GetPe());
	PopStack();
	SetFFileChenge();
}

static void __fastcall DrawLineUndone(HDC hDC, const node &bas)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			DrawLineUndone(hDC, ((group *)pc)->GetCont());
		} else if (pc->IsElem()) {
			Draw(hDC, (element *)pc);
		}
	}
}

static void FnLineValu2()
{
	((line *)gDrawLine.Ref())->Expand(((line *)gDrawLine.Ref())->GetPe(), pos(gValueX, gValueY));
	if (((line *)gDrawLine.Ref())->GetPos().IsEq(((line *)gDrawLine.Ref())->GetPe())
			|| IsOutSize((element *)gDrawLine.Ref())) {
		FnLineMv2();
		MessageBeep(MB_OK);
	} else {
		line *li, *li1;
		gPcb->Add(li1 = (li = (line *)gDrawLine.Ref())->Copy());
		pos p;
		if (!(p = GetAng8(li1->GetPos(), pos(gValueX, gValueY))).IsEq(li->GetPe())) {
			li1->Expand(li1->GetPe(), p);
		}
		SetOwnTmpUndone(undone::New(li1));
		HDC hDC = GetScrHandle()->GetScrDC();
		SelectObject(hDC, gPcb->GetPen(li));
		SetROP2(hDC, R2_MERGEPEN);
		Draw(hDC, li1);
		li->Expand(li->GetPos(), p);
		gClicBase.Set(gValueX, gValueY);
	}
}

static void FnLineKey2()
{
	switch (gEditKey) {
	case EDKEY_LY:
		int rt;
		rt = CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 1);
		((line *)gDrawLine.Ref())->ChengeLy(gEditLayer);
		if (rt) {
			GetScrHandle()->Invalidate();
		}
		break;
	case EDKEY_SEL:
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELELINE, LineSelDlg, 1) == 2) {
			((element *)gDrawLine.Ref())->ChengeAp(IsFLineSq() ? gEditLineSq : gEditLineCir);
		}
		break;
	case EDKEY_VAL:
		EdKeyValSub(0);
		break;
	case EDKEY_RTN:
		FnLineRtn2();
		break;
	}
}

static void FnLineUnd2()
{
	if (! gTmpUndone.IsNull()) {
		undone *pu = (undone *)gTmpUndone.Prev();
		const node &nb = pu->GetDataBase();
		if (((undone::data *)nb.Prev())->GetAdd()->IsLine()) {
			line *li = (line *)((undone::data *)nb.Prev())->GetAdd()->Copy();
			container *pc = (container*)gDrawLine.Ref();
			pc->Del();
			gDrawLine.Add(li);
			pos p;
			GetLineMousePos1(p);
			li->Expand(li->GetPe(), p);
			gClicBase = li->GetPos();
		} else {
			// C[߂
		}
		HDC hDC = GetScrHandle()->GetScrDC();
		SetROP2(hDC, R2_BLACK);
		for (undone::data *pud = (undone::data *)nb.Prev(); pud != &nb; pud = pud->Prev()) {
			container *pc;
			if ((pc = pud->GetAdd())->IsElem()) {
				Draw(hDC, (element *)pc);
			} else if (pc->IsGroup()) {
				DrawLineUndone(hDC, ((group *)pc)->GetCont());
			}
		}
		FlushScr();
		pu->Undone()->Del();
		gEditLayer = ((element *)gDrawLine.Ref())->GetLy();
		*(IsFLineSq() ? &gEditLineSq : &gEditLineCir) = ((element *)gDrawLine.Ref())->GetAp();
	} else {
		line *li = (line *)gDrawLine.Ref();
		li->Expand(li->GetPos(), li->GetPe());
		gDrawLand.Add(gDrawLine.Ref()->Cut());
		PopStack();
	}
}

static void FnLineEsc2()
{
	line *li;
	gDrawLand.Add((li = (line *)gDrawLine.Ref()->Cut()));
	li->Expand(li->GetPos(), li->GetPe());
	HDC hDC = GetScrHandle()->GetScrDC();
	SetROP2(hDC, R2_BLACK);
	while (! gTmpUndone.IsNull()) {
		undone *pu = (undone *)gTmpUndone.Prev();
		const node &nb = pu->GetDataBase();
		for (undone::data *pud = (undone::data *)nb.Prev(); pud != &nb; pud = pud->Prev()) {
			container *pc;
			if ((pc = pud->GetAdd())->IsElem()) {
				Draw(hDC, (element *)pc);
			} else if (pc->IsGroup()) {
				DrawLineUndone(hDC, ((group *)pc)->GetCont());
			}
		}
		pu->Undone()->Del();
	}
	PopStack();
}

static void __fastcall ArcSetSnt(double *siz, pos p)
{
	if (siz[0] > STEP_MINI && GetCirAp(gEditLineCir) + ERR_MINI < siz[0]) {
		double *lo, *hi;
		if (siz[0] <= siz[1]) {
			lo = &siz[0];
			hi = &siz[1];
		} else {
			lo = &siz[1];
			hi = &siz[0];
		}
		while (*hi - *lo > 360.0) {
			*hi -= 360.0;
		}
		while (*hi > 360.0) {
			*lo -= 360.0;
			*hi -= 360.0;
		}
		while (*lo <= -360.0) {
			*lo += 360.0;
			*hi += 360.0;
		}
		arc *pa;
		if (IsOutSize(pa = arc::New(gEditLayer, gEditLineCir, p,
									siz[0], siz[1] * M_PI / 180.0, siz[2] * M_PI / 180.0))) {
			pa->Del();
			MessageBeep(MB_OK);
		} else {
			gPcb->Add(pa);
			AddOwnUndone(undone::New(pa));
			SetFFileChenge();
			HDC hDC = GetScrHandle()->GetScrDC();
			SelectObject(hDC, gPcb->GetPen(gEditLayer));
			SetROP2(hDC, R2_MERGEPEN);
			Draw(hDC, pa);
		}
	} else {
		MessageBeep(MB_OK);
	}
}

static void DrawArcSenter()
{
	double xx, yy;
	Pos2NoSnap(gClicBase, xx, yy);
	if (xx >= 0.0 && xx <= GetScrHandle()->GetScrX() && yy >= 0.0 && yy <= GetScrHandle()->GetScrY()) {
		int x = int(xx), y = int(yy);
		HDC hDC = GetScrHandle()->GetScrDC();
		SelectObject(hDC, gSelePen);
		SetROP2(hDC, R2_COPYPEN);
		MoveToEx(hDC, x - 2, y - 2, NULL);
		LineTo(hDC, x + 3, y + 3);
		MoveToEx(hDC, x + 2, y - 2, NULL);
		LineTo(hDC, x - 3, y + 3);
	}
}

static void FnArcLu()
{
	if (IsFEditArcFree()) {
		PushStack(1, IDST_ARC1);
		SetButtonOnState();
		gApp->EnableButton(IDB_LAYER, FALSE);
		gApp->EnableButton(IDB_SELECT, FALSE);
		EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_GRAYED);
		EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_AP, MF_GRAYED);
		DrawMenuBar(GetAppWnd());
		gApp->UpdateButton();
		GetMouseSnapPos(gClicBase);
		DrawArcSenter();
	} else if (CallDialogBoxParam(gApp->Handle(), IDD_ARCIN, ArcInDlg, (LPARAM)gArcSize) == 0) {
		pos p;
		GetMouseSnapPos(p);
		ArcSetSnt(gArcSize, p);
	}
}

static void FnArcValu()
{
	if (! IsFEditArcFree() || gValueX < 0.0 || gValueX > GetSizeX() || gValueY < 0.0 || gValueY > GetSizeY()) {
		if (CallDialogBoxParam(gApp->Handle(), IDD_ARCIN, ArcInDlg, (LPARAM)gArcSize) == 0) {
			ArcSetSnt(gArcSize, pos(gValueX, gValueY));
		}
	} else {
		PushStack(1, IDST_ARC1);
		SetButtonOnState();
		gApp->EnableButton(IDB_LAYER, FALSE);
		gApp->EnableButton(IDB_SELECT, FALSE);
		EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_GRAYED);
		EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_AP, MF_GRAYED);
		DrawMenuBar(GetAppWnd());
		gApp->UpdateButton();
		gClicBase.Set(gValueX, gValueY);
		DrawArcSenter();
	}
}

static void FnArcKey()
{
	switch (gEditKey) {
	case EDKEY_LY:
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 1)) {
			GetScrHandle()->Invalidate();
		}
		SetLyApMenu();
		break;
	case EDKEY_SEL:
		CallDialogBox(gApp->Handle(), IDD_SELEARC, ArcSelDlg);
		SetLyApMenu();
		break;
	case EDKEY_VAL:
		EdKeyValSub(1);
		break;
	}
}

static void FnArcEsc()
{
	FnSetEsc();
	ResetStack(MOD_ARC, IDST_ARC);
	gApp->PushButton(IDB_ARC, TRUE);
	gApp->UpdateButton();
	gEditMode = EDMOD_ARC;
	SetLyApMenu();
}

static void FnArcLu1()
{
	pos p;
	GetMouseSnapPos(p);
	if (p.IsEq(gClicBase)) {
		MessageBeep(MB_OK);
	} else {
		double dx, dy, ang = atan2(dy = p.y - gClicBase.y, dx = p.x - gClicBase.x);
		double r = sqrt(dx * dx + dy * dy);
		if (GetCirAp(gEditLineCir) + ERR_MINI < r) {
			gDrawArc.Add(arc::New(gEditLayer, gEditLineCir, gClicBase, r, ang, ang));
			PushStack(2, IDST_ARC2);
		} else {
			MessageBeep(MB_OK);
		}
	}
}

static void FnArcValu1()
{
	pos p(gValueX, gValueY);
	if (gValueX < 0.0 || gValueX > GetSizeX() || gValueY < 0.0 || gValueY > GetSizeY() || p.IsEq(gClicBase)) {
		MessageBeep(MB_OK);
	} else {
		double dx, dy,
			   ang = atan2(dy = p.y - gClicBase.y, dx = p.x - gClicBase.x);
		double r = sqrt(dx * dx + dy * dy);
		if (GetCirAp(gEditLineCir) + ERR_MINI < r) {
			gDrawArc.Add(arc::New(gEditLayer, gEditLineCir, gClicBase, r, ang, ang));
			ClrFEditArcSt();
			PushStack(2, IDST_ARC2);
		} else {
			MessageBeep(MB_OK);
		}
	}
}

static void FnArcKey1()
{
	if (gEditKey == EDKEY_VAL) {
		EdKeyValSub(0);
	}
}

static void FnArcUnd1()
{
	PopStack();
}

static void FnArcEsc1()
{
	PopStack();
}

static int __fastcall ChengeArcAngle(const pos &p, arc *ar)
{
	if (ar->GetPos() == p) {
		return 1;
	}
	double ang = atan2(p.y - ar->GetPos().y, p.x - ar->GetPos().x);
	while (ang < 0.0) {
		ang += M_PI * 2.0;
	}
	if (IsFEditArcSt()) {
		if (ar->GetSt() > M_PI * 3.0 / 2.0 && ang < M_PI / 2.0) {
			ang += M_PI * 2.0;
		} else if (ar->GetSt() < M_PI / 2.0 && ang > M_PI * 3.0 / 2.0) {
			ang -= M_PI * 2.0;
		}
		if (ang > ar->GetEnd()) {
			ClrFEditArcSt();
		}
		ar->ChengeSt(ang);
	} else {
		if (ang < ar->GetSt()) {
			ang += M_PI * 2.0;
		}
		if (ar->GetEnd() - ar->GetSt() < M_PI / 2.0) {
			if (ang - ar->GetSt() > M_PI) {
				ang -= M_PI * 2.0;
			}
			if (ang < ar->GetSt()) {
				SetFEditArcSt();
			}
		} else if (ang - ar->GetSt() >= M_PI * 2.0) {
			ang -= M_PI * 2.0;
		}
		ar->ChengeEnd(ang);
	}
	return 0;
}

static void FnArcMv2()
{
	pos p;
	GetMouseHitPos(p);
	ChengeArcAngle(p, (arc *)gDrawArc.Ref());
}

static void __fastcall FnArcEnd(const pos &p)
{
	ChengeArcAngle(p, (arc *)gDrawArc.Ref());
	arc *pa = (arc *)gDrawArc.Ref();
	if (IsOutSize(pa)) {
		MessageBeep(MB_OK);
	} else {
		gPcb->Add(pa->Cut());
		AddOwnUndone(undone::New(pa));
		SetFFileChenge();
		HDC hDC = GetScrHandle()->GetScrDC();
		SelectObject(hDC, gPcb->GetPen(gEditLayer));
		SetROP2(hDC, R2_MERGEPEN);
		Draw(hDC, pa);
		PopStack(2);
	}
}

static void FnArcLu2()
{
	pos p;
	GetMouseSnapPos(p);
	if (p.IsEq(gClicBase)) {
		MessageBeep(MB_OK);
	} else {
		FnArcEnd(p);
	}
}

static void FnArcValu2()
{
	pos p(gValueX, gValueY);
	if (p.IsEq(gClicBase)) {
		MessageBeep(MB_OK);
	} else {
		FnArcEnd(p);
	}
}

static void FnArcUnd2()
{
	((element *)gDrawArc.Ref())->Del();
	PopStack();
}

static void FnArcEsc2()
{
	((element *)gDrawArc.Ref())->Del();
	PopStack(2);
}

static void __fastcall AddPcbPad(const pos &p)
{
	int ly = -1;
	group *grp = group::New(gPadIndex);
	HDC hDC = GetScrHandle()->GetScrDC();
	SetROP2(hDC, R2_MERGEPEN);
	for (ly_ap_t *pl = (ly_ap_t *)gDrawPad.Ref(); pl != &gDrawPad; pl = pl->Ref()) {
		element *pe;
		grp->AddGroup(pe = land::New(pl->GetLy(), pl->GetAp(), 0, p));
		if (gPcb->IsFDisp(pe)) {
			if (ly != pl->GetLy()) {
				SelectObject(hDC, gPcb->GetPen(ly = pl->GetLy()));
			}
			Draw(hDC, pe);
		}
	}
	if (IsOutSize((element *)grp)) {
		grp->Del();
		MessageBeep(MB_OK);
	} else {
		gPcb->Add(grp);
		AddOwnUndone(undone::New(grp));
		SetFFileChenge();
	}
}
static void FnPadLu()
{
	pos p;
	GetMouseSnapPos(p);
	AddPcbPad(p);
}

static void FnPadValu()
{
	AddPcbPad(pos(gValueX, gValueY));
}

static void FnPadKey()
{
	switch (gEditKey) {
	case EDKEY_LY:
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 0)) {
			GetScrHandle()->Invalidate();
		}
		SetLyApMenu();
		break;
	case EDKEY_SEL:
		if (! gPad.IsNull()) {
			int rt;
			if ((rt = CallDialogBoxParam(gApp->Handle(), IDD_SELEPAD, PadSeleDlg, (LPARAM)gEditPad)) >= 0) {
				ClearDraws();
				gEditPad = rt;
				pad *pl = (pad *)gPad.Ref();
				for (int i = 0; i < rt; ++i) {
					pl = pl->Ref();
				}
				const node *nb = &pl->GetLyAp();
				for (ly_ap_t *pp = (ly_ap_t *)nb->Ref(); pp != nb; pp = pp->Ref()) {
					gDrawPad.Add(ly_ap_t::New(pp->GetLy(), pp->GetAp()));
				}
			}
		}
		SetLyApMenu();
		break;
	case EDKEY_VAL:
		EdKeyValSub(1);
		break;
	}
}

static void FnPadEsc()
{
	pos p;
	FnSetEsc(p);
	ResetStack(MOD_PAD, IDST_PAD);
	gApp->PushButton(IDB_PAD, TRUE);
	gApp->UpdateButton();
	if (! gPad.IsNull()) {
		pad *pl = (pad *)gPad.Ref();
		int i = 0;
		while (i < gEditPad) {
			if (pl == &gPad) {
				return;
			}
			pl = pl->Ref();
			++i;
		}
		const node *nb = &pl->GetLyAp();
		for (ly_ap_t *pp = (ly_ap_t *)nb->Ref(); pp != nb; pp = pp->Ref()) {
			gDrawPad.Add(ly_ap_t::New(pp->GetLy(), pp->GetAp()));
		}
	}
	gEditMode = EDMOD_PAD;
	SetLyApMenu();
}

static group *__fastcall CreateText(pos p, textin_t &text, int ly, int ap, int *err = NULL)
{
	FILE *fp;
	const unsigned char *str;
	if (*(str = (unsigned char *)text.buff) == 0 || (fp = GetFontFile()) == NULL) {
		if (err) {
			*err = 1;
		}
		return NULL;
	}
	if (err) {
		*err = 0;
	}
	if (text.wi < 1e-2) {
		text.wi = 1e-2;
	}
	if (text.hi < 32e-3) {
		text.hi = 32e-3;
	}
	group *grp = group::New(gTextIndex);
	double ang = fmod(text.ang * M_PI / 180.0, M_PI * 2.0);
	double wi = text.f ? -text.wi : text.wi;
	grp->SetExt(group::ext_text::New(p, ang, text.hi, wi));
	double co = cos(ang);
	double si = sin(ang);
	double dx = wi * co;
	double dy = wi * si;
	if (wi < 0.0) {
		si = -si;
	}
	double adFont = text.hi / FONT_SIZE;
	BYTE buff[513];
	ConvFontCh();
	ConvFontCh(*str++);
	do {
		unsigned ch;
		if ((ch = ConvFontCh(*str)) != 0) {
			if (FontSeek(ch, fp) == 0 && fread(buff, 1, 513, fp) > 2) {
				int xx, yy;
				double ox = 0.0, oy = 0.0;
				int dwn = ((font_elem *)buff)->FindFirst(xx, yy);
				while (dwn >= 0) {
					double x = xx * adFont, y = yy * adFont;
					if ((fabs(ox - x) >= ERR_MINI || fabs(oy - y) >= ERR_MINI) && dwn) {
						double dx1, dx2;
						if (text.f) {
							dx1 = oy * si - ox * co;
							dx2 = y * si - x * co;
						} else {
							dx1 = ox * co - oy * si;
							dx2 = x * co - y * si;
						}
						line *pl;
						if (IsOutSize(pl = line::New(ly, ap, 0,
													 pos(p.x + dx1, p.y + oy * co + ox * si), pos(p.x + dx2, p.y + y * co + x * si)))) {
							if (err) {
								*err = 1;
							}
							pl->Del();
						} else {
							grp->AddGroup(pl);
						}
					}
					ox = x;
					oy = y;
					dwn = ((font_elem *)buff)->FindNext(xx, yy);
				}
			}
			p.Move(dx, dy);
		}
	} while (*str++);
	fclose(fp);
	if (grp->IsNullGroup()) {
		grp->Del();
		if (err) {
			*err = 1;
		}
		return NULL;
	} else {
		grp->SetAttrib(text.buff);
	}
	return grp;
}

static void __fastcall PcbAddText(pos po)
{
	if (CallDialogBoxParam(gApp->Handle(), IDD_TEXTIN, TextInDlg, (LPARAM)&gTEXT_IN) == 0 && *gTEXT_IN.buff) {
		int err;
		group *grp;
		if ((grp = CreateText(po, gTEXT_IN, gEditLayer, gEditText, &err)) != NULL) {
			HDC hDC = GetScrHandle()->GetScrDC();
			SelectObject(hDC, gPcb->GetPen(gEditLayer));
			SetROP2(hDC, R2_MERGEPEN);
			const node &bas = grp->GetCont();
			for (element *pe = (element *)bas.Ref(); pe != &bas; pe = pe->Ref()) {
				Draw(hDC, pe);
			}
			gPcb->Add(grp);
			AddOwnUndone(undone::New(grp));
			SetFFileChenge();
		}
		if (err) {
			MessageBeep(MB_OK);
		}
	}
	SetLyApMenu();
}

static void FnTextLu()
{
	pos p;
	GetMouseSnapPos(p);
	PcbAddText(p);
}

static void FnTextValu()
{
	PcbAddText(pos(gValueX, gValueY));
}

static void FnTextKey()
{
	apsele_t ap;
	switch (gEditKey) {
	case EDKEY_LY:
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 1)) {
			GetScrHandle()->Invalidate();
		}
		SetLyApMenu();
		break;
	case EDKEY_SEL:
		ap.sq = 0;
		ap.pap = &gEditText;
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELELAND, ApSeleDlg, (LPARAM)&ap)) {
			SetLyApMenu();
		}
		break;
	case EDKEY_VAL:
		EdKeyValSub(1);
		break;
	}
}

static void FnTextEsc()
{
	FnSetEsc();
	ResetStack(MOD_TEXT, IDST_TEXT);
	gApp->PushButton(IDB_TEXT, TRUE);
	gApp->UpdateButton();
	gEditMode = EDMOD_TEXT;
	SetLyApMenu();
}

static void __fastcall BeginEditMove(const pos &p)
{
	PushStack(1, IDST_EDIT2);
	SetButtonOnState();
	gApp->EnableButton(IDB_LAYER, FALSE);
	EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_GRAYED);
	DrawMenuBar(GetAppWnd());
	gApp->EnableButton(IDB_VALUEIN, TRUE);
	gApp->EnableButton(IDB_CROSSIN, TRUE);
	if (gEditEditMode != ED_LINE2 && gEditEditMode != ED_GROUP && !((container*)gEditSeles.Ref())->IsLine()) {
		gApp->EnableButton(IDB_SELECT, FALSE);
	}
	if (gEditEditMode == ED_GROUP) {
		gApp->EnableButton(IDB_ROTATE, TRUE);
		if (((container *)gEditSeles.Ref())->IsGroup() && ((group *)gEditSeles.Ref())->IsExtParts()) {
			gApp->EnableButton(IDB_MIRROR, TRUE);
		}
	} else if (((container *)gEditSeles.Ref())->IsRect()) {
		gApp->EnableButton(IDB_ROTATE, TRUE);
	}
	gApp->UpdateButton();
	DrawEditSeleCros(GetScrHandle()->GetScrDC(), R2_COPYPEN);
	gClicBase = gEditBase = p;
}

void edit_pos::Init()
{
	((int *)gTempMem)[1] = *(int *)gTempMem = 0;
	pedit_pos ped = GetPosR2Top();
	top = &ped[1];
	ped[1].next = ped->next = (pedit_pos)INVALID_HANDLE_VALUE;
	ped[1].pe = ped->pe = NULL;
	ped[1].p.Set(-1e200, -1e200);
	ped->p.Set(-1e200, -1e200);
}

int __fastcall edit_pos::Add(container *pc, const pos *pp)
{
	pedit_pos ped;
	if (IsArcR2()) {
		ped = GetPosR2Top();
	} else {
		ped = GetPosTop();
	}
	if (ped->next != INVALID_HANDLE_VALUE) {
		if (pc->IsLine() && pp) {
			for (;;) {
				if (ped->p.x > -1e100 && ped->p.IsEq(*pp)) {
					edit_elem *pel = ped->pe;
					for (;;) {
						if (pel->pc == pc) {
							return 0;
						}
						if (pel->next == NULL) {
							break;
						}
						pel = pel->next;
					}
					if (top > gTempMemMax) {
						return -1;
					}
					pel->next = (edit_elem *)top;
					pel = pel->next;
					top = &pel[1];
					pel->pc = pc;
					pel->next = NULL;
					return 0;
				}
				if (ped->next == NULL) {
					break;
				}
				ped = ped->next;
			}
		} else {
			for (;;) {
				for (edit_elem *pel = ped->pe; pel; pel = pel->next) {
					if (pel->pc == pc) {
						return 0;
					}
				}
				if (ped->next == NULL) {
					break;
				}
				ped = ped->next;
			}
		}
		if (top > gTempMemMax) {
			return -1;
		}
		ped->next = (pedit_pos)top;
		ped = ped->next;
		top = &ped[1];
		((pedit_pos)top)->next = NULL;
	} else if (ped == top) {
		top = &((pedit_pos)top)[1];
		((pedit_pos)top)->next = NULL;
	}
	ped->next = NULL;
	if (pp) {
		ped->p = *pp;
	} else {
		ped->p.Set(-1e200, -1e200);
	}
	if (top > gTempMemMax) {
		ped->pe = NULL;
		return -1;
	}
	ped->pe = (edit_elem *)top;
	top = &ped->pe[1];
	((edit_pos *)top)->next = NULL;
	ped->pe->next = NULL;
	ped->pe->pc = pc;
	return 0;
}

static land *__fastcall GetGroupOrigPos(const node *pn)
{
	while (!pn->IsNull() && ((container *)pn->Ref())->IsGroup() && pn->Ref()->Ref() == pn) {
		pn = &((group *)pn->Ref())->GetCont();
	}
	land *pl = NULL;
	for (container *pc = (container *)pn->Ref(); pc != pn; pc = pc->Ref()) {
		if (pc->IsLand() && !((land *)pc)->IsSq() && gPcb->IsFDatum((land *)pc)) {
			if (pl) {
				return NULL;
			}
			pl = (land *)pc;
		}
	}
	return pl;
}

pos __fastcall GetGroupOrigPos(const node &bas)
{
	if (!bas.IsNull() && bas.Ref()->Ref() == &bas
			&& ((container *)bas.Ref())->IsGroup() && ((group *)bas.Ref())->GetExt()) {
		return ((group *)bas.Ref())->GetExt()->GetPb();
	}
	land *pl = GetGroupOrigPos(&bas);
	if (pl) {
		return pl->GetPos();
	}
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup() && (pl = GetGroupOrigPos(&((group *)pc)->GetCont())) != NULL) {
			return pl->GetPos();
		}
	}
	pos lo, hi;
	GetMinMaxPoint(bas, lo, hi);
	double x, y;
	Pos2Scr(pos((lo.x + hi.x) / 2.0, (lo.y + hi.y) / 2.0), x, y);
	Scr2Snap(x, y, lo);
	return lo;
}

static const pos __fastcall GetPushPos(const node &bas)
{
	node *top = (node*)&bas;
	while (!top->IsNull() && top->Ref()->Ref() == top && ((container*)top->Ref())->IsGroup()) {
		const node &b = ((group*)top->Ref())->GetCont();
		if (!b.IsNull() && b.Ref()->Ref() == &b && ((container*)b.Ref())->IsGroup()) {
			top = (node*)&b;
		} else {
			break;
		}
	}
	if (!top->IsNull() && top->Ref()->Ref() == top && ((container*)top->Ref())->IsGroup()) {
		if (((group*)top->Ref())->GetExt()) {
			return ((group*)top->Ref())->GetExt()->GetPb();
		}
		top = (node*)&((group*)top->Ref())->GetCont();
	}
	pos p;
	int n = 0;
	for (container *pc = (container*)top->Ref(); pc != top; pc = pc->Ref()) {
		if (pc->IsLand() && gPcb->IsFDatum((element*)pc) && !((element*)pc)->IsSq()) {
			++n;
			p = ((land*)pc)->GetPos();
		}
	}
	if (n == 1) {
		return p;
	}
	pos hi;
	GetMinMaxPoint(bas, p, hi);
	double x, y;
	Pos2Scr(pos((p.x + hi.x) / 2.0, (p.y + hi.y) / 2.0), x, y);
	Scr2Snap(x, y, p);
	return p;
}

static inline int IsPaintLy(int ly, const node &nb)
{
	if (nb.Ref() != &nb && ((container *)nb.Ref())->IsGroup()
			&& ((group *)nb.Ref())->GetCont().Ref() != &((group *)nb.Ref())->GetCont()
			&& ((container *)((group *)nb.Ref())->GetCont().Ref())->IsElem()
			&& ((element *)((group *)nb.Ref())->GetCont().Ref())->GetLy() == ly) {
		return 1;
	}
	return 0;
}

static int __fastcall IsPaintGroup(const node &bas)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			if (((group *)pc)->IsNullGroup() || (((group *)pc)->GetName() != gPaintedMado
												 && ((group *)pc)->GetName() != gPaintedLoop && ((group *)pc)->GetName() != gPaintedLines)) {
				return 0;
			}
		} else {
			return 0;
		}
	}
	return 1;
}

#if 1
static int __fastcall IsInPaintedLines(const node &bas, const pos &p)
{
	double ap = GetCirAp(((line *)bas.Ref())->GetAp());
	double ap2 = ap * ap;
	for (line *pl = (line *)bas.Ref(); pl != &bas; pl = pl->Ref()) {
		if ((pl->GetHi().x - p.x) * (pl->GetLo().x - p.x) >= 0.0 || (pl->GetHi().y - p.y) * (pl->GetLo().y - p.y) >= 0.0) {
			continue;
		}
		double dx, dy = pl->GetPe().y - pl->GetPos().y;
		if (fabs(dy) < ERR_MINI) {
			if (fabs(p.y - pl->GetPos().y) < ap && (p.x - pl->GetPos().x) * (p.x - pl->GetPe().x) <= 0.0) {
				return 1;
			}
		} else if (fabs(dx = pl->GetPe().x - pl->GetPos().x) < ERR_MINI) {
			if (fabs(p.x - pl->GetPos().x) < ap && (p.y - pl->GetPos().y) * (p.y - pl->GetPe().y) <= 0.0) {
				return 1;
			}
		} else {
			double a = dy / dx;
			double b = pl->GetPos().y - a * pl->GetPos().x;
			double x = (p.y + p.x / a - b) / (a + 1.0 / a);
			double y = a * x + b;
			dx = x - p.x;
			dy = y - p.y;
			if (dx * dx + dy * dy < ap2
					&& (pl->GetPos().x - x) * (pl->GetPe().x - x) <= 0.0
					&& (pl->GetPos().y - y) * (pl->GetPe().y - y) <= 0.0) {
				return 1;
			}
		}
	}
	return 0;
}

int __fastcall IsPointInPaint(const node &bas, const pos &p)
{
	for (group *pg = (group *)bas.Ref(); pg != &bas; pg = pg->Ref()) {
		if (pg->IsGroup()) {
			const node &nb = pg->GetCont();
			if (pg->GetName() == gPaintedLines) {
				if (((container *)pg->GetCont().Ref())->IsLine() && IsInPaintedLines(nb, p)) {
					return 1;
				}
			} else {
				for (element *pe = (element *)nb.Ref(); pe != &nb; pe = pe->Ref()) {
					if (pe->IsElem() && pe->IsEqPos(p)) {
						return 1;
					}
				}
			}
		}
	}
	return 0;
}
#else
static int __fastcall GetPointInArc(arc *pa, double y, double x[])
{
	double dy = y - pa->GetPos().y;
	int n = 0;
	double ang = atan2(dy, -pa->GetR());
	if (ang > pa->GetEnd()) {
		ang -= M_PI * 2.0;
	}
	while (ang < pa->GetSt()) {
		ang += M_PI * 2.0;
	}
	if (ang <= pa->GetEnd()) {
		x[n++] = pa->GetPos().x + pa->GetR() * cos(ang);
	}
	ang = atan2(dy, pa->GetR());
	if (ang > pa->GetEnd()) {
		ang -= M_PI * 2.0;
	}
	while (ang < pa->GetSt()) {
		ang += M_PI * 2.0;
	}
	if (ang <= pa->GetEnd()) {
		x[n++] = pa->GetPos().x + pa->GetR() * cos(ang);
	}
	return n;
}

#define TMPCKSZ		2000

static double *__fastcall GetPointInPaint(const node &bas, const pos &p, double xbuf[], double *lim)
{
	double *xp = xbuf, *max = lim, mini = 1e200;
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup() && (((group *)pc)->GetName() == gPaintedLoop || ((group *)pc)->GetName() == gPaintedMado)) {
			const node &b = ((group *)pc)->GetCont();
			for (element *pe = (element *)b.Ref(); pe != &b; pe = pe->Ref()) {
				double ap = GetCirAp(pe->GetAp());
				if ((pe->GetHi().y - ap - p.y) * (pe->GetLo().y + ap - p.y) <= 0.0) {
					if (pe->IsArc()) {
						double x[2];
						int n = GetPointInArc((arc *)pe, p.y, x);
						for (int i = 0; i < n; ++i) {
							double xtm = x[i];
							if (xtm < mini) {
								mini = xtm;
							}
							double *ptr = xbuf;
							do {
								if (ptr >= xp) {
									*xp++ = xtm;
									if (xp >= max) {
										goto exit_loop;
									}
									break;
								}
							} while (fabs(*ptr++ - xtm) > ERR_MINI);
						}
					} else if (pe->IsLine()) {
						double x, dx = ((line *)pe)->GetPe().x - pe->GetPos().x;
						if (fabs(dx) < ERR_MINI) {
							if ((x = pe->GetPos().x) < mini) {
								mini = x;
							}
							double *ptr = xbuf;
							do {
								if (ptr >= xp) {
									*xp++ = x;
									if (xp >= max) {
										goto exit_loop;
									}
									break;
								}
							} while (fabs(*ptr++ - x) > ERR_MINI);
						} else {
							double dy = ((line *)pe)->GetPe().y - pe->GetPos().y;
							if (fabs(dy) < ERR_MINI) {
								x = pe->GetLo().x + ap;
								double *ptr = max - 1;
								do {
									if (ptr < max) {
										*(max = ptr) = x;
										if (xp >= max) {
											goto exit_loop;
										}
										break;
									}
								} while (fabs(*ptr-- - x) > ERR_MINI);
								x = pe->GetHi().x - ap;
							} else {
								double a = dy / dx;
								double b = pe->GetPos().y - a * pe->GetPos().x;
								x = (p.y - b) / a;
							}
							double *ptr = xbuf;
							do {
								if (ptr >= xp) {
									*xp++ = x;
									if (xp >= max) {
										goto exit_loop;
									}
									break;
								}
							} while (fabs(*ptr++ - x) > ERR_MINI);
						}
					}
				}
			}
		}
	}
exit_loop:
	for (double *ptr = max; ptr < lim; ++ptr) {
		if (fabs(*ptr - mini) < ERR_MINI) {
			*ptr = 2e200;
		}
	}
	for (double *ptr = max; ptr < lim; ++ptr) {
		double x = *ptr;
		for (double *pck = xbuf; pck < xp; ++pck) {
			if (fabs(*pck - x) < ERR_MINI) {
				*pck = 1e100;
			}
		}
	}
	return xp;
}

int __fastcall IsPointInPaint(const node &bas, const pos &p)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup() && (((group *)pc)->GetName() == gPaintedLoop || ((group *)pc)->GetName() == gPaintedMado)) {
			const node &b = ((group *)pc)->GetCont();
			for (element *pe = (element *)b.Ref(); pe != &b; pe = pe->Ref()) {
				if (pe->IsEqPos(p)) {
					return 1;
				}
			}
		}
	}
	double x[TMPCKSZ], *max;
	max = GetPointInPaint(bas, p, x, x + TMPCKSZ);
	int n = 0;
	for (double *ptr = x; ptr < max; ++ptr) {
		if (p.x <= *ptr) {
			++n;
		}
	}
	return n & 1;
}

#endif

static inline int IsInPaint(const pos &p, group *paint)
{
	return IsPointInPaint(paint->GetCont(), p);
}

static int __fastcall IsHitPosGrp(int ly, const pos &p, group *pg)
{
	if (!gPcb->IsFEdit(ly) || pg->GetName() == gTextIndex || pg->GetName() == gDrcIndex) {
		return 0;
	}
	if (pg->GetName() == gPaintedIndex) {
		if (IsPaintGroup(pg->GetCont()) && IsPaintLy(ly, pg->GetCont()) && IsPointInPaint(pg->GetCont(), p)) {
			return 1;
		}
	} else {
		const node &bas = pg->GetCont();
		for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
			if (pc->IsGroup()) {
				if (IsHitPosGrp(ly, p, (group *)pc)) {
					return 1;
				}
			} else {
				if (((element *)pc)->GetLy() == ly && ((element *)pc)->IsEqPos(p)) {
					return 1;
				}
			}
		}
	}
	return 0;
}

static int __fastcall IsInPaint(int ly, element *pe, const node &paint, const pos **p = NULL)
{
	if (!gPcb->IsFEdit(ly)) {
		return 0;
	}
	if (pe->IsElem() && pe->GetLy() == ly) {
		switch (pe->GetType()) {
		case container::T_LAND:
			if (IsPointInPaint(paint, ((land *)pe)->GetPos())) {
				if (p) {
					*p = &((land *)pe)->GetPos();
				}
				return 1;
			}
			break;
		case container::T_LINE:
			if (IsPointInPaint(paint, ((line *)pe)->GetPos())) {
				if (p) {
					*p = &((line *)pe)->GetPos();
				}
				return 1;
			} else if (IsPointInPaint(paint, ((line *)pe)->GetPe())) {
				if (p) {
					*p = &((line *)pe)->GetPe();
				}
				return 1;
			}
			break;
		case container::T_ARC:
			if (IsPointInPaint(paint, ((arc *)pe)->GetStPos())) {
				if (p) {
					*p = &((arc *)pe)->GetStPos();
				}
				return 1;
			} else if (IsPointInPaint(paint, ((arc *)pe)->GetEndPos())) {
				if (p) {
					*p = &((arc *)pe)->GetEndPos();
				}
				return 1;
			}
			break;
		case container::T_RECT:
			if (IsPointInPaint(paint, ((rect *)pe)->GetPos())) {
				if (p) {
					*p = &((rect *)pe)->GetPos();
				}
				return 1;
			} else if (IsPointInPaint(paint, ((rect *)pe)->GetP1())) {
				if (p) {
					*p = &((rect *)pe)->GetP1();
				}
				return 1;
			} else if (IsPointInPaint(paint, ((rect *)pe)->GetP2())) {
				if (p) {
					*p = &((rect *)pe)->GetP2();
				}
				return 1;
			}
			break;
		}
	}
	return 0;
}

static int __fastcall IsInPaint(int ly, const node &bas, const node &paint)
{
	if (!gPcb->IsFEdit(ly)) {
		return 0;
	}
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			if (((group *)pc)->GetName() == gPaintedIndex) {
				if (IsPaintGroup(((group *)pc)->GetCont()) && IsPaintLy(ly, ((group *)pc)->GetCont())) {
					const node &nb = ((group *)pc)->GetCont();
					for (container *pcon = (container *)nb.Ref(); pcon != &nb; pcon = pcon->Ref()) {
						if (pcon->IsGroup()
								&& (((group *)pcon)->GetName() == gPaintedLoop
									|| ((group *)pcon)->GetName() == gPaintedMado)) {
							const node &no = ((group *)pcon)->GetCont();
							for (container *pac = (container *)no.Ref(); pac != &no; pac = pac->Ref()) {
								if (IsInPaint(ly, (element *)pac, paint)) {
									return 1;
								}
							}
						}
					}
				}
			} else if (((group *)pc)->GetName() != gTextIndex && ((group *)pc)->GetName() != gDrcIndex
					   && IsInPaint(ly, ((group *)pc)->GetCont(), paint)) {
				return 1;
			}
		} else if (IsInPaint(ly, (element *)pc, paint)) {
			return 1;
		}
	}
	return 0;
}

int __fastcall GetAllHit(int ly, const pos *p, undone *und);

static int __fastcall GetAllHit(group *grp, undone *und)
{
	if (grp->GetName() == gPaintedIndex) {
		const node &bas = grp->GetCont();
		if (!IsPaintGroup(bas) || !((container *)bas.Ref())->IsGroup() || !((container *)((group *)bas.Ref())->GetCont().Ref())->IsElem()) {
			return 0;
		}
		int ly = ((element *)((group *)bas.Ref())->GetCont().Ref())->GetLy();
		node tmp(0);
		const node &nb = gPcb->GetElems();
		for (container *pc = (container *)nb.Ref(); pc != &nb;) {
			container *next = pc->Ref();
			const pos *pp;
			if (pc->IsGroup()) {
				if (((group *)pc)->GetName() != gTextIndex && ((group *)pc)->GetName() != gDrcIndex
						&& IsInPaint(ly, ((group *)pc)->GetCont(), bas)) {
					container *hit = pc->Copy();
					und->AddData(hit);
					und->AddData(pc->Cut(), pc->Ref());
					tmp.Add(hit);
				}
			} else if (IsInPaint(ly, (element *)pc, bas, &pp)) {
				container *hit = pc->Copy();
				und->AddData(hit);
				und->AddData(pc->Cut(), pc->Ref());
				gEditSeles.Add(hit);
				if (edit_pos::Add(hit, pp)) {
					return -1;
				}
			}
			pc = next;
		}
		while (!tmp.IsNull()) {
			container *hit = (container *)tmp.Ref()->Cut();
			if (edit_pos::Add(hit)) {
				return -1;
			}
			gEditSeles.Add(hit);
			GetAllHit((group *)hit, und);
		}
	} else if (grp->GetName() != gTextIndex && grp->GetName() != gDrcIndex) {
		const node &bas = grp->GetCont();
		for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
			switch (pc->GetType()) {
			case container::T_LAND:
				if (GetAllHit(((land *)pc)->GetLy(), &((land *)pc)->GetPos(), und)) {
					return -1;
				}
				break;
			case container::T_LINE:
				if (GetAllHit(((line *)pc)->GetLy(), &((line *)pc)->GetPos(), und)) {
					return -1;
				}
				if (GetAllHit(((line *)pc)->GetLy(), &((line *)pc)->GetPe(), und)) {
					return -1;
				}
				break;
			case container::T_ARC:
				if (GetAllHit(((arc *)pc)->GetLy(), &((arc *)pc)->GetStPos(), und)) {
					return -1;
				}
				if (GetAllHit(((arc *)pc)->GetLy(), &((arc *)pc)->GetEndPos(), und)) {
					return -1;
				}
				break;
			case container::T_RECT:
				if (GetAllHit(((rect *)pc)->GetLy(), &((rect *)pc)->GetPos(), und)) {
					return -1;
				}
				if (GetAllHit(((rect *)pc)->GetLy(), &((rect *)pc)->GetP1(), und)) {
					return -1;
				}
				if (GetAllHit(((rect *)pc)->GetLy(), &((rect *)pc)->GetP2(), und)) {
					return -1;
				}
				break;
			case container::T_GROUP:
				if (GetAllHit((group *)pc, und)) {
					return -1;
				}
				break;
			}
		}
	}
	return 0;
}

static void __fastcall GetAllHitTh(const pos *p, undone *und)
{
	const node &bas = gPcb->GetElems();
	for (container *pc = (container *)bas.Ref(); pc != &bas;) {
		container *pt = pc;
		pc = pc->Ref();
		if (pt->IsElem() && gPcb->IsFEdit((element *)pt)
			&& gPcb->IsFPtn((element*)pt) && ((element *)pt)->IsEqPos(*p)) {
			container *hit = pt->Copy();
			if (edit_pos::Add(hit, p)) {
				hit->Del();
				return;
			}
			und->AddData(hit);
			und->AddData(pt->Cut(), pc);
			gEditSeles.Add(hit);
		}
	}
}

static int __fastcall IsHitPad(const pos *p, const node &bas)
{
	for (element *pe = (element *)bas.Ref(); pe != &bas; pe = pe->Ref()) {
		if (pe->IsGroup() && IsHitPad(p, ((group *)pe)->GetCont())) {
			return 1;
		} else if (pe->IsLand() && gPcb->IsFEdit(pe) && gPcb->IsFHole(pe) && !gPcb->IsFNoTh(pe) && pe->IsEqPos(*p)) {
			return 1;
		}
	}
	return 0;
}

static container *__fastcall GetAllHitElem(const pos *p, undone *und)
{
	const node &bas = gPcb->GetElems();
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsLand() && gPcb->IsFEdit((element *)pc)
				&& gPcb->IsFHole((element *)pc) && !gPcb->IsFNoTh((element *)pc) && ((element *)pc)->IsEqPos(*p)) {
			container *hit = pc->Copy();
			if (edit_pos::Add(hit, p)) {
				hit->Del();
				return NULL;
			}
			und->AddData(hit);
			und->AddData(pc->Cut(), pc->Ref());
			gEditSeles.Add(hit);
			GetAllHitTh(p, und);
			break;
		} else if (pc->IsGroup() && IsHitPad(p, ((group *)pc)->GetCont())) {
			return pc;
		}
	}
	return NULL;
}

static int __fastcall GetAllHit(int ly, const pos *p, undone *und)
{
	if (!gPcb->IsFEdit(ly)) {
		return 0;
	}
	node tmp(0);
	if (gPcb->IsFHole(ly) && !gPcb->IsFNoTh(ly)) {
		GetAllHitTh(p, und);
	} else {
		container *pc = GetAllHitElem(p, und);
		if (pc) {
			container *hit = pc->Copy();
			und->AddData(hit);
			und->AddData(pc->Cut(), pc->Ref());
			tmp.Add(hit);
		}
	}
	const node &bas = gPcb->GetElems();
	for (container *pc = (container *)bas.Ref(); pc != &bas;) {
		container *next = pc->Ref();
		if (pc->IsGroup()) {
			if (IsHitPosGrp(ly, *p, (group *)pc)) {
				container *hit = pc->Copy();
				und->AddData(hit);
				und->AddData(pc->Cut(), pc->Ref());
				tmp.Add(hit);
			}
		} else if (((element *)pc)->GetLy() == ly && ((element *)pc)->IsEqPos(*p)) {
			container *hit = pc->Copy();
			if (edit_pos::Add(hit, p)) {
				hit->Del();
				return -1;
			}
			und->AddData(hit);
			und->AddData(pc->Cut(), pc->Ref());
			gEditSeles.Add(hit);
		}
		pc = next;
	}
	while (tmp.Ref() != &tmp) {
		group *grp = (group *)tmp.Ref()->Cut();
		gEditSeles.Add(grp);
		if (edit_pos::Add(grp) || GetAllHit(grp, und)) {
			return -1;
		}
	}
	return 0;
}

static void __fastcall SetLineExpand(int ly, line *li, const node &bas)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			if (((group *)pc)->GetName() == gPaintedIndex && IsPaintLy(ly, ((group *)pc)->GetCont())) {
				if (IsPointInPaint(((group *)pc)->GetCont(), li->GetPos())) {
					edit_pos::Add(li, &li->GetPos());
				}
				if (IsPointInPaint(((group *)pc)->GetCont(), li->GetPe())) {
					edit_pos::Add(li, &li->GetPe());
				}
			} else if (((group *)pc)->GetName() != gTextIndex && ((group *)pc)->GetName() != gDrcIndex) {
				SetLineExpand(ly, li, ((group *)pc)->GetCont());
			}
		} else if (((element *)pc)->GetLy() == ly && (pc->GetOwner() || !pc->IsLine())) {
			if (((element *)pc)->IsEqPos(li->GetPos())) {
				if (edit_pos::Add(li, &li->GetPos())) {
					return;
				}
			} else if (((element *)pc)->IsEqPos(li->GetPe())) {
				if (edit_pos::Add(li, &li->GetPe())) {
					return;
				}
			}
		}
	}
}

static void SetLineExpand()
{
	pedit_pos po = edit_pos::GetPosTop();
	if (po->next == INVALID_HANDLE_VALUE) {
		po->next = NULL;
	}
	for (; po; po = po->next) {
		for (edit_pos::edit_elem *pel = po->pe; pel; pel = pel->next) {
			if (pel->pc->IsLine()) {
				line *li = (line *)pel->pc;
				SetLineExpand(li->GetLy(), li, gEditSeles);
			}
		}
	}
	if ((po = edit_pos::GetPosR2Top())->next == INVALID_HANDLE_VALUE) {
		po->next = NULL;
		return;
	}
	if (edit_pos::IsArcR()) {
		for (; po; po = po->next) {
			for (edit_pos::edit_elem *pel = po->pe; pel; pel = pel->next) {
				if (pel->pc->IsLine()) {
					line *li = (line *)pel->pc;
					SetLineExpand(li->GetLy(), li, gEditSeles);
				}
			}
		}
	}
}

static int __fastcall CheckDualArc(int ly, const pos &p, element **ppe = NULL)
{
	int n = 0;
	const node &bas = gPcb->GetElems();
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsArc() && ((arc *)pc)->GetLy() == ly && ((arc *)pc)->IsEqPos(p)) {
			if (ppe) {
				*ppe = (element *)pc;
			}
			++n;
		}
	}
	return n;
}

static int __fastcall CheckDualArc(const pos &p, element **ppe, int f)
{
	int n;
	if ((*ppe)->IsArc()) {
		if (f) {
			n = CheckDualArc((*ppe)->GetLy(), p);
		} else {
			n = CheckDualArc((*ppe)->GetLy(), ((arc *)*ppe)->GetStPos());
			n += CheckDualArc((*ppe)->GetLy(), ((arc *)*ppe)->GetEndPos());
			--n;
		}
	} else if ((*ppe)->GetOwner() && ((group *)(*ppe)->GetTopOwner())->GetName() == gPadIndex) {
		const node &bas = ((group *)(*ppe)->GetTopOwner())->GetCont();
		n = 0;
		for (element *pe = (element *)bas.Ref(); pe != &bas; pe = pe->Ref()) {
			if (pe->IsElem() && gPcb->IsFEdit(pe)) {
				n += CheckDualArc(pe->GetLy(), p, ppe);
			}
		}
	} else {
		n = CheckDualArc((*ppe)->GetLy(), p, ppe);
	}
	return n > 1;
}

static void SetEditLine2(pos &p, line *li, undone *und)
{
	gEditEditMode = ED_LINE2;
	GetMouseSnapPos(p);
	HDC hDC = GetScrHandle()->GetScrDC();
	SetROP2(hDC, R2_BLACK);
	Draw(hDC, li);
	container *pe;
	gEditSeles.Add(pe = line::New(li->GetLy(), li->GetAp(), li->IsSq(), li->GetPos(), p));
	edit_pos::Add(pe, &p);
	pe->CopyOwner(li);
	und->AddData(pe);
	gEditSeles.Add(pe = line::New(li->GetLy(), li->GetAp(), li->IsSq(), p, li->GetPe()));
	edit_pos::Add(pe, &p);
	pe->CopyOwner(li);
	und->AddData(pe);
}

static element *__fastcall CheckWindowGroup(element *pe)
{
	const group *pg = pe->GetOwner();
	if (pg) {
		if (pg->GetName() == gPaintedLoop || pg->GetName() == gPaintedLines || pg->GetName() == gPaintedMado) {
			pg = pg->GetOwner();
		}
		if (pg && pg->GetName() == gTextIndex || pg->GetName() == gPadIndex
				|| pg->GetName() == gPaintedIndex || pg->GetName() == gDrcIndex) {
			return (element *)pg;
		}
	}
	return pe;
}

static edit_edit_t CheckLine2()
{
	container *pc = (container *)gEditSeles.Ref();
	if ((pc = pc->Ref()) == &gEditSeles || pc->Ref() != &gEditSeles || !pc->IsLine()) {
		return ED_EXPAND;
	}
	return ED_LINE2;
}

static void FnEditLu()
{
	pos p;
	GetMouseHitPos(p);
	element *pe;
	undone *und = NULL;
	edit_pos::Init();
	HCURSOR cursor = SetCursor(LoadCursor(NULL, MAKEINTRESOURCE(IDC_WAIT)));
	int fNoCheck = 0;
	if ((pe = gPcb->IsHit(p)) == NULL) {
		SetCursor(cursor);
		MessageBeep(MB_OK);
		return;
	} else {
		int f = p.Distance(pe->GetNearPos(p)) < GetHitInPoint(pe);
		if ((GetScrHandle()->GetMouseKey() & (MK_SHIFT | MK_CONTROL))
				|| IsFEditOwn() || (!f && pe->IsLine() && !pe->GetOwner())) {
			if ((fNoCheck = pe->GetOwner() && (GetScrHandle()->GetMouseKey() & MK_CONTROL)) != 0) {
				pe = CheckWindowGroup(pe);
			} else {
				pe = (element *)pe->GetTopOwner();
			}
			und = undone::New(pe->Cut(), pe->Ref());
			switch (pe->GetType()) {
			case container::T_LAND:
			case container::T_RECT:
				gEditEditMode = ED_MOVE;
				gEditSeles.Add(pe->Copy());
				edit_pos::Add((container *)gEditSeles.Ref(), pe->IsGroup() ? NULL : &pe->GetNearPos(p));
				break;
			case container::T_GROUP:
				gEditEditMode = ED_GROUP;
				gEditSeles.Add(pe->Copy());
				edit_pos::Add((container *)gEditSeles.Ref(), pe->IsGroup() ? NULL : &pe->GetNearPos(p));
				break;
			case container::T_LINE:
				if (f) {
					gEditEditMode = ED_EXPAND;
					gEditSeles.Add(pe->Copy());
					edit_pos::Add((container *)gEditSeles.Ref(), &pe->GetNearPos(p));
				} else {
					SetEditLine2(p, (line *)pe, und);
					pe = (element *)gEditSeles.Ref();
				}
				break;
			case container::T_ARC:
				gEditSeles.Add(pe->Copy());
				if (f) {
					gEditEditMode = ED_ARC_ANG;
					if (((arc *)pe)->GetNearPos(p).IsEq(((arc *)pe)->GetStPos())) {
						SetFEditArcSt();
					} else {
						ClrFEditArcSt();
					}
				} else {
					gEditEditMode = ED_ARC_R;
				}
				break;
			default:
				goto edlu_err;
			}
		} else {
			if (CheckDualArc(pe->GetNearPos(p), &pe, f)) {
				goto edlu_err;
			}
			pe = (element *)pe->GetTopOwner();
			und = undone::New(pe->Cut(), pe->Ref());
			switch (pe->GetType()) {
			case container::T_LAND:
				gEditEditMode = ED_MOVE;
				gEditSeles.Add(pe->Copy());
				edit_pos::Add((container *)gEditSeles.Ref(), &pe->GetNearPos(p));
				GetAllHit(pe->GetLy(), &pe->GetNearPos(p), und);
				break;
			case container::T_LINE:
				if (f) {
					gEditSeles.Add(pe->Copy());
					edit_pos::Add((container *)gEditSeles.Ref(), &pe->GetNearPos(p));
					GetAllHit(pe->GetLy(), &pe->GetNearPos(p), und);
					gEditEditMode = CheckLine2();
				} else {
					SetEditLine2(p, (line *)pe, und);
					pe = (element *)gEditSeles.Ref();
				}
				break;
			case container::T_ARC:
				gEditSeles.Add(pe->Copy());
				if (f) {
					gEditEditMode = ED_ARC_ANG;
					GetAllHit(pe->GetLy(), &((arc *)pe)->GetNearPos(p), und);
					if (((arc *)pe)->GetNearPos(p).IsEq(((arc *)pe)->GetStPos())) {
						SetFEditArcSt();
					} else {
						ClrFEditArcSt();
					}
				} else {
					gEditEditMode = ED_ARC_R;
					edit_pos::SetArcR();
					GetAllHit(pe->GetLy(), &((arc *)pe)->GetStPos(), und);
					edit_pos::SetArcR2();
					GetAllHit(pe->GetLy(), &((arc *)pe)->GetEndPos(), und);
				}
				break;
			case container::T_RECT:
				gEditEditMode = ED_MOVE;
				gEditSeles.Add(pe->Copy());
				edit_pos::Add((container *)gEditSeles.Ref(), &pe->GetNearPos(p));
				if (pe->GetNearPos(p).IsEq(((rect *)pe)->GetP1())) {
					GetAllHit(pe->GetLy(), &((rect *)pe)->GetP1(), und);
					GetAllHit(pe->GetLy(), &((rect *)pe)->GetPos(), und);
					GetAllHit(pe->GetLy(), &((rect *)pe)->GetP2(), und);
				} else if (pe->GetNearPos(p).IsEq(((rect *)pe)->GetPos())) {
					GetAllHit(pe->GetLy(), &((rect *)pe)->GetPos(), und);
					GetAllHit(pe->GetLy(), &((rect *)pe)->GetP1(), und);
					GetAllHit(pe->GetLy(), &((rect *)pe)->GetP2(), und);
				} else {
					GetAllHit(pe->GetLy(), &((rect *)pe)->GetP2(), und);
					GetAllHit(pe->GetLy(), &((rect *)pe)->GetPos(), und);
					GetAllHit(pe->GetLy(), &((rect *)pe)->GetP1(), und);
				}
				break;
			case container::T_GROUP:
				gEditEditMode = ED_GROUP;
				gEditSeles.Add(pe->Copy());
				edit_pos::Add((container *)gEditSeles.Ref());
				GetAllHit((group *)pe, und);
				break;
			default:
				goto edlu_err;
			}
		}
	}
	if (!fNoCheck) {
		SetLineExpand();
	} else if (edit_pos::GetPosTop()->next == INVALID_HANDLE_VALUE) {
		edit_pos::GetPosTop()->next = NULL;
	}
	if (edit_pos::GetPosR2Top()->next == INVALID_HANDLE_VALUE) {
		edit_pos::GetPosR2Top()->next = NULL;
	}
	if (und) {
		gTmpUndone.Add(und);
	}
	if (pe->IsGroup()) {
		if (((group *)pe)->GetExt()) {
			p = ((group *)pe)->GetExt()->GetPb();
		} else if (((group *)pe)->GetName() == gPadIndex) {
			p = ((land *)((group *)pe)->GetCont().Ref())->GetPos();
		} else {
			p = GetGroupOrigPos(((group *)pe)->GetCont());
		}
	} else if (pe->IsRect() || pe->IsLand()) {
		p = pe->GetPos();
	} else if (pe->IsLine()) {
		p = pe->GetNearPos(p);
	} else {
		GetMouseSnapPos(p);
	}
	SetCursor(cursor);
	BeginEditMove(p);
	return;
edlu_err:
	SetCursor(cursor);
	if (und) {
		und->Undone()->Del();
	}
	while (!gEditSeles.IsNull()) {
		((container *)gEditSeles.Ref())->Del();
	}
	MessageBeep(MB_OK);
}

static void FnEditKey()
{
	if (gEditKey == EDKEY_LY) {
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 0)) {
			GetScrHandle()->Invalidate();
		}
		SetLyApMenu();
	} else if (gEditKey == EDKEY_SEL) {
		CallDialogBox(gApp->Handle(), IDD_SELEEDIT, SeleEditDlg);
	}
}

static void FnEditEsc()
{
	if (!gTmpUndone.IsNull()) {
		((undone *)gTmpUndone.Prev()->Cut())->Undone()->Del();
	}
	while (!gEditSeles.IsNull()) {
		((container *)gTmpUndone.Ref()->Cut())->Del();
	}
	FnSetEsc();
	ResetStack(MOD_EDIT, IDST_EDIT1);
	gApp->PushButton(IDB_EDIT, TRUE);
	gApp->EnableButton(IDB_VALUEIN, FALSE);
	gApp->EnableButton(IDB_CROSSIN, FALSE);
	gApp->UpdateButton();
	gEditMode = EDMOD_NON;
	SetLyApMenu();
}

#define EDIT_MIRRO	0x80

static void __fastcall EditMoveSub(pos p)
{
	double dx;
	double dy;
	if (gEditEditMode == ED_ARC_R) {
		arc *pa = (arc *)gEditSeles.Ref();
		if (p.Distance(pa->GetPos()) < GetCirAp(pa->GetAp()) - ERR_MINI) {
			double dx = p.x - pa->GetPos().x;
			double dy = p.y - pa->GetPos().y;
			double l = sqrt(dx * dx + dy * dy) + ERR_MINI;
			p.Set(GetCirAp(pa->GetAp()) * dx / l, GetCirAp(pa->GetAp()) * dy / l);
		}
		pos p1, p2;
		if (edit_pos::IsArcR2()) {
			p1 = pa->GetStPos();
			p2 = pa->GetEndPos();
		} else {
			p2 = pa->GetStPos();
			p1 = pa->GetEndPos();
		}
		pa->ChengeR(p.Distance(pa->GetPos()));
		dx = pa->GetStPos().x - p1.x;
		dy = pa->GetStPos().y - p1.y;
		for (pedit_pos mem = edit_pos::GetPosTop(); mem; mem = mem->next) {
			for (edit_pos::edit_elem *pel = mem->pe; pel; pel = pel->next) {
				if (pel->pc->IsLine()) {
					((line *)pel->pc)->Expand(mem->p, dx, dy);
				} else {
					pel->pc->Move(dx, dy);
				}
			}
			if (mem->p.x > -1e100) {
				mem->p.Move(dx, dy);
			}
		}
		dx = pa->GetEndPos().x - p2.x;
		dy = pa->GetEndPos().y - p2.y;
		for (pedit_pos mem = edit_pos::GetPosR2Top(); mem; mem = mem->next) {
			for (edit_pos::edit_elem *pel = mem->pe; pel; pel = pel->next) {
				if (pel->pc->IsLine()) {
					((line *)pel->pc)->Expand(mem->p, dx, dy);
				} else {
					pel->pc->Move(dx, dy);
				}
			}
			if (mem->p.x > -1e100) {
				mem->p.Move(dx, dy);
			}
		}
	} else {
		if (gEditEditMode == ED_ARC_ANG) {
			pos po;
			arc *ar = (arc *)gEditSeles.Ref();
			if (IsFEditArcSt()) {
				po = ar->GetStPos();
			} else {
				po = ar->GetEndPos();
			}
			ChengeArcAngle(p, ar);
			if (IsFEditArcSt()) {
				dx = ar->GetStPos().x - po.x;
				dy = ar->GetStPos().y - po.y;
			} else {
				dx = ar->GetEndPos().x - po.x;
				dy = ar->GetEndPos().y - po.y;
			}
		} else {
			dx = p.x - gEditBase.x;
			dy = p.y - gEditBase.y;
		}
		for (pedit_pos mem = edit_pos::GetPosTop(); mem; mem = mem->next) {
			for (edit_pos::edit_elem *pel = mem->pe; pel; pel = pel->next) {
				if (pel->pc->IsLine()) {
					((line *)pel->pc)->Expand(mem->p, dx, dy);
				} else {
					pel->pc->Move(dx, dy);
				}
			}
			if (mem->p.x > -1e100) {
				mem->p.Move(dx, dy);
			}
		}
	}
	gEditBase = p;
}

static void FnEditMv1()
{
	HDC hDC = GetScrHandle()->GetScrDC();
	DrawEditSeleCros(hDC);
	pos p;
	GetMouseSnapPos(p);
	EditMoveSub(p);
	DrawEditSeleCros(hDC);
	FlushScr();
}

static void __fastcall DrawNodeBase(HDC hDC, const node &bas)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsElem()) {
			if (gPcb->IsFDisp((element *)pc)) {
				SelectObject(hDC, gPcb->GetPen((element *)pc));
				Draw(hDC, (element *)pc);
			}
		} else if (pc->IsGroup()) {
			DrawNodeBase(hDC, ((group *)pc)->GetCont());
		}
	}
}

static void __fastcall EditAddPcb(const pos &p)
{
	HDC hDC = GetScrHandle()->GetScrDC();
	DrawEditSeleCros(hDC);
	EditMoveSub(p);
	if (IsOutSize(gEditSeles)) {
		MessageBeep(MB_OK);
		pos pp;
		GetMouseSnapPos(pp);
		EditMoveSub(pp);
		DrawEditSeleCros(hDC);
	} else {
		undone *pu = MoveTmpUndoToUndone(1);
		SetROP2(hDC, R2_MERGEPEN);
		DrawNodeBase(hDC, gEditSeles);
		int n = 0;
		while (! gEditSeles.IsNull()) {
			element *pc = (element *)gEditSeles.Ref()->Cut();
			if (pc->IsLine() && pc->GetPos().IsEq(((line*)pc)->GetPe()) && (n++ || !gEditSeles.IsNull())) {
				pc->Del();
				n = 0x80000001;
			} else {
				gPcb->Add(pc);
				pu->AddData((container *)pc);
			}
		}
		if (n < 0) {
			gApp->Invalidate();
		}
		AddOwnUndone(pu);
		PopStack();
		SetFFileChenge();
	}
}

static inline int IsEditLineAng8(container *pc)
{
	return pc->IsLine() && gEditEditMode == ED_EXPAND && (IsFEditAng8() || (GetScrHandle()->GetMouseKey() & MK_SHIFT));
}

#define MAX_EDIT_HIT_ELEM	256

static int __fastcall GetEditHitLyNum(const node &bas, const pos &p, element *pe[])
{
	int n = 0;
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsLand() || pc->IsRect()) {
			if (((element *)pc)->GetPos() != p) {
				return 0;
			}
			pe[n] = (element *)pc;
			if (++n >= MAX_EDIT_HIT_ELEM) {
				break;
			}
		} else {
			return 0;
		}
	}
	return n;
}

static int __fastcall GetEditHitPos(pos &p, const node &bas, element *pe[], int num)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsElem()) {
			if (gPcb->IsFEdit((element *)pc) && ((element *)pc)->IsHit(p)) {
				for (int i = 0; i < num; ++i) {
					if (((element *)pc)->GetLy() != pe[i]->GetLy()) {
						continue;
					}
					double err = GetHitInPoint((element *)pc);
					if (pc->IsLand()) {
						p = ((land *)pc)->GetPos();
						return 0;
					} else if (pc->IsLine()) {
						if (p.Distance2(((line *)pc)->GetPos()) < p.Distance2(((line *)pc)->GetPe())) {
							if (p.Distance(((line *)pc)->GetPos()) < err) {
								p = ((line *)pc)->GetPos();
								return 0;
							}
						} else {
							if (p.Distance(((line *)pc)->GetPe()) < err) {
								p = ((line *)pc)->GetPe();
								return 0;
							}
						}
					} else if (pc->IsRect()) {
						if (p.Distance2(((rect *)pc)->GetP1()) < p.Distance2(((rect *)pc)->GetP2())) {
							if (p.Distance2(((rect *)pc)->GetPos()) < p.Distance2(((rect *)pc)->GetP1())) {
								if (p.Distance(((rect *)pc)->GetPos()) < err) {
									p = ((rect *)pc)->GetPos();
									return 0;
								}
							} else {
								if (p.Distance(((rect *)pc)->GetP1()) < err) {
									p = ((rect *)pc)->GetP1();
									return 0;
								}
							}
						} else {
							if (p.Distance2(((rect *)pc)->GetPos()) < p.Distance2(((rect *)pc)->GetP2())) {
								if (p.Distance(((rect *)pc)->GetPos()) < err) {
									p = ((rect *)pc)->GetPos();
									return 0;
								}
							} else {
								if (p.Distance(((rect *)pc)->GetP2()) < err) {
									p = ((rect *)pc)->GetP2();
									return 0;
								}
							}
						}
					} else if (pc->IsArc()) {
						if (p.Distance2(((arc *)pc)->GetStPos()) < p.Distance2(((arc *)pc)->GetEndPos())) {
							if (p.Distance(((arc *)pc)->GetStPos()) < err) {
								p = ((arc *)pc)->GetStPos();
								return 0;
							}
						} else {
							if (p.Distance(((arc *)pc)->GetEndPos()) < err) {
								p = ((arc *)pc)->GetEndPos();
								return 0;
							}
						}
					}
				}
			}
		} else if (GetEditHitPos(p, ((group *)pc)->GetCont(), pe, num) == 0) {
			return 0;
		}
	}
	return 1;
}

static int __fastcall GetEditHitPos(double x, double y, pos &p)
{
	if (IsFLineHitHit()) {
		return 1;
	}
	element *pe[MAX_EDIT_HIT_ELEM];
	int n;
	if (((container *)gEditSeles.Ref())->IsLine()
			|| ((container *)gEditSeles.Ref())->IsLand() || ((container *)gEditSeles.Ref())->IsRect()) {
		*pe = (element *)gEditSeles.Ref();
		n = 1;
	} else if (((container *)gEditSeles.Ref())->IsGroup()) {
		if (!((container *)((group *)gEditSeles.Ref())->GetCont().Ref())->IsLand()
				&& !((container *)((group *)gEditSeles.Ref())->GetCont().Ref())->IsRect()) {
			return 1;
		}
		if ((n = GetEditHitLyNum(((group *)gEditSeles.Ref())->GetCont(),
								 ((element *)((group *)gEditSeles.Ref())->GetCont().Ref())->GetPos(), pe)) == 0) {
			return 1;
		}
	} else {
		return 1;
	}
	Scr2NoSnap(x, y, p);
	return GetEditHitPos(p, gPcb->GetElems(), pe, n);
}

static int __fastcall IsHitEditPos(pos& p)
{
	if (!((container*)gEditSeles.Ref())->IsLine() || IsFLineHitHit()) {
		return 0;
	}
	pos pm;
	GetMouseHitPos(pm);
	element *pe = gPcb->IsHit(pm);
	if (pe && ((element*)gEditSeles.Ref())->GetLy() == pe->GetLy()) {
		if (pm.Distance(pe->GetNearPos(pm)) < GetApHitLen(pe)) {
			p = pe->GetNearPos(pm);
			return 1;
		}
	}
	return 0;
}

static void FnEditLu1()
{
	pos p;
	if (gEditEditMode == ED_LINE2
			&& !(GetScrHandle()->GetMouseKey() & MK_CONTROL)
			&& (IsFEditAng8() || (GetScrHandle()->GetMouseKey() & MK_SHIFT))) {
		line *pl = (line *)gEditSeles.Ref();
		if (pl->Ref() == &gEditSeles) {
			GetMouseSnapPos(p);
			p = GetAng8(gEditBase.IsEq(pl->GetPos()) ? pl->GetPe() : pl->GetPos(), p);
		} else {
			pos p1 = pl->GetFarPos(gEditBase);
			pos p2 = pl->Ref()->GetFarPos(gEditBase);
			if (fabs(p1.x - p2.x) < ERR_MINI) {
				GetMouseHitPos(p);
				if (fabs(p.x - p1.x) > fabs(p1.y - p2.y) / 4.0) {
					if (p.x > p1.x) {
						p.x = p1.x + fabs(p1.y - p2.y) / 2.0;
					} else {
						p.x = p1.x - fabs(p1.y - p2.y) / 2.0;
					}
					p.y = (p1.y + p2.y) / 2.0;
				} else {
					p.Set(p1.x, (p1.y + p2.y) / 2.0);
					if (IsSnap()) {
						double x, y;
						Pos2NoSnap(p, x, y);
						Scr2Snap(x, y, p);
						p.x = p1.x;
					}
				}
			} else if (fabs(p1.y - p2.y) < ERR_MINI) {
				GetMouseHitPos(p);
				if (fabs(p.y - p1.y) > fabs(p1.x - p2.x) / 4.0) {
					if (p.y > p1.y) {
						p.y = p1.y + fabs(p1.x - p2.x) / 2.0;
					} else {
						p.y = p1.y - fabs(p1.x - p2.x) / 2.0;
					}
					p.x = (p1.x + p2.x) / 2.0;
				} else {
					p.Set((p1.x + p2.x) / 2.0, p1.y);
					if (IsSnap()) {
						double x, y;
						Pos2NoSnap(p, x, y);
						Scr2Snap(x, y, p);
						p.y = p1.y;
					}
				}
			} else {
				p.Set(p1.x, p2.y);
				double len = gEditBase.Distance2(p);
				pos px(p2.x, p1.y);
				double er;
				if (len > (er = gEditBase.Distance2(px))) {
					len = er;
					p = px;
				}
				if ((er = fabs(p1.x - p2.x) - fabs(p1.y - p2.y)) >= 0.0) {
					if (er == 0) {
						p = gEditBase;
					} else {
						double dy = p2.y - p1.y;
						if (dy * (p2.x - p1.x) < 0.0) {
							dy = -dy;
						}
						px.Set(p1.x + dy, p2.y);
						if ((er = gEditBase.Distance2(px)) < len) {
							len = er;
							p = px;
						}
						px.Set(p2.x - dy, p1.y);
						if (gEditBase.Distance2(px) < len) {
							p = px;
						}
					}
				} else {
					double dx = p2.x - p1.x;
					if (dx * (p2.y - p1.y) < 0.0) {
						dx = - dx;
					}
					px.Set(p2.x, dx + p1.y);
					if ((er = gEditBase.Distance2(px)) < len) {
						len = er;
						p = px;
					}
					px.Set(p1.x, p2.y - dx);
					if (gEditBase.Distance2(px) < len) {
						p = px;
					}
				}
			}
		}
	} else if (!(IsEditLineAng8((container *)gEditSeles.Ref()) || (GetScrHandle()->GetMouseKey() & MK_SHIFT))) {
		if ((GetScrHandle()->GetMouseKey() & MK_CONTROL)
				|| GetEditHitPos(GetScrHandle()->GetMouseX(), GetScrHandle()->GetMouseY(), p)) {
			GetMouseSnapPos(p);
		}
	} else {
		GetMouseSnapPos(p);
		if (!(GetScrHandle()->GetMouseKey() & MK_CONTROL)
				&& !IsHitEditPos(p) && IsEditLineAng8((container *)gEditSeles.Ref())) {
			p = GetAng8(((line *)gEditSeles.Ref())->GetFarPos(p), p);
		}
	}
	EditAddPcb(p);
}

static void FnEditValu1()
{
	EditAddPcb(pos(gValueX, gValueY));
}

static void EditMirror()
{
	HDC hDC = GetScrHandle()->GetScrDC();
	DrawEditSeleCros(hDC);
	for (container *pc = (container *)gEditSeles.Ref(); pc != &gEditSeles; pc = pc->Ref()) {
		if (!pc->IsLine()) {
			pc->Mirror(gEditBase);
		}
	}
	double px2 = 2.0 * gEditBase.x;
	for (pedit_pos mem = edit_pos::GetPosTop(); mem; mem = mem->next) {
		if (mem->p.x > -1e100) {
			pos pb = mem->p;
			pos pn(px2 - pb.x, pb.y);
			for (edit_pos::edit_elem *pel = mem->pe; pel; pel = pel->next) {
				if (pel->pc->IsLine()) {
					((line *)pel->pc)->Expand(pb, pn);
				}
			}
			mem->p = pn;
		}
	}
	for (pedit_pos mem = edit_pos::GetPosR2Top(); mem; mem = mem->next) {
		if (mem->p.x > -1e100) {
			pos pb = mem->p;
			pos pn(px2 - pb.x, pb.y);
			for (edit_pos::edit_elem *pel = mem->pe; pel; pel = pel->next) {
				if (pel->pc->IsLine()) {
					((line *)pel->pc)->Expand(pb, pn);
				}
			}
			mem->p = pn;
		}
	}
	edit_pos::ChengArcR2();
	if (((container *)gEditSeles.Ref())->IsGroup() && ((group *)gEditSeles.Ref())->IsExtParts()) {
		SwapParts(((group *)gEditSeles.Ref())->GetCont());
	}
	DrawEditSeleCros(hDC);
	FlushScr();
}

static void EditRotate()
{
	static double ang1 = M_PI / 2.0;
	HDC hDC = GetScrHandle()->GetScrDC();
	DrawEditSeleCros(hDC);
	double ang;
	if (IsFEditRot() || (GetScrHandle()->GetMouseKey() & MK_CONTROL)) {
		if (CallDialogBoxParam(gApp->Handle(), IDD_GETROT, GetRotAngDlg, (LPARAM)&ang1)) {
			DrawEditSeleCros(hDC);
			FlushScr();
			return;
		}
		ang = ang1;
	} else {
		ang = M_PI / 2.0;
	}
	for (container *pc = (container *)gEditSeles.Ref(); pc != &gEditSeles; pc = pc->Ref()) {
		if (!pc->IsLine()) {
			pc->Rotate(gEditBase, ang);
		}
	}
	double co = cos(ang), si = sin(ang);
	for (pedit_pos mem = edit_pos::GetPosTop(); mem; mem = mem->next) {
		if (mem->p.x > -1e100) {
			double dx, dy;
			pos pb = mem->p, pn;
			pn.x = (dx = pb.x - gEditBase.x) * co - (dy = pb.y - gEditBase.y) * si + gEditBase.x;
			pn.y = dy * co + dx * si + gEditBase.y;
			for (edit_pos::edit_elem *pel = mem->pe; pel; pel = pel->next) {
				if (pel->pc->IsLine()) {
					((line *)pel->pc)->Expand(pb, pn);
				}
			}
			mem->p = pn;
		}
	}
	DrawEditSeleCros(hDC);
	FlushScr();
}

static void FnEditKey1()
{
	switch (gEditKey) {
	case EDKEY_SEL:
		if (gEditEditMode == ED_LINE2 || gEditEditMode == ED_GROUP || ((container*)gEditSeles.Ref())->IsLine()) {
			CallDialogBoxParam(gApp->Handle(), IDD_SELEEDIT1, SeleEdit1Dlg, gApp->IsButtonEnable(IDB_ROTATE));
		}
		break;
	case EDKEY_VAL:
		if (gEditEditMode == ED_ARC_R || gEditEditMode == ED_ARC_ANG) {
			valuein_t val;
			arc *pa = (arc *)gEditSeles.Ref();
			if ((val.f = gEditEditMode == ED_ARC_R) != 0) {
				val.x = pa->GetR();
			} else if (IsFEditArcSt()) {
				val.x = pa->GetSt() * 180.0 / M_PI;
			} else {
				val.x = pa->GetEnd() * 180.0 / M_PI;
			}
			if (CallDialogBoxParam(gApp->Handle(), IDD_EDITARCIN, EditArcInDlg, (LPARAM)&val) == 0) {
				double x, y;
				if (val.f) {
					x = pa->GetPos().x + val.x;
					y = pa->GetPos().y;
				} else {
					x = pa->GetPos().x + pa->GetR() * cos(val.x *= M_PI / 180.0);
					y = pa->GetPos().y + pa->GetR() * sin(val.x);
				}
				RunEditValue(x, y);
			}
		} else {
			EdKeyValSub(0);
		}
		break;
	case EDKEY_MIRROR:
		if (gEditEditMode == ED_GROUP) {
			EditMirror();
		}
		break;
	case EDKEY_ROT:
		if (gEditEditMode == ED_GROUP || ((container *)gEditSeles.Ref())->IsRect()) {
			EditRotate();
		}
		break;
	}
}

static void FnEditUnd1()
{
	DrawTmpUndoneElem();
	while (! gTmpUndone.IsNull()) {
		((undone *)gTmpUndone.Prev())->Undone()->Del();
	}
	while (! gEditSeles.IsNull()) {
		((container *)gEditSeles.Ref())->Del();
	}
	PopStack();
}

static container *__fastcall GetEraseGroup(container *pc)
{
	if (pc->GetOwner()) {
		group *pg = pc->GetOwner();
		if (pg->GetName() == gTextIndex || pg->GetName() == gPadIndex) {
			return (container *)pg;
		}
		if (pg->GetOwner() && pg->GetOwner()->GetName() == gPaintedIndex) {
			return (container *)pg->GetOwner();
		}
	}
	return pc;
}

static void FnEraseLu()
{
	pos p;
	GetMouseHitPos(p);
	element *pe;
	if ((pe = gPcb->IsHit(p)) != NULL) {
		if (GetScrHandle()->GetMouseKey() & MK_CONTROL) {
			pe = (element *)GetEraseGroup(pe);
		} else {
			pe = (element *)pe->GetTopOwner();
		}
		if (IsFDeleteFast()) {
			AddOwnUndone(undone::New(pe->Cut(), pe->Ref()));
			gEditSeles.Add(pe);
			DrawEditSeleCros(GetScrHandle()->GetScrDC(), R2_BLACK);
			gEditSeles.Ref()->Cut();
			SetFFileChenge();
		} else {
			PushStack(2, IDST_ERASE2);
			SetButtonOnState();
			gApp->EnableButton(IDB_LAYER, FALSE);
			EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_GRAYED);
			DrawMenuBar(GetAppWnd());
			gApp->UpdateButton();
			SetOwnTmpUndone(undone::New(pe->Cut(), pe->Ref()));
			gEditSeles.Add(pe);
			DrawEditSeleCros(GetScrHandle()->GetScrDC(), R2_COPYPEN);
		}
	} else {
		MessageBeep(MB_OK);
	}
}

static void FnEraseLMv()
{
	GetScrHandle()->BeginSeleBox();
	PushStack(1, IDST_ERASE1);
	SetButtonOnState();
	gApp->EnableButton(IDB_LAYER, FALSE);
	EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_GRAYED);
	DrawMenuBar(GetAppWnd());
	gApp->EnableButton(IDB_ZOOMIN, FALSE);
	gApp->EnableButton(IDB_ZOOMOUT, FALSE);
	gApp->EnableButton(IDB_SELECT, FALSE);
	gApp->UpdateButton();
}

static void FnEraseKey()
{
	if (gEditKey ==  EDKEY_LY) {
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 0)) {
			GetScrHandle()->Invalidate();
		}
		SetLyApMenu();
	} else if (gEditKey == EDKEY_SEL) {
		CallDialogBox(gApp->Handle(), IDD_DELETE_FAST, EraseSetDlg);
	}
}

static void FnEraseEsc()
{
	FnSetEsc();
	ResetStack(MOD_ERASE, IDST_ERASE);
	GetScrHandle()->OnLdMove();
	gApp->PushButton(IDB_ERASE, TRUE);
//	gApp->EnableButton(IDB_SELECT, FALSE);
	gApp->EnableButton(IDB_VALUEIN, FALSE);
	gApp->EnableButton(IDB_CROSSIN, FALSE);
	gApp->UpdateButton();
	gEditMode = EDMOD_NON;
	SetLyApMenu();
}

static void FnEraseLu1()
{
	GetScrHandle()->EndSeleBox();
	pos p, p1 = GetScrHandle()->GetSeleBase();
	GetMouseHitPos(p);
	undone *und;
	if ((und = GetInBoxElems(p, p1, gEditSeles, gPcb->GetElems())) != NULL) {
		SetOwnTmpUndone(und);
		DrawEditSeleCros(GetScrHandle()->GetScrDC(), R2_COPYPEN);
		MoveStack(2, IDST_ERASE2);
	} else {
		PopStack();
		MessageBeep(MB_OK);
	}
}

static void FnEraseEsc1()
{
	GetScrHandle()->EndSeleBox();
	PopStack();
}

static void FnEraseLu2()
{
	DrawEditSeleCros(GetScrHandle()->GetScrDC());
	while (! gEditSeles.IsNull()) {
		gEditSeles.Ref()->Cut();
	}
	MoveTmpUndoToUndone();
	SetFFileChenge();
	PopStack();
}

static void FnEraseKey2()
{
	if (gEditKey == EDKEY_RTN) {
		FnEraseLu2();
		FlushScr();
	}
}

static void FnEraseEsc2()
{
	DrawTmpUndoneElem();
	while (! gTmpUndone.IsNull()) {
		((undone *)gTmpUndone.Prev())->Undone()->Del();
	}
	while (! gEditSeles.IsNull()) {
		gEditSeles.Ref()->Cut();
	}
	PopStack();
}

static void __fastcall SeleDrawAndSetEditSele(element *pel)
{
	gEditCross = pel->Cut();
	HDC hDC = GetScrHandle()->GetScrDC();
	SelectObject(hDC, gSelePen);
	SetROP2(hDC, R2_COPYPEN);
	Draw(hDC, pel);
	FlushScr();
}

static int __fastcall IsNoPaint(element *pe)
{
	group *grp = pe->GetOwner();
	return grp == NULL || (grp = grp->GetOwner()) == NULL || grp->GetName() != gPaintedIndex;
}

static void FnParaLu()
{
	GetMouseHitPos(gClicBase);
	element *elem;
	if ((elem = gPcb->IsHit(gClicBase)) != NULL) {
		if ((GetScrHandle()->GetMouseKey() & MK_SHIFT) || IsFParaInt()) {
			SeleDrawAndSetEditSele(elem);
			PushStack(1, IDST_PARA1);
			SetButtonOnState();
			gApp->EnableButton(IDB_LAYER, FALSE);
			gApp->EnableButton(IDB_SELECT, FALSE);
			EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_GRAYED);
			EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_AP, MF_GRAYED);
			DrawMenuBar(GetAppWnd());
			gApp->UpdateButton();
		} else if (((elem->IsLine() && elem->GetPos() != ((line *)elem)->GetPe())
					|| (elem->IsArc() && ((arc *)elem)->GetR() >= (STEP_MINI / 2.0))) && IsNoPaint(elem)) {
			SeleDrawAndSetEditSele(elem);
			if (IsFParaOwn()) {
				PushStack(2, IDST_PARA2);
			} else {
				PushStack(3, IDST_PARA3);
			}
			SetButtonOnState();
			gApp->EnableButton(IDB_LAYER, FALSE);
			gApp->EnableButton(IDB_SELECT, FALSE);
			EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_GRAYED);
			EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_AP, MF_GRAYED);
			DrawMenuBar(GetAppWnd());
			gApp->UpdateButton();
		} else {
			MessageBeep(MB_OK);
		}
	} else {
		MessageBeep(MB_OK);
	}
}

static void FnParaKey()
{
	if (gEditKey ==  EDKEY_LY) {
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 1)) {
			GetScrHandle()->Invalidate();
		}
		SetLyApMenu();
	} else if (gEditKey == EDKEY_SEL) {
		CallDialogBox(gApp->Handle(), IDD_SELEPARA, SeleParaDlg);
	}
}

static void FnParaEsc()
{
	FnSetEsc();
	ResetStack(MOD_PARA, IDST_PARA0);
	gApp->PushButton(IDB_PARA, TRUE);
	gApp->EnableButton(IDB_VALUEIN, FALSE);
	gApp->EnableButton(IDB_CROSSIN, FALSE);
	gApp->UpdateButton();
	gEditMode = EDMOD_PARA;
	SetLyApMenu();
}

static void RedrawAndBackEditCross()
{
	HDC hDC = GetScrHandle()->GetScrDC();
	SelectObject(hDC, gPcb->GetPen(gEditCross));
	SetROP2(hDC, R2_COPYPEN);
	Draw(hDC, gEditCross);
	FlushScr();
	gEditCross->Ref()->Add(gEditCross);
	gEditCross = NULL;
}

static void FnParaLu1()
{
	pos p;
	GetMouseHitPos(p);
	element *elem;
	if ((elem = gPcb->IsHit(p)) != NULL) {
		pos px[2];
		int rt = GetInterval(gEditCross, elem, px);
		RedrawAndBackEditCross();
		if (rt || px[0].IsEq(px[1])) {
			DispMessage(IDS_SHRT_MSG);
		} else {
			static parain_t in = { { 1.0 }, { 10.0 }, 1, parain_t::T_KAN };
			if (CallDialogBoxParam(gApp->Handle(), IDD_SENTLINE, SentLineDlg, (LPARAM)&in) == 0) {
				HDC hDC = GetScrHandle()->GetScrDC();
				SelectObject(hDC, gPcb->GetPen(gEditLayer));
				SetROP2(hDC, R2_MERGEPEN);
				pos p0((px[0].x + px[1].x) / 2.0, (px[0].y + px[1].y) / 2.0);
				double ang = atan2(px[1].y - px[0].y, px[1].x - px[0].x);
				double stepX = in.pitch * cos(ang);
				double stepY = in.pitch * sin(ang);
				ang += M_PI / 2.0;
				double dx = in.len / 2.0 * cos(ang);
				double dy = in.len / 2.0 * sin(ang);
				line *pli;
				undone *tmUnd = undone::New();
				if (in.n & 1) {
					if (IsOutSize(pli = line::New(gEditLayer, gEditLineCir, 0,
												  pos(p0.x - dx, p0.y - dy), pos(p0.x + dx, p0.y + dy)))) {
						pli->Del();
					} else {
						Draw(hDC, pli);
						gPcb->Add(pli);
						tmUnd->AddData(pli);
					}
					px[0] = px[1] = p0;
				} else {
					px[0].Set(p0.x - stepX / 2.0, p0.y - stepY / 2.0);
					px[1].Set(p0.x + stepX / 2.0, p0.y + stepY / 2.0);
				}
				for (int i = in.n >> 1; --i >= 0;) {
					px[0].Move(stepX, stepY);
					px[1].Move(-stepX, -stepY);
					if (IsOutSize(pli = line::New(gEditLayer, gEditLineCir, 0,
												  pos(px[0].x - dx, px[0].y - dy), pos(px[0].x + dx, px[0].y + dy)))) {
						pli->Del();
					} else {
						Draw(hDC, pli);
						gPcb->Add(pli);
						tmUnd->AddData(pli);
					}
					if (IsOutSize(pli = line::New(gEditLayer, gEditLineCir, 0,
												  pos(px[1].x - dx, px[1].y - dy), pos(px[1].x + dx, px[1].y + dy)))) {
						pli->Del();
					} else {
						Draw(hDC, pli);
						gPcb->Add(pli);
						tmUnd->AddData(pli);
					}
				}
				AddOwnUndone(tmUnd);
				SetFFileChenge();
			}
		}
		PopStack();
	} else {
		MessageBeep(MB_OK);
	}
}

static void FnParaUnd1()
{
	RedrawAndBackEditCross();
	PopStack();
}

static void FnParaLu2()
{
	pos p;
	GetMouseHitPos(p);
	static parain_t in = { { 1.0 }, { 10.0 }, 1, parain_t::T_ARC };
	element *pel = gEditCross;
	if (gEditCross->IsLine()) {
		in.mod = in.T_SING;
	} else {
		in.mod = in.T_ARC;
	}
	RedrawAndBackEditCross();
	if (CallDialogBoxParam(gApp->Handle(), IDD_SENTLINE, SentLineDlg, (LPARAM)&in) == 0) {
		HDC hDC = GetScrHandle()->GetScrDC();
		SelectObject(hDC, gPcb->GetPen(gEditLayer));
		SetROP2(hDC, R2_MERGEPEN);
		undone *und = undone::New();
		if (pel->IsLine()) {
			double a, dx = ((line *)pel)->GetPe().x - ((line *)pel)->GetPos().x;
			pos p0;
			if (fabs(dx) < ERR_MINI) {
				p0.Set(((line *)pel)->GetPos().x, p.y);
			} else if ((a = (((line *)pel)->GetPe().y - ((line *)pel)->GetPos().y) / dx), fabs(a) < ERR_MINI) {
				p0.Set(p.x, ((line *)pel)->GetPos().y);
			} else {
				double x = (p.y - ((line *)pel)->GetPos().y + (p.x - ((line *)pel)->GetPos().x) / a) / (a + 1 / a);
				p0.Set(((line *)pel)->GetPos().x + x, ((line *)pel)->GetPos().y + a * x);
			}
			if (p == p0) {
				MessageBeep(MB_OK);
			} else {
				double ang = atan2(p.y - p0.y, p.x - p0.x);
				double pitX = in.pitch * cos(ang);
				double pitY = in.pitch * sin(ang);
				ang = atan2(((line *)pel)->GetPe().y - ((line *)pel)->GetPos().y,
							((line *)pel)->GetPe().x - ((line *)pel)->GetPos().x);
				double dx = in.len * cos(ang) / 2.0;
				double dy = in.len * sin(ang) / 2.0;
				for (int i = in.n; --i >= 0;) {
					p0.Move(pitX, pitY);
					if (IsOutSize(pel = line::New(gEditLayer, gEditLineCir, 0,
												  pos(p0.x - dx, p0.y - dy), pos(p0.x + dx, p0.y + dy)))) {
						pel->Del();
					} else {
						Draw(hDC, pel);
						gPcb->Add(pel);
						und->AddData(pel);
					}
				}
			}
		} else if (pel->IsArc()) {
			double r = ((arc *)pel)->GetR();
			double dr = p.Distance(pel->GetPos()) >= r ? in.pitch : -in.pitch;
			double st = ((arc *)pel)->GetSt();
			double en = ((arc *)pel)->GetEnd();
			p = pel->GetPos();
			for (int i = in.n; --i >= 0;) {
				if ((r += dr), r <= 0.0) {
					break;
				}
				if (IsOutSize(pel = arc::New(gEditLayer, gEditLineCir, p, r, st, en))
						|| GetCirAp(gEditLineCir) >= r + ERR_MINI) {
					pel->Del();
				} else {
					Draw(hDC, pel);
					gPcb->Add(pel);
					und->AddData(pel);
				}
			}
		}
		AddOwnUndone(und);
		SetFFileChenge();
	}
	PopStack();
}

static inline int IsParaLine(element *pe, int ly)
{
	return pe->GetLy() == ly && ((pe->IsLine() && !pe->GetPos().IsEq(((line *)pe)->GetPe()))
								 || (pe->IsArc() && ((arc *)pe)->GetEnd() - ((arc *)pe)->GetSt() >= STEP_MINI * M_PI / 180.0
									 && ((arc *)pe)->GetR() > STEP_MINI)) && IsNoPaint(pe);
}

static element *__fastcall GetParaHit(const pos &p, const node &bas, int ly)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			element *pe;
			if ((pe = GetParaHit(p, ((group *)pc)->GetCont(), ly)) != NULL) {
				return pe;
			}
		} else if (IsParaLine((element *)pc, ly) && ((element *)pc)->IsHit(p)) {
			return (element *)pc;
		}
	}
	return NULL;
}

static int __fastcall GetParaHitList(const pos &p, int ly,
									 element *end, const node &bas, container *pc, element **ppElem)
{
	element **ptr = ppElem;
	while (pc != &bas) {
		if (pc->IsGroup()) {
			if (GetParaHitList(p, ly, end,
							   ((group *)pc)->GetCont(), (container *)((group *)pc)->GetCont().Ref(), ppElem) == 0) {
				return 0;
			}
		} else if (pc->tmp.work == 0 && IsParaLine((element *)pc, ly) && ((element *)pc)->IsEqPos(p)) {
			pc->tmp.work = 1;
			*ppElem = (element *)pc;
			*++ppElem = NULL;
			if ((element *)pc == end) {
				return 0;
			} else if (GetParaHitList(((element *)pc)->GetFarPos(p), ly, end,
									  gPcb->GetElems(), (container *)gPcb->GetElems().Ref(), ppElem) == 0) {
				return 0;
			} else {
				*--ppElem = NULL;
			}
		}
		pc = pc->Ref();
	}
	while (*ptr) {
		(*ptr)->tmp.work = 0;
		*ptr++ = NULL;
	}
	return 1;
}

static int __fastcall GetParaHitList(const pos &p, int ly, element *end)
{
	ResetMem();
	*(element **)gTempMemPtr = NULL;
	gPcb->ClearWork();
	return GetParaHitList(p, ly, end, gPcb->GetElems(), (container *)gPcb->GetElems().Ref(), (element **)gTempMemPtr);
}

static void FnParaLu3()
{
	GetMouseHitPos(gEditBase);
	element *pe;
	if (gEditCross->IsHit(gEditBase)) {
		*(int *)gTempMem = 0;
		PushStack(4, IDST_PARA4);
	} else if ((pe = GetParaHit(gEditBase, gPcb->GetElems(), gEditCross->GetLy())) != NULL) {
		int rt;
		if (gEditCross->IsLine()) {
			if ((rt = GetParaHitList(gEditCross->GetPos(), gEditCross->GetLy(), pe)) != 0) {
				rt = GetParaHitList(((line *)gEditCross)->GetPe(), gEditCross->GetLy(), pe);
			}
		} else if ((rt = GetParaHitList(((arc *)gEditCross)->GetStPos(), gEditCross->GetLy(), pe)) != 0) {
			rt = GetParaHitList(((arc *)gEditCross)->GetEndPos(), gEditCross->GetLy(), pe);
		}
		if (rt) {
			MessageBeep(MB_OK);
			FnParaUnd1();
		} else {
			HDC hDC = GetScrHandle()->GetScrDC();
			SelectObject(hDC, gSelePen);
			SetROP2(hDC, R2_COPYPEN);
			ResetMem();
			undone *und = undone::New();
			while (*(element **)gTempMemPtr) {
				element *pe = *(element **)gTempMemPtr;
				*(element ***)&gTempMemPtr = (element **)gTempMemPtr + 1;
				und->AddData(pe->Cut(), pe->Ref());
				gEditSeles.Add(pe);
				Draw(hDC, pe);
			}
			SetOwnTmpUndone(und);
			PushStack(4, IDST_PARA4);
		}
	} else {
		MessageBeep(MB_OK);
	}
}

static void __fastcall InParaLine(const pos &p, HDC hDC, const parain_t &in)
{
	pos px1, px2;
	double a, dx = ((line *)gEditCross)->GetPe().x - gEditCross->GetPos().x, pitX, pitY;
	if (fabs(dx) < ERR_MINI) {
		px1.Set(gEditCross->GetPos().x, gClicBase.y);
		px2.Set(gEditCross->GetPos().x, gEditBase.y);
		pitY = 0.0;
		pitX = p.x < gEditCross->GetPos().x ? -in.pitch : in.pitch;
	} else if ((a = (((line *)gEditCross)->GetPe().y - gEditCross->GetPos().y) / dx), fabs(a) < ERR_MINI) {
		px1.Set(gClicBase.x, gEditCross->GetPos().y);
		px2.Set(gEditBase.x, gEditCross->GetPos().y);
		pitX = 0.0;
		pitY = p.y < gEditCross->GetPos().y ? -in.pitch : in.pitch;
	} else {
		double b = gEditCross->GetPos().y - a * gEditCross->GetPos().x;
		double aa = a + 1 / a;
		double x = (gClicBase.y + gClicBase.x / a - b) / aa;
		px1.Set(x, a * x + b);
		x = (gEditBase.y + gEditBase.x / a - b) / aa;
		px2.Set(x, a * x + b);
		x = (p.y + p.x / a - b) / aa;
		double ang = atan2(p.y - (a * x + b), p.x - x);
		pitX = in.pitch * cos(ang);
		pitY = in.pitch * sin(ang);
	}
	undone *und = undone::New();
	for (int n = in.n; --n >= 0;) {
		px1.Move(pitX, pitY);
		px2.Move(pitX, pitY);
		line *li;
		if (IsOutSize(li = line::New(gEditLayer, gEditLineCir, 0, px1, px2))) {
			li->Del();
		} else {
			Draw(hDC, li);
			gPcb->Add(li);
			und->AddData(li);
		}
	}
	AddOwnUndone(und);
	SetFFileChenge();
}

static double __fastcall AdjustAng(double an, arc *pa)
{
	double ang = pa->GetSt();
	while (an < ang) {
		an += M_PI * 2.0;
	}
	ang = pa->GetEnd();
	while (an > ang) {
		an -= M_PI * 2.0;
	}
	return an;
}

static void __fastcall InParaArc(const pos &p, HDC hDC, const parain_t &in)
{
	pos p0 = ((arc *)gEditCross)->GetPos();
	if (gEditBase == p0 || gClicBase == p0) {
		MessageBeep(MB_OK);
	} else {
		double r, dr;
		if (p0.Distance(p) < (r = ((arc *)gEditCross)->GetR())) {
			dr = -in.pitch;
		} else {
			dr = in.pitch;
		}
		double st = AdjustAng(atan2(gClicBase.y - p0.y, gClicBase.x - p0.x), (arc *)gEditCross);
		double en = AdjustAng(atan2(gEditBase.y - p0.y, gEditBase.x - p0.x), (arc *)gEditCross);
		if (st > en) {
			Swap(st, en);
		}
		arc *pa;
		undone *und = undone::New();
		for (int n = in.n; --n >= 0;) {
			if ((r += dr), r <= 0.0) {
				break;
			}
			if (IsOutSize(pa = arc::New(gEditLayer, gEditLineCir, p0, r, st, en))
					|| GetCirAp(gEditLineCir) >= r + ERR_MINI) {
				pa->Del();
			} else {
				Draw(hDC, pa);
				gPcb->Add(pa);
				und->AddData(pa);
			}
		}
		AddOwnUndone(und);
		SetFFileChenge();
	}
}

static para_t *__fastcall SetParaLineTmpSt()
{
	para_t *pp = (para_t *)ResetMem();
	pp->f = 0;
	if (gEditCross->IsLine()) {
		pp->mode = pp->T_LINE;
		if (((element *)gEditSeles.Ref())->IsEqPos(gEditCross->GetPos())) {
			pp->pn = gEditCross->GetPos();
		} else {
			pp->pn = ((line *)gEditCross)->GetPe();
		}
		double x, y, dx = ((line *)gEditCross)->GetPe().x - gEditCross->GetPos().x;
		if (fabs(dx) < ERR_MINI) {
			pp->f |= pp->B_V;
			x = gEditCross->GetPos().x;
			y = gClicBase.y;
		} else if ((pp->da.li.a = (((line *)gEditCross)->GetPe().y - gEditCross->GetPos().y) / dx),
				   fabs(pp->da.li.a) < ERR_MINI) {
			pp->da.li.a = 0.0;
			pp->f |= pp->B_H;
			x = gClicBase.x;
			y = gEditCross->GetPos().y;
		} else {
			double b = gEditCross->GetPos().y - pp->da.li.a * gEditCross->GetPos().x;
			x = (gClicBase.y + gClicBase.x / pp->da.li.a - b) / (pp->da.li.a + 1.0 / pp->da.li.a);
			y = pp->da.li.a * x + b;
		}
		SetPos(pp->da.li.px, x, y);
		if (pp->f & pp->B_V) {
			if (pp->pn.y > pp->da.li.px.y) {
				pp->da.li.a = 1e100;
			} else {
				pp->da.li.a = -1e100;
			}
		}
	} else {
		pp->mode = pp->T_ARC;
		if (((element *)gEditSeles.Ref())->IsEqPos(((arc *)gEditCross)->GetStPos())) {
			pp->pn = ((element *)gEditSeles.Ref())->GetNearPos(((arc *)gEditCross)->GetStPos());
			pp->da.ar.st = ((arc *)gEditCross)->GetSt();
		} else {
			pp->f |= pp->B_ST;
			pp->pn = ((element *)gEditSeles.Ref())->GetNearPos(((arc *)gEditCross)->GetEndPos());
			pp->da.ar.end = ((arc *)gEditCross)->GetEnd();
		}
		pp->da.ar.r = ((arc *)gEditCross)->GetR();
		pp->da.ar.pb = gEditCross->GetPos();
		double ang = atan2(gClicBase.y - gEditCross->GetPos().y, gClicBase.x - gEditCross->GetPos().x);
		while (ang < ((arc *)gEditCross)->GetSt()) {
			ang += M_PI * 2.0;
		}
		if (pp->f & pp->B_ST) {
			pp->da.ar.st = ang;
		} else {
			pp->da.ar.end = ang;
		}
	}
	return pp;
}

static void __fastcall SetParaLineTmpEnd(para_t *pp, element *pe)
{
	para_t *ppn = pp + 1;
	ppn->f = 0;
	if (pe->IsLine()) {
		ppn->mode = pp->T_LINE;
		ppn->da.li.px = pp->pn;
		double dx, dy;
		if (pe->GetPos().IsEq(pp->pn)) {
			dx = ((line *)pe)->GetPe().x - pe->GetPos().x;
			dy = ((line *)pe)->GetPe().y - pe->GetPos().y;
		} else {
			dx = pe->GetPos().x - ((line *)pe)->GetPe().x;
			dy = pe->GetPos().y - ((line *)pe)->GetPe().y;
		}
		if (fabs(dx) < ERR_MINI) {
			ppn->f |= pp->B_V;
			ppn->pn.x = pe->GetPos().x;
			ppn->pn.y = gEditBase.y;
			if (dy >= 0.0) {
				ppn->da.li.a = 1e100;
			} else {
				ppn->da.li.a = -1e100;
			}
		} else if (ppn->da.li.a = dy / dx, fabs(ppn->da.li.a) < ERR_MINI) {
			ppn->da.li.a = 0.0;
			ppn->f |= pp->B_H;
			ppn->pn.x = gEditBase.x;
			ppn->pn.y = pe->GetPos().y;
		} else {
			double b = pe->GetPos().y - ppn->da.li.a * pe->GetPos().x;
			ppn->pn.x = (gEditBase.y + gEditBase.x / ppn->da.li.a - b) / (ppn->da.li.a + 1.0 / ppn->da.li.a);
			ppn->pn.y = ppn->da.li.a * ppn->pn.x + b;
		}
	} else {
		ppn->mode = pp->T_ARC;
		if (((arc *)pe)->GetStPos().IsEq(pp->pn)) {
			ppn->f |= pp->B_ST;
			ppn->da.ar.st = ((arc *)pe)->GetSt();
		} else {
			ppn->da.ar.end = ((arc *)pe)->GetEnd();
		}
		ppn->da.ar.pb = pe->GetPos();
		ppn->da.ar.r = ((arc *)pe)->GetR();
		if (ppn->f & pp->B_ST) {
			ppn->da.ar.end = atan2(gEditBase.y - pe->GetPos().y, gEditBase.x - pe->GetPos().x);
		} else {
			ppn->da.ar.st = atan2(gEditBase.y - pe->GetPos().y, gEditBase.x - pe->GetPos().x);
		}
	}
	ppn[1].mode = pp->T_NON;
}

static int __fastcall SetParaLineTmp()
{
	if (gEditSeles.IsNull()) {
		return 0;
	}
	para_t *pp = SetParaLineTmpSt();
	for (element *pe = (element *)gEditSeles.Ref(); pe->Ref() != &gEditSeles; pe = pe->Ref()) {
		para_t *ppn = pp + 1;
		ppn->f = 0;
		if (pe->IsLine()) {
			ppn->mode = pp->T_LINE;
			ppn->da.li.px = pp->pn;
			if (pe->GetPos().IsEq(pp->pn)) {
				ppn->pn = ((line *)pe)->GetPe();
			} else {
				ppn->pn = pe->GetPos();
			}
			double dx = ppn->pn.x - ppn->da.li.px.x;
			if (fabs(dx) < ERR_MINI) {
				ppn->f |= pp->B_V;
				if (ppn->pn.y > ppn->da.li.px.y) {
					ppn->da.li.a = 1e100;
				} else {
					ppn->da.li.a = -1e100;
				}
			} else if (ppn->da.li.a = (ppn->pn.y - ppn->da.li.px.y) / dx, fabs(ppn->da.li.a) < ERR_MINI) {
				ppn->da.li.a = 0.0;
				ppn->f |= pp->B_H;
			}
		} else {
			ppn->mode = pp->T_ARC;
			if (((pos)pp->pn).Distance2(((arc *)pe)->GetStPos()) < ((pos)pp->pn).Distance2(((arc *)pe)->GetEndPos())) {
				ppn->f |= pp->B_ST;
				ppn->pn = pe->Ref()->GetNearPos(((arc *)pe)->GetEndPos());
			} else {
				ppn->pn = pe->Ref()->GetNearPos(((arc *)pe)->GetStPos());
			}
			ppn->da.ar.pb = pe->GetPos();
			ppn->da.ar.r = ((arc *)pe)->GetR();
			ppn->da.ar.st = ((arc *)pe)->GetSt();
			ppn->da.ar.end = ((arc *)pe)->GetEnd();
		}
		pp = ppn;
	}
	SetParaLineTmpEnd(pp, (element *)gEditSeles.Prev());
	return 1;
}

static double __fastcall AdjustAng(double ang)
{
	while (ang >= M_PI * 2.0) {
		ang -= M_PI * 2.0;
	}
	while (ang < 0.0) {
		ang += M_PI * 2.0;
	}
	return ang;
}

static void __fastcall SetParaPitch(const pos &px, para_t *pp, double pit)
{
	if (pp->mode & pp->T_LINE) {
		if (pp->f & pp->B_V) {
			if (pp->pn.x < px.x) {
				pp->da.li.dx = pit;
			} else {
				pp->da.li.dx = -pit;
			}
			pp->da.li.dy = 0.0;
		} else if (pp->f & pp->B_H) {
			pp->da.li.dx = 0.0;
			if (pp->pn.y < px.y) {
				pp->da.li.dy = pit;
			} else {
				pp->da.li.dy = -pit;
			}
		} else {
			double b = pp->pn.y - pp->da.li.a * pp->pn.x;
			double x = (px.y + px.x / pp->da.li.a - b) / (pp->da.li.a + 1.0 / pp->da.li.a);
			double y = pp->da.li.a * x + b;
			x = px.x - x;
			y = px.y - y;
			double l = sqrt(x * x + y * y);
			pp->da.li.dx = pit * x / l;
			pp->da.li.dy = pit * y / l;
		}
	} else {
		double dx = px.x - pp->da.ar.pb.x;
		double dy = px.y - pp->da.ar.pb.y;
		if (sqrt(dx * dx + dy * dy) < pp->da.ar.r) {
			pp->da.ar.dr = -pit;
		} else {
			pp->da.ar.dr = pit;
		}
	}
}

static void __fastcall SetParaLineLinePx(pos &px, para_t *pp, para_t *ppn)
{
	if (!(pp->f & ppn->f & (pp->B_V | pp->B_H)) && fabs(pp->da.li.a - ppn->da.li.a) >= ERR_MINI) {
		double dx = pp->pn.x - pp->da.li.px.x;
		double dy = pp->pn.y - pp->da.li.px.y;
		double l = sqrt(dx * dx + dy * dy);
		double co = dx / l;
		double si = dy / l;
		dx = ppn->pn.x - pp->pn.x;
		dy = ppn->pn.y - pp->pn.y;
		double y = dy * co - dx * si;
		dx = dx * co + dy * si;
		l = sqrt(dx * dx + y * y);
		co = dx / l;
		si = y / l;
		px.Set(pp->pn.x + pp->da.li.dx * co - pp->da.li.dy * si, pp->pn.y + pp->da.li.dy * co + pp->da.li.dx * si);
	}
}

static void __fastcall SetParaLineArcPx(pos &px, para_t *pp, para_t *ppn)
{
	double dx = pp->pn.x - pp->da.li.px.x;
	double dy = pp->pn.y - pp->da.li.px.y;
	double l = sqrt(dx * dx + dy * dy);
	double co = dx / l;
	double si = dy / l;
	dx = ppn->da.ar.pb.x - pp->pn.x;
	dy = ppn->da.ar.pb.y - pp->pn.y;
	double x = dx * co + dy * si;
	dy = dy * co - dx * si;
	l = sqrt(x * x + dy * dy);
	co = x / l;
	si = dy / l;
	if (ppn->f & pp->B_ST) {
		px.Set(pp->pn.x + (pp->da.li.dy * co + pp->da.li.dx * si), pp->pn.y - (pp->da.li.dx * co - pp->da.li.dy * si));
	} else {
		px.Set(pp->pn.x - (pp->da.li.dy * co + pp->da.li.dx * si), pp->pn.y + (pp->da.li.dx * co - pp->da.li.dy * si));
	}
}

static void __fastcall SetParaArcLinePx(pos &px, para_t *pp, para_t *ppn)
{
	double dx = pp->pn.x - pp->da.ar.pb.x;
	double dy = pp->pn.y - pp->da.ar.pb.y;
	double l = sqrt(dx * dx + dy * dy);
	double co = dx / l;
	double si = dy / l;
	dx = ppn->pn.x - pp->pn.x;
	dy = ppn->pn.y - pp->pn.y;
	double x = dx * co + dy * si;
	dy = dy * co - dx * si;
	l = sqrt(x * x + dy * dy);
	double co1 = x / l;
	double si1 = dy / l;
	double y;
	if (pp->f & pp->B_ST) {
		x = pp->da.ar.dr * si1;
		y = -pp->da.ar.dr * co1;
	} else {
		x = -pp->da.ar.dr * si1;
		y = pp->da.ar.dr * co1;
	}
	px.Set(pp->pn.x + x * co - y * si, pp->pn.y + y * co + x * si);
}

static void __fastcall SetParaArcArcPx(pos &px, para_t *pp, para_t *ppn)
{
	if (((pos)pp->da.ar.pb).IsEq(ppn->da.ar.pb)) {
		double dx = px.x - pp->da.ar.pb.x;
		double dy = px.y - pp->da.ar.pb.y;
		double l = sqrt(dx * dx + dy * dy);
		if (pp->f & pp->B_ST) {
			px.Set(pp->da.ar.pb.x + l * cos(pp->da.ar.end), pp->da.ar.pb.y + l * sin(pp->da.ar.end));
		} else {
			px.Set(pp->da.ar.pb.x + l * cos(pp->da.ar.st), pp->da.ar.pb.y + l * sin(pp->da.ar.st));
		}
	} else if (ppn->f & pp->B_ST) {
		double dx = ppn->da.ar.r * cos(ppn->da.ar.st);
		double dy = ppn->da.ar.r * sin(ppn->da.ar.st);
		if (pp->f & pp->B_ST) {
			if (pp->da.ar.dr > 0.0) {
				px.Set(ppn->da.ar.pb.x + dx * 2.0, ppn->da.ar.pb.y + dy * 2.0);
			} else {
				px.Set(ppn->da.ar.pb.x + dx / 2.0, ppn->da.ar.pb.y + dy / 2.0);
			}
		} else {
			if (pp->da.ar.dr > 0.0) {
				px.Set(ppn->da.ar.pb.x + dx / 2.0, ppn->da.ar.pb.y + dy / 2.0);
			} else {
				px.Set(ppn->da.ar.pb.x + dx * 2.0, ppn->da.ar.pb.y + dy * 2.0);
			}
		}
	} else {
		double dx = ppn->da.ar.r * cos(ppn->da.ar.end);
		double dy = ppn->da.ar.r * sin(ppn->da.ar.end);
		if (pp->f & pp->B_ST) {
			if (pp->da.ar.dr > 0.0) {
				px.Set(ppn->da.ar.pb.x + dx / 2.0, ppn->da.ar.pb.y + dy / 2.0);
			} else {
				px.Set(ppn->da.ar.pb.x + dx * 2.0, ppn->da.ar.pb.y + dy * 2.0);
			}
		} else {
			if (pp->da.ar.dr > 0.0) {
				px.Set(ppn->da.ar.pb.x + dx * 2.0, ppn->da.ar.pb.y + dy * 2.0);
			} else {
				px.Set(ppn->da.ar.pb.x + dx / 2.0, ppn->da.ar.pb.y + dy / 2.0);
			}
		}
	}
}

static void __fastcall SetPxBase(pos &px, para_t *pp)
{
	if (pp->mode == pp->T_LINE) {
		if (pp->f & pp->B_V) {
			px.y = pp->da.li.px.y;
		} else if (pp->f & pp->B_H) {
			px.x = pp->da.li.px.x;
		} else {
			double dx = pp->pn.x - pp->da.li.px.x;
			double dy = pp->pn.y - pp->da.li.px.y;
			double a = dy / dx;
			double b = pp->pn.y - a * pp->pn.x;
			double x = (px.y + px.x / a - b) / (a + 1.0 / a);
			px.Set(pp->da.li.px.x + px.x - x, pp->da.li.px.y + px.y - (a * x + b));
		}
	}
}

static void __fastcall SetParaPitch(pos px, double pit)
{
	para_t *pp = (para_t *)gTempMem;
	if (pp->mode == pp->T_NON) {
		return;
	}
	SetPxBase(px, pp);
	for (para_t *ppn = pp + 1; ppn->mode != pp->T_NON; pp = ppn, ++ppn) {
		SetParaPitch(px, pp, pit);
		if (pp->mode == pp->T_LINE) {
			if (ppn->mode == pp->T_LINE) {
				SetParaLineLinePx(px, pp, ppn);
			} else {
				SetParaLineArcPx(px, pp, ppn);
			}
		} else {
			if (ppn->mode == pp->T_LINE) {
				SetParaArcLinePx(px, pp, ppn);
			} else {
				SetParaArcArcPx(px, pp, ppn);
			}
		}
	}
	SetParaPitch(px, pp, pit);
}

static void __fastcall SetParaStPx(para_t *pp, double mul, pos &px)
{
	if (pp->mode == pp->T_LINE) {
		px.Set(pp->da.li.px.x + pp->da.li.dx * mul, pp->da.li.px.y + pp->da.li.dy * mul);
	} else {
		double ang;
		if (pp->f & pp->B_ST) {
			ang = pp->da.ar.st;
		} else {
			ang = pp->da.ar.end;
		}
		double r = pp->da.ar.r + pp->da.ar.dr * mul;
		if (r < 0.0) {
			r = fabs(r);
			ang += M_PI;
		}
		px.Set(pp->da.ar.pb.x + r * cos(ang), pp->da.ar.pb.y + r * sin(ang));
	}
}

static element *__fastcall GetParaLineLine(para_t *pp, para_t *ppn, double mul, pos &px)
{
	pos pb;
	if (IsFParaSt()) {
		pb = px;
	} else {
		pb.Set(px.x, px.y);
	}
	if (pp->f & pp->B_V) {
		if (ppn->f & pp->B_V) {
			px.y = pp->pn.y;
		} else if (ppn->f & pp->B_H) {
			px.Set(pp->pn.x + mul * pp->da.li.dx, pp->pn.y + mul * ppn->da.li.dy);
		} else {
			double b = pp->pn.y + mul * ppn->da.li.dy - ppn->da.li.a * (pp->pn.x + mul * ppn->da.li.dx);
			px.x = pp->pn.x + mul * pp->da.li.dx;
			px.y = ppn->da.li.a * px.x + b;
		}
	} else if (pp->f & pp->B_H) {
		if (ppn->f & pp->B_V) {
			px.Set(pp->pn.x + mul * ppn->da.li.dx, pp->pn.y + mul * pp->da.li.dy);
		} else if (ppn->f & pp->B_H) {
			px.x = pp->pn.x;
		} else {
			double b = pp->pn.y + mul * ppn->da.li.dy - ppn->da.li.a * (pp->pn.x + mul * ppn->da.li.dx);
			px.y = pp->pn.y + mul * pp->da.li.dy;
			px.x = (px.y - b) / ppn->da.li.a;
		}
	} else {
		double b = px.y - pp->da.li.a * px.x;
		if (ppn->f & pp->B_V) {
			px.x = pp->pn.x + mul * ppn->da.li.dx;
			px.y = pp->da.li.a * px.x + b;
		} else if (ppn->f & pp->B_H) {
			px.y = pp->pn.y + mul * ppn->da.li.dy;
			px.x = (px.y - b) / pp->da.li.a;
		} else if (fabs(pp->da.li.a - ppn->da.li.a) < ERR_MINI) {
			px.Set(pp->pn.x + mul * pp->da.li.dx, pp->pn.y + mul * pp->da.li.dy);
		} else {
			double b1 = pp->pn.y + mul * ppn->da.li.dy - ppn->da.li.a * (pp->pn.x + mul * ppn->da.li.dx);
			px.x = (b1 - b) / (pp->da.li.a - ppn->da.li.a);
			px.y = pp->da.li.a * px.x + b;
		}
	}
	pos pn(px.x, px.y);
	if (pb.IsEq(pn)) {
		return NULL;
	}
	return line::New(gEditLayer, gEditLineCir, 0, pb, pn);
}

static element *__fastcall GetParaLineArc(para_t *pp, para_t *ppn, double mul, pos &px)
{
	pos p1, p2;
	double r = ppn->da.ar.r + mul * ppn->da.ar.dr;
	if (pp->f & pp->B_V) {
		p1.x = p2.x = px.x;
		double l = fabs(px.x - ppn->da.ar.pb.x);
		if (l > r + 1e-8) {
			return NULL;
		}
		if ((l = r * r - l * l) < ERR_MINI) {
			l = 0.0;
		} else {
			l = sqrt(l);
		}
		p1.y = ppn->da.ar.pb.y + l;
		p2.y = ppn->da.ar.pb.y - l;
	} else {
		double b = px.y - ppn->da.ar.pb.y - pp->da.li.a * (px.x - ppn->da.ar.pb.x);
		double ab = pp->da.li.a * b;
		double a2_1 = pp->da.li.a * pp->da.li.a + 1;
		double sq = a2_1 * r * r - b * b;
		if (sq < -ERR_MINI) {
			return NULL;
		}
		if (sq < ERR_MINI) {
			sq = 0.0;
		} else {
			sq = sqrt(sq);
		}
		p1.x = (-ab + sq) / a2_1;
		p2.x = (-ab - sq) / a2_1;
		p1.y = pp->da.li.a * p1.x + b + ppn->da.ar.pb.y;
		p2.y = pp->da.li.a * p2.x + b + ppn->da.ar.pb.y;
		p1.x += ppn->da.ar.pb.x;
		p2.x += ppn->da.ar.pb.x;
	}
	pos pb;
	if (IsFParaSt()) {
		pb = px;
	} else {
		pb.Set(px.x, px.y);
	}
	if (((pos)pp->pn).Distance2(p1) < ((pos)pp->pn).Distance2(p2)) {
		px = p1;
	} else {
		px = p2;
	}
	p1.Set(px.x, px.y);
	if (pb.IsEq(px)) {
		return NULL;
	}
	return line::New(gEditLayer, gEditLineCir, 0, pb, px);
}

static element *__fastcall GetParaArcLine(para_t *pp, para_t *ppn, double mul, pos &px)
{
	double r = pp->da.ar.r + mul * pp->da.ar.dr;
	pos p1, p2;
	if (ppn->f & pp->B_V) {
		p1.x = p2.x = ppn->pn.x + mul * ppn->da.li.dx;
		double l = fabs(p1.x - pp->da.ar.pb.x);
		if (l > r + 1e-8) {
			return NULL;
		}
		if ((l = r * r - l * l) < ERR_MINI) {
			l = 0.0;
		} else {
			l = sqrt(l);
		}
		p1.y = pp->da.ar.pb.y + l;
		p2.y = pp->da.ar.pb.y - l;
	} else {
		pos p0(ppn->da.li.px.x + mul * ppn->da.li.dx, ppn->da.li.px.y + mul * ppn->da.li.dy);
		double b = p0.y - pp->da.ar.pb.y - ppn->da.li.a * (p0.x - pp->da.ar.pb.x);
		double ab = ppn->da.li.a * b;
		double a2_1 = ppn->da.li.a * ppn->da.li.a + 1;
		double sq = a2_1 * r * r - b * b;
		if (sq < -ERR_MINI) {
			return NULL;
		}
		if (sq < ERR_MINI) {
			sq = 0.0;
		} else {
			sq = sqrt(sq);
		}
		p1.x = (-ab + sq) / a2_1;
		p2.x = (-ab - sq) / a2_1;
		p1.y = ppn->da.li.a * p1.x + b + pp->da.ar.pb.y;
		p2.y = ppn->da.li.a * p2.x + b + pp->da.ar.pb.y;
		p1.x += pp->da.ar.pb.x;
		p2.x += pp->da.ar.pb.x;
	}
	double ang = AdjustAng(atan2(px.y - pp->da.ar.pb.y, px.x - pp->da.ar.pb.x));
	if (p1.Distance2(pp->pn) > p2.Distance2(pp->pn)) {
		p1 = p2;
	}
	double ang1 = AdjustAng(atan2(p1.y - pp->da.ar.pb.y, p1.x - pp->da.ar.pb.x));
	if (fabs(ang - ang1) < ERR_MINI) {
		return NULL;
	}
	px = p1;
	if (pp->f & pp->B_ST) {
		if (ang > ang1) {
			ang -= M_PI * 2.0;
		}
	} else {
		if (ang < ang1) {
			ang += M_PI * 2.0;
		}
	}
	if (GetCirAp(gEditLineCir) >= r + ERR_MINI) {
		return NULL;
	}
	return arc::New(gEditLayer, gEditLineCir, pp->da.ar.pb, r, ang, ang1);
}

static element *__fastcall GetParaArcArc(para_t *pp, para_t *ppn, double mul, pos &px)
{
	double r1 = pp->da.ar.r + mul * pp->da.ar.dr;
	if (r1 <= 0.0) {
		return NULL;
	}
	if (((pos)pp->da.ar.pb).IsEq(ppn->da.ar.pb)) {
		if (pp->f & pp->B_ST) {
			px.Set(pp->da.ar.pb.x + r1 * cos(pp->da.ar.end), pp->da.ar.pb.y + r1 * sin(pp->da.ar.end));
		} else {
			px.Set(pp->da.ar.pb.x + r1 * cos(pp->da.ar.st), pp->da.ar.pb.y + r1 * sin(pp->da.ar.st));
		}
		if (GetCirAp(gEditLineCir) >= r1 + ERR_MINI) {
			return NULL;
		}
		return arc::New(gEditLayer, gEditLineCir, pp->da.ar.pb, r1, pp->da.ar.st, pp->da.ar.end);
	}
	double r2 = ppn->da.ar.r + mul * ppn->da.ar.dr;
	if (r2 <= 0.0) {
		return NULL;
	}
	double l = Distance(pp->da.ar.pb, ppn->da.ar.pb);
	if (r1 + r2 <= l || fabs(r1 - r2) >= l) {
		return NULL;
	}
	double dx = ppn->da.ar.pb.x - pp->da.ar.pb.x;
	double dy = ppn->da.ar.pb.y - pp->da.ar.pb.y;
	l = sqrt(dx * dx + dy * dy);
	double x = (l * l + r1 * r1 - r2 * r2) / (2.0 * l);
	double co1 = dx / l;
	double si1 = dy / l;
	double h = r1 * r1 - x * x;
	if (h < ERR_MINI) {
		h = 0.0;
	} else {
		h = sqrt(h);
	}
	double co2 = x / r1;
	double si2 = h / r1;
	pos p1(r1 * (co1 * co2 + si1 * si2), r1 * (si1 * co2 - co1 * si2));
	pos p2(r1 * (co1 * co2 - si1 * si2), r1 * (si1 * co2 + co1 * si2));
	double ang = AdjustAng(atan2(px.y - pp->da.ar.pb.y, px.x - pp->da.ar.pb.x));
	p1.x += pp->da.ar.pb.x;
	p1.y += pp->da.ar.pb.y;
	p2.x += pp->da.ar.pb.x;
	p2.y += pp->da.ar.pb.y;
	if (p1.Distance2(pp->pn) > p2.Distance2(pp->pn)) {
		p1 = p2;
	}
	double ang1 = AdjustAng(atan2(p1.y - pp->da.ar.pb.y, p1.x - pp->da.ar.pb.x));
	if (fabs(ang - ang1) < ERR_MINI) {
		return NULL;
	}
	px = p1;
	if (pp->f & pp->B_ST) {
		if (ang > ang1) {
			ang -= M_PI * 2.0;
		}
	} else {
		if (ang < ang1) {
			ang += M_PI * 2.0;
		}
	}
	if (GetCirAp(gEditLineCir) >= r1 + ERR_MINI) {
		return NULL;
	}
	return arc::New(gEditLayer, gEditLineCir, pp->da.ar.pb, r1, ang, ang1);
}

static element *__fastcall GetParaElemEnd(para_t *pp, double mul, pos &px)
{
	if (pp->mode == pp->T_LINE)
		return line::New(gEditLayer, gEditLineCir, 0, px,
						 pos(pp->pn.x + mul * pp->da.li.dx, pp->pn.y + mul * pp->da.li.dy));
	double r = pp->da.ar.r + mul * pp->da.ar.dr;
	if (r <= 0.0) {
		return NULL;
	}
	double ang = AdjustAng(atan2(px.y - pp->da.ar.pb.y, px.x - pp->da.ar.pb.x));
	if (pp->f & pp->B_ST) {
		if (ang > pp->da.ar.end) {
			ang -= M_PI * 2.0;
		}
		if (GetCirAp(gEditLineCir) >= r + ERR_MINI) {
			return NULL;
		}
		return arc::New(gEditLayer, gEditLineCir, pp->da.ar.pb, r, ang, pp->da.ar.end);
	}
	if (ang < pp->da.ar.st) {
		ang += M_PI * 2.0;
	}
	if (GetCirAp(gEditLineCir) >= r + ERR_MINI) {
		return NULL;
	}
	return arc::New(gEditLayer, gEditLineCir, pp->da.ar.pb, r, pp->da.ar.st, ang);
}

static void __fastcall InParaLines(HDC hDC, parain_t &in)
{
	if (((para_t *)gTempMem)->mode == para_t::T_NON) {
		return;
	}
	undone *und = undone::New();
	for (int mul = 1; mul <= in.n; ++mul) {
		para_t *pp = (para_t *)gTempMem;
		element *el;
		pos px;
		SetParaStPx(pp, mul, px);
		SetFParaSt();
		for (para_t *ppn = pp + 1; ppn->mode != pp->T_NON; pp = ppn, ++ppn) {
			if (pp->mode == pp->T_LINE) {
				if (ppn->mode == pp->T_LINE) {
					el = GetParaLineLine(pp, ppn, mul, px);
				} else {
					el = GetParaLineArc(pp, ppn, mul, px);
				}
			} else {
				if (ppn->mode == pp->T_LINE) {
					el = GetParaArcLine(pp, ppn, mul, px);
				} else {
					el = GetParaArcArc(pp, ppn, mul, px);
				}
			}
			if (el) {
				if (IsOutSize(el)) {
					el->Del();
					el = NULL;
				} else {
					Draw(hDC, el);
					und->AddData(el);
					gPcb->Add(el);
				}
			} else {
				break;
			}
			ClrFParaSt();
		}
		if (el && (el = GetParaElemEnd(pp, mul, px)) != NULL) {
			if (IsOutSize(el)) {
				el->Del();
			} else {
				Draw(hDC, el);
				und->AddData(el);
				gPcb->Add(el);
			}
		}
	}
	AddOwnUndone(und);
}

static HDC ParaDrawReSeles()
{
	HDC hDC = GetScrHandle()->GetScrDC();
	SetROP2(hDC, R2_COPYPEN);
	DrawEditCrossNorm(hDC, gEditCross);
	DrawEditSelesNorm(hDC, gEditSeles);
	FlushScr();
	while (! gTmpUndone.IsNull()) {
		((undone *)gTmpUndone.Prev())->Undone()->Del();
	}
	gEditCross->Ref()->Add(gEditCross);
	return hDC;
}

static void FnParaLu4()
{
	pos p;
	GetMouseHitPos(p);
	static parain_t in = { { 1.0 }, { 10.0 }, 1, parain_t::T_KU };
	if (CallDialogBoxParam(gApp->Handle(), IDD_SENTLINE, SentLineDlg, (LPARAM)&in) == 0) {
		int mod = SetParaLineTmp();
		HDC hDC = ParaDrawReSeles();
		SetROP2(hDC, R2_MERGEPEN);
		SelectObject(hDC, gPcb->GetPen(gEditLayer));
		if (mod == 0) {
			if (gEditCross->IsLine()) {
				InParaLine(p, hDC, in);
			} else {
				InParaArc(p, hDC, in);
			}
		} else {
			SetParaPitch(p, in.pitch);
			InParaLines(hDC, in);
		}
		SetFFileChenge();
	} else {
		ParaDrawReSeles();
	}
	gEditCross = NULL;
	PopStack(2);
}

static void FnParaUnd4()
{
	HDC hDC = GetScrHandle()->GetScrDC();
	SelectObject(hDC, gPcb->GetPen(gEditCross));
	SetROP2(hDC, R2_COPYPEN);
	for (element *pe = (element *)gEditSeles.Ref(); pe != &gEditSeles; pe = pe->Ref()) {
		Draw(hDC, pe);
	}
	FlushScr();
	if (! gTmpUndone.IsNull()) {
		((undone *)gTmpUndone.Prev())->Undone()->Del();
		GetScrHandle()->Invalidate();
	}
	PopStack();
}

static void FnParaEsc4()
{
	DrawOwnUndone((undone *)gTmpUndone.Prev());
	if (! gTmpUndone.IsNull()) {
		((undone *)gTmpUndone.Prev())->Undone()->Del();
	}
	gEditCross->Ref()->Add(gEditCross);
	gEditCross = NULL;
	PopStack(2);
	GetScrHandle()->Invalidate();
}

static double __fastcall GetAngleIn(double an, double a1, double a2)
{
	while (an > a2) {
		an -= M_PI * 2.0;
	}
	while (an < a1) {
		an += M_PI * 2.0;
	}
	return an;
}

static void FnConnLu()
{
	GetMouseHitPos(gClicBase);
	element *elem;
	int f = 0;
	if ((elem = gPcb->IsHit(gClicBase)) != NULL) {
		if (GetScrHandle()->GetMouseKey() & MK_CONTROL || IsFConnDivi()) {
			element *e1, *e2;
			if (elem->GetOwner()) {
				f = 1;
			} else if (elem->IsLine()) {
				const pos &p = elem->GetPos();
				const pos &pe = ((line *)elem)->GetPe();
				pos px;
				double dx = pe.x - p.x, dy = pe.y - p.y;
				if (fabs(dx) < ERR_MINI) {
					if ((p.y - gClicBase.y) * (pe.y - gClicBase.y) >= 0.0) {
						f = 1;
					} else {
						px.Set(p.x, gClicBase.y);
					}
				} else if (fabs(dy) < ERR_MINI) {
					if ((p.x - gClicBase.x) * (pe.x - gClicBase.x) >= 0.0) {
						f = 1;
					} else {
						px.Set(gClicBase.x, p.y);
					}
				} else {
					double a = dy / dx;
					double x = ((gClicBase.y - p.y) + (gClicBase.x - p.x) / a) / (a + 1.0 / a);
					px.Set(p.x + x, p.y + a * x);
				}
				if (f == 0) {
					e1 = line::New(elem->GetLy(), elem->GetAp(), elem->IsSq(), p, px);
					e2 = line::New(elem->GetLy(), elem->GetAp(), elem->IsSq(), px, pe);
				}
			} else if (elem->IsArc()) {
				double al = ((arc *)elem)->GetSt();
				double ah = ((arc *)elem)->GetEnd();
				double ang = GetAngleIn(atan2(gClicBase.y - elem->GetPos().y, gClicBase.x - elem->GetPos().x), al, ah);
				e1 = arc::New(elem->GetLy(), elem->GetAp(), elem->GetPos(), ((arc *)elem)->GetR(), al, ang);
				e2 = arc::New(elem->GetLy(), elem->GetAp(), elem->GetPos(), ((arc *)elem)->GetR(), ang, ah);
			} else {
				f = 1;
			}
			if (f == 0) {
				HDC hDC = GetScrHandle()->GetScrDC();
				SetROP2(hDC, R2_BLACK);
				Draw(hDC, elem);
				SelectObject(hDC, gPcb->GetPen(elem));
				SetROP2(hDC, R2_MERGEPEN);
				Draw(hDC, e1);
				Draw(hDC, e2);
				undone *und = undone::New(elem->Cut(), elem->Ref());
				und->AddData(e1);
				gPcb->Add(e1);
				und->AddData(e2);
				gPcb->Add(e2);
				AddOwnUndone(und);
				SetFFileChenge();
			} else {
				MessageBeep(MB_OK);
			}
		} else if (GetScrHandle()->GetMouseKey() & MK_SHIFT) {
			if (elem->IsLine() && !elem->IsSq() && !elem->GetOwner()) {
				SeleDrawAndSetEditSele(elem);
				if (*GetString(IDST_CONNR)) {
					PushStack(5, IDST_CONNR);
				} else {
					PushStack(5, IDST_CONN5);
				}
				SetButtonOnState();
				gApp->EnableButton(IDB_SELECT, FALSE);
				gApp->EnableButton(IDB_LAYER, FALSE);
				EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_GRAYED);
				EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_AP, MF_GRAYED);
				DrawMenuBar(GetAppWnd());
				gApp->UpdateButton();
				GetScrHandle()->OffLdMove();
			} else {
				MessageBeep(MB_OK);
			}
		} else if ((elem->IsLine() || elem->IsArc()) && ! elem->GetOwner()) {
			SeleDrawAndSetEditSele(elem);
			PushStack(1, IDST_CONN1);
			SetButtonOnState();
			gApp->EnableButton(IDB_SELECT, FALSE);
			gApp->EnableButton(IDB_LAYER, FALSE);
			EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_GRAYED);
			EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_AP, MF_GRAYED);
			DrawMenuBar(GetAppWnd());
			gApp->UpdateButton();
			GetScrHandle()->OffLdMove();
		} else {
			MessageBeep(MB_OK);
		}
	} else {
		MessageBeep(MB_OK);
	}
}

static void FnConnLMv()
{
	if (IsFConnExpand() || IsFConnDivi() || (GetScrHandle()->GetMouseKey() & MK_CONTROL)) {
		MessageBeep(MB_OK);
	} else {
		if (GetScrHandle()->GetMouseKey() & MK_SHIFT) {
			SetFConnR();
		} else {
			ClrFConnR();
		}
		PushStack(2, IDST_CONN2);
		SetButtonOnState();
		gApp->EnableButton(IDB_LAYER, FALSE);
		gApp->EnableButton(IDB_SELECT, FALSE);
		EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_GRAYED);
		EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_AP, MF_GRAYED);
		DrawMenuBar(GetAppWnd());
		gApp->EnableButton(IDB_ZOOMIN, FALSE);
		gApp->EnableButton(IDB_ZOOMOUT, FALSE);
		gApp->UpdateButton();
		GetScrHandle()->OffLdMove();
		GetScrHandle()->BeginSeleLine();
	}
}

static void FnConnKey()
{
	if (gEditKey == EDKEY_LY) {
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 0)) {
			GetScrHandle()->Invalidate();
		}
		SetLyApMenu();
	} else if (gEditKey == EDKEY_SEL) {
		CallDialogBox(gApp->Handle(), IDD_SELECONN, SeleConnDlg);
	}
}

static void FnConnEsc()
{
	FnSetEsc();
	ResetStack(MOD_CONN, IDST_CONN);
	GetScrHandle()->OnLdMove();
	gApp->PushButton(IDB_CONNECT, TRUE);
	gApp->EnableButton(IDB_VALUEIN, FALSE);
	gApp->EnableButton(IDB_CROSSIN, FALSE);
	gApp->UpdateButton();
	gEditMode = EDMOD_NON;
	SetLyApMenu();
}

static void __fastcall SetConnExpandLine(node &bas, const pos &p)
{
	double dx = ((line *)gEditCross)->GetPe().x - gEditCross->GetPos().x;
	double dy = ((line *)gEditCross)->GetPe().y - gEditCross->GetPos().y;
	pos pn, pb = gEditCross->GetFarPos(gClicBase);
	if (fabs(dx) < ERR_MINI) {
		pn.Set(gEditCross->GetPos().x, p.y);
	} else if (fabs(dy) < ERR_MINI) {
		pn.Set(p.x, gEditCross->GetPos().y);
	} else {
		double a = dy / dx;
		double x = ((p.y - pb.y) + (p.x - pb.x) / a) / (a + 1.0 / a);
		pn.Set(pb.x + x, pb.y + a * x);
	}
	bas.Add(line::New(gEditCross->GetLy(), gEditCross->GetAp(), gEditCross->IsSq(), pb, pn));
}

static arc *__fastcall GetConnExpandArc(const pos &px, const arc *ar, const pos &pa, int f = 0)
{
	double ast, aen, ang = atan2(pa.y - ar->GetPos().y, pa.x - ar->GetPos().x);
	if (px.Distance2(ar->GetStPos()) < px.Distance2(ar->GetEndPos())) {
		if (ar->GetSt() < ar->GetEnd()) {
			ast = ang;
			aen = ar->GetEnd();
		} else {
			ast = ar->GetEnd();
			aen = ang;
		}
	} else {
		if (ar->GetSt() < ar->GetEnd()) {
			ast = ar->GetSt();
			aen = ang;
		} else {
			ast = ang;
			aen = ar->GetSt();
		}
	}
	while (ast > aen) {
		ast -= M_PI * 2.0;
	}
	while (ast < - M_PI) {
		ast += M_PI * 2.0;
		aen += M_PI * 2.0;
	}
	if (f) {
		return arc::New(ar->GetLy(), ar->GetAp(), ar->GetPos(), ar->GetR(), ast, aen);
	}
	return arc::New(ar->GetLy(), ar->GetAp(), ar->GetPos(), ar->GetR(), ast, aen);
}

static int __fastcall SetConnEqLines(line *l1, line *l2, node &bas)
{
	if (l1->GetLy() != l2->GetLy() || l1->GetAp() != l2->GetAp() || (l1->IsSq() == 0) != (l2->IsSq() == 0)) {
		return 1;
	}
	pos pl = l1->GetPos(), ph = l1->GetPe();
	if (fabs(pl.x - ph.x) < ERR_MINI) {
		if (ph.y < pl.y) {
			Swap(pl, ph);
		}
		if (l2->GetPos().y < pl.y) {
			pl = l2->GetPos();
		}
		if (l2->GetPe().y < pl.y) {
			pl = l2->GetPe();
		}
		if (ph.y < l2->GetPos().y) {
			ph = l2->GetPos();
		}
		if (ph.y < l2->GetPe().y) {
			ph = l2->GetPe();
		}
	} else {
		if (ph.x < pl.x) {
			Swap(pl, ph);
		}
		if (l2->GetPos().x < pl.x) {
			pl = l2->GetPos();
		}
		if (l2->GetPe().x < pl.x) {
			pl = l2->GetPe();
		}
		if (ph.x < l2->GetPos().x) {
			ph = l2->GetPos();
		}
		if (ph.x < l2->GetPe().x) {
			ph = l2->GetPe();
		}
	}
	bas.Add(line::New(l1->GetLy(), l1->GetAp(), l1->IsSq(), pl, ph));
	return 0;
}

static double __fastcall GetNearAng(const double &an0, double ang)
{
	while (an0 - M_PI > ang) {
		ang += M_PI * 2.0;
	}
	while (an0 + M_PI < ang) {
		ang -= M_PI * 2.0;
	}
	return ang;
}

static inline double GetNearAng(const double &ang1, const double &ang2, double ang)
{
	return (fabs(GetNearAng(ang1, ang) - ang1) < fabs(GetNearAng(ang2, ang) - ang2)) ? ang1 : ang2;
}

static void __fastcall SetConnExpandArc(node &bas, const pos &p)
{
	double ang = atan2(gClicBase.y - gEditCross->GetPos().y, gClicBase.x - gEditCross->GetPos().x);
	ang = GetNearAng(((arc *)gEditCross)->GetSt(), ((arc *)gEditCross)->GetEnd(), ang);
	double ang1 = atan2(p.y - gEditCross->GetPos().y, p.x - gEditCross->GetPos().x);
	if (fabs(ang - ((arc *)gEditCross)->GetSt()) < fabs(ang - ((arc *)gEditCross)->GetEnd())) {
		ang = GetNearAng(((arc *)gEditCross)->GetSt(), ang1);
		ang1 = ((arc *)gEditCross)->GetEnd();
		if (ang > ang1) {
			ang -= M_PI * 2.0;
		}
	} else {
		ang = ((arc *)gEditCross)->GetSt();
		ang1 = GetNearAng(((arc *)gEditCross)->GetEnd(), ang1);
		if (ang1 < ang) {
			ang1 += M_PI * 2.0;
		}
	}
	bas.Add(arc::New(gEditCross->GetLy(), gEditCross->GetAp(),
					 gEditCross->GetPos(), ((arc *)gEditCross)->GetR(), ang, ang1));
}

static void FnConnLu1()
{
	pos p;
	GetMouseHitPos(p);
	element *elem;
	node bas(0);
	int f = 0;
	if ((elem = gPcb->IsHit(p)) == NULL) {
		if (! IsFConnExpand() && !(GetScrHandle()->GetMouseKey() & MK_SHIFT)) {
			f = 1;
		} else if (gEditCross->IsLine()) {
			SetConnExpandLine(bas, p);
		} else {	// Arc
			SetConnExpandArc(bas, p);
		}
	} else if (IsFConnExpand() || (GetScrHandle()->GetMouseKey() & MK_SHIFT)) {
		if (gEditCross->IsLine()) {
			if (elem->IsLand()) {			// Line - Lnad
				SetConnExpandLine(bas, elem->GetPos());
			} else if (elem->IsLine()) {	// Line - Line
				pos px;
				if ((f = GetCrossLineLinePos((line *)gEditCross, (line *)elem, px)) == 0)
					bas.Add(line::New(gEditCross->GetLy(), gEditCross->GetAp(),
									  gEditCross->IsSq(), gEditCross->GetFarPos(gClicBase), px));
			} else if (elem->IsArc()) {	// Line - Arc
				pos px[2];
				if ((f = GetCrossLineArcPos((line *)gEditCross, (arc *)elem, px)) == 0)
					bas.Add(line::New(gEditCross->GetLy(), gEditCross->GetAp(), gEditCross->IsSq(),
									  p.Distance2(px[0]) < p.Distance2(px[1]) ? px[0] : px[1],
									  gEditCross->GetFarPos(gClicBase)));
			} else {
				f = 1;
			}
		} else if (elem->IsLine()) {	// Arc - Line
			pos px[2];
			if ((f = GetCrossLineArcPos((line *)elem, (arc *)gEditCross, px)) == 0)
				bas.Add(GetConnExpandArc(gClicBase,
										 (arc *)gEditCross, gClicBase.Distance2(px[0]) < gClicBase.Distance2(px[1]) ? px[0] : px[1]));
		} else if (elem->IsArc()) {		//Arc - Arc
			pos px[2];
			if ((f = GetCrossArcArcPos((arc *)gEditCross, (arc *)elem, px)) == 0)
				bas.Add(GetConnExpandArc(gClicBase,
										 (arc *)gEditCross, gClicBase.Distance2(px[0]) < gClicBase.Distance2(px[1]) ? px[0] : px[1]));
		} else {
			f = 1;
		}
		elem = NULL;
	} else if (! elem->GetOwner()) {	// connect
		pos px[2];
		if (elem->IsLine()) {
			if (gEditCross->IsLine()) {	// Line - Line
				if ((f = GetCrossLineLinePos((line *)gEditCross, (line *)elem, px[0])) == 0) {
					bas.Add(line::New(gEditCross->GetLy(), gEditCross->GetAp(),
									  gEditCross->IsSq(), gEditCross->GetFarPos(gClicBase), px[0]));
					bas.Add(line::New(elem->GetLy(), elem->GetAp(), elem->IsSq(), elem->GetFarPos(p), px[0]));
				} else if (f == 2) {
					f = SetConnEqLines((line *)gEditCross, (line *)elem, bas);
				}
			} else if ((f = GetCrossLineArcPos((line *)elem, (arc *)gEditCross, px)) == 0) {	// Arc - Line
				if (gClicBase.Distance2(px[0]) > gClicBase.Distance2(px[1])) {
					px[0] = px[1];
				}
				bas.Add(GetConnExpandArc(gClicBase, (arc *)gEditCross, px[0], 1));
				bas.Add(line::New(elem->GetLy(), elem->GetAp(), elem->IsSq(), elem->GetFarPos(p), px[0]));
			}
		} else if (elem->IsArc()) {
			if (gEditCross->IsLine()) {	// Line - Arc
				if ((f = GetCrossLineArcPos((line *)gEditCross, (arc *)elem, px)) == 0) {
					if (p.Distance2(px[0]) > p.Distance2(px[1])) {
						px[0] = px[1];
					}
					bas.Add(line::New(gEditCross->GetLy(), gEditCross->GetAp(),
									  gEditCross->IsSq(), gEditCross->GetFarPos(gClicBase), px[0]));
					bas.Add(GetConnExpandArc(p, (arc *)elem, px[0], 1));
				}
			} else if ((f = GetCrossArcArcPos((arc *)gEditCross, (arc *)elem, px)) == 0) {	// Arc - Arc
				if (p.Distance2(px[0]) > p.Distance2(px[1])) {
					px[0] = px[1];
				}
				bas.Add(GetConnExpandArc(gClicBase, (arc *)gEditCross, px[0], 1));
				bas.Add(GetConnExpandArc(p, (arc *)elem, px[0], 1));
			} else if (f == 2) {
				double st = 0.0, end = 0.0;
				if (((arc *)elem)->GetSt() < ((arc *)gEditCross)->GetSt()) {
					if (((arc *)elem)->GetEnd() <= ((arc *)gEditCross)->GetSt()) {
						st = ((arc *)elem)->GetSt();
						end = ((arc *)gEditCross)->GetEnd();
						f = 0;
					}
				} else {
					if (((arc *)gEditCross)->GetEnd() <= ((arc *)elem)->GetSt()) {
						st = ((arc *)gEditCross)->GetSt();
						end = ((arc *)elem)->GetEnd();
						f = 0;
					}
				}
				if (f == 0) {
					if (end > st + M_PI * 2.0) {
						end -= M_PI * 2.0;
					}
					bas.Add(arc::New(elem->GetLy(), elem->GetAp(), elem->GetPos(), ((arc *)elem)->GetR(), st, end));
				}
			}
		} else {
			f = 1;
		}
	} else {
		f = 1;
	}
	if (f) {
		MessageBeep(MB_OK);
	} else if (bas.IsNull()) {
		MessageBeep(MB_OK);
		FnConnEsc1();
	} else {
		HDC hDC = GetScrHandle()->GetScrDC();
		SetROP2(hDC, R2_BLACK);
		Draw(hDC, gEditCross);
		undone *und = undone::New(gEditCross, gEditCross->Ref());
		gEditCross = NULL;
		if (elem) {
			Draw(hDC, elem);
			elem->Cut();
			und->AddData(elem, elem->Ref());
		}
		SetROP2(hDC, R2_MERGEPEN);
		while (! bas.IsNull()) {
			if (IsOutSize(elem = (element *)bas.Ref()->Cut())) {
				elem->Del();
			} else {
				gPcb->Add(elem);
				SelectObject(hDC, gPcb->GetPen(elem));
				Draw(hDC, elem);
				und->AddData(elem);
			}
		}
		AddOwnUndone(und);
		GetScrHandle()->OnLdMove();
		SetFFileChenge();
		PopStack();
	}
}

static void FnConnUnd1()
{
	RedrawAndBackEditCross();
	GetScrHandle()->OnLdMove();
	PopStack();
}

static void __fastcall BeginConnSeleLine(const pos &p)
{
	gSeleP1 = GetScrHandle()->GetSeleBase();
	gSeleP2 = p;
	gEditFlags |= F_CONDSP;
	DrawEditSeleCros(GetScrHandle()->GetScrDC(), R2_COPYPEN);
}

static void EndConnSeleLine()
{
	DrawEditSeleCros(GetScrHandle()->GetScrDC(), R2_BLACK);
	FlushScr();
	gEditFlags &= ~F_CONDSP;
}

static void FnConnLu2()
{
	pos p;
	GetMouseHitPos(p);
	GetScrHandle()->EndSeleLine();
	BeginConnSeleLine(p);
	GetScrHandle()->OnLdMove();
	PushStack(3, IDST_CONN3);
}

static void FnConnUnd2()
{
	GetScrHandle()->EndSeleLine();
	GetScrHandle()->OnLdMove();
	PopStack();
}

static void FnConnLMv3()
{
	GetScrHandle()->BeginSeleLine();
	GetScrHandle()->OffLdMove();
	PushStack(4, IDST_CONN4);
}

static void FnConnUnd3()
{
	EndConnSeleLine();
	GetScrHandle()->OffLdMove();
	GetScrHandle()->BackSeleLine(gSeleP1);
	PopStack();
}

static void FnConnEsc3()
{
	EndConnSeleLine();
	PopStack(2);
}

typedef struct {
	pos pb, pn;
	double dx, dy, a;
	conns_t **cn;
} cross_check_t;

static int __fastcall GetCrossLineLine(const cross_check_t &cc, const pos &p1, const pos &p2, pos &px)
{
	double dx = p2.x - p1.x;
	double dy = p2.y - p1.y;
	if (fabs(cc.dx) < ERR_MINI) {
		if (fabs(dx) < ERR_MINI) {
			return 1;
		}
		if (fabs(dy) < ERR_MINI) {
			px.Set(cc.pb.x, p1.y);
		} else {
			px.Set(cc.pb.x, p1.y + (cc.pb.x - p1.x) * dy / dx);
		}
	} else if (fabs(cc.dy) < ERR_MINI) {
		if (fabs(dx) < ERR_MINI) {
			px.Set(p1.x, cc.pb.y);
		} else if (fabs(dy) < ERR_MINI) {
			return 1;
		} else {
			px.Set(p1.x + (cc.pb.y - p1.y) * dx / dy, cc.pb.y);
		}
	} else {
		if (fabs(dx) < ERR_MINI) {
			px.Set(p1.x, cc.pb.y + (p1.x - cc.pb.x) * cc.a);
		} else if (fabs(dy) < ERR_MINI) {
			px.Set(cc.pb.x + (p1.y - cc.pb.y) / cc.a, p1.y);
		} else {
			double a = dy / dx;
			if (fabs(a - cc.a) < ERR_MINI) {
				return 1;
			}
			double x = (cc.pb.y - p1.y - cc.a * (cc.pb.x - p1.x)) / (a - cc.a);
			px.Set(p1.x + x, p1.y + a * x);
		}
	}
	return (cc.pb.x - px.x) * (cc.pn.x - px.x) > 0.0
		   || (cc.pb.y - px.y) * (cc.pn.y - px.y) > 0.0
		   || (p1.x - px.x) * (p2.x - px.x) > 0.0
		   || (p1.y - px.y) * (p2.y - px.y) > 0.0;
}

static int __fastcall GetCrossLines(cross_check_t &cc, const node &bas)
{
	int n = 0;
	for (element *pe = (element *)bas.Ref(); pe != &bas; pe = pe->Ref()) {
		if (pe->IsGroup()) {
//			int rt;
//			if ((rt = GetCrossLines(cc, ((group *)pe)->GetCont())) < 0) {
//				return -1;
//			}
//			n += rt;
		} else if (gPcb->IsFEdit(pe) && pe->IsLine() && !pe->IsSq() && pe->GetOwner() == NULL) {
			pos p;
			if (GetCrossLineLine(cc, pe->GetPos(), ((line *)pe)->GetPe(), p) == 0) {
				if (pe->tmp.work) {
					return -1;
				}
				pe->tmp.work = 1;
				++n;
				(*cc.cn)->px = p;
				(*cc.cn)->el = pe;
				++*cc.cn;
			}
		}
	}
	return n;
}

static int __fastcall GetCrossLines(const pos &p1, const pos &p2, conns_t **cn, const node &bas)
{
	cross_check_t cc;
	if (p1.y < p2.y) {
		cc.pb = p1;
		cc.pn = p2;
	} else {
		cc.pb = p2;
		cc.pn = p1;
	}
	cc.cn = cn;
	cc.dy = cc.pn.y - cc.pb.y;
	if (fabs(cc.dx = cc.pn.x - cc.pb.x) > ERR_MINI) {
		cc.a = cc.dy / cc.dx;
	}
	int rt;
	conns_t *cnBase = *cn;
	if ((rt = GetCrossLines(cc, bas)) > 0) {
		conns_t *lim = &cnBase[rt];
		for (conns_t *pc = cnBase; pc < lim; ++pc) {
			pc->l = p1.Distance2(pc->px);
		}
		for (conns_t *pc = cnBase; pc < lim; ++pc) {
			conns_t *pcl = pc;
			double ll = pc->l;
			for (conns_t *pcc = pc + 1; pcc < lim; ++pcc) {
				if (pcc->l < ll) {
					pcl = pcc;
					ll = pcl->l;
				}
			}
			if (pcl != pc) {
				conns_t sw = *pcl;
				*pcl = *pc;
				*pc = sw;
			}
		}
	}
	return rt;
}

typedef struct {
	line *li;
	double dx;
	double dy;
	double a;
	double b;
	const pos *pn;
} deltaLine_t;

static int GetXPoint(deltaLine_t de[2], pos &p)
{
	de[0].dx = de[0].li->GetPe().x - de[0].li->GetPos().x;
	de[0].dy = de[0].li->GetPe().y - de[0].li->GetPos().y;
	de[1].dx = de[1].li->GetPe().x - de[1].li->GetPos().x;
	de[1].dy = de[1].li->GetPe().y - de[1].li->GetPos().y;
	if (fabs(de[0].dx) < ERR_MINI) {
		if (fabs(de[1].dx) < ERR_MINI) {
			return -1;
		}
		p.x = de[0].li->GetPos().x;
		de[1].a = de[1].dy / de[1].dx;
		p.y = de[1].a * p.x + (de[1].b = de[1].li->GetPos().y - de[1].a * de[1].li->GetPos().x);
	} else if (fabs(de[1].dx) < ERR_MINI) {
		p.x = de[1].li->GetPos().x;
		de[0].a = de[0].dy / de[0].dx;
		p.y = de[0].a * p.x + (de[0].b = de[0].li->GetPos().y - de[0].a * de[0].li->GetPos().x);
	} else {
		de[0].a = de[0].dy / de[0].dx;
		de[1].a = de[1].dy / de[1].dx;
		if (fabs(de[0].a - de[1].a) < ERR_MINI) {
			return -1;
		}
		de[0].b = de[0].li->GetPos().y - de[0].a * de[0].li->GetPos().x;
		de[1].b = de[1].li->GetPos().y - de[1].a * de[1].li->GetPos().x;
		p.x = (de[1].b - de[0].b) / (de[0].a - de[1].a);
		p.y = de[0].a * p.x + de[0].b;
	}
	return 0;
}

static double GetParaB(deltaLine_t &de, const pos &px, const pos &pn)
{
	double l = sqrt(de.dx * de.dx + de.dy * de.dy);
	double si = de.dy / l;
	double co = de.dx / l;
	double r;
	if ((pn.y - px.y) *co - (pn.x - px.x) * si < 0) {
		r = -gConnR;
	} else {
		r = gConnR;
	}
	if (fabs(de.dx) < ERR_MINI) {
		if (de.dy < 0.0) {
			return r;
		}
		return -r;
	}
	return (px.y + r * co) - de.a * (px.x - r * si);
}

static int GetConnCirPos(deltaLine_t de[2], pos &px, pos &p)
{
	if (gConnR < 0.01 || GetXPoint(de, px)) {
		return -1;
	}
	de[0].pn = &de[0].li->GetFarPos(px);
	de[1].pn = &de[1].li->GetFarPos(px);;
	double b1 = GetParaB(de[0], px, *de[1].pn);
	double b2 = GetParaB(de[1], px, *de[0].pn);
	if (fabs(de[0].dx) < ERR_MINI) {
		p.x = de[0].li->GetPos().x + b1;
		p.y = de[1].a * p.x + b2;
	} else if (fabs(de[1].dx) < ERR_MINI) {
		p.x = de[1].li->GetPos().x + b2;
		p.y = de[0].a * p.x + b1;
	} else {
		p.x = (b2 - b1) / (de[0].a - de[1].a);
		p.y = de[0].a * p.x + b1;
	}
	return 0;
}

static void GetNormPos(const deltaLine_t &de, const pos& pb, pos& p)
{
	if (fabs(de.dx) < ERR_MINI) {
		p.x = de.li->GetPos().x;
		p.y = pb.y;
	} else if (fabs(de.dy) < ERR_MINI) {
		p.y = de.li->GetPos().y;
		p.x = pb.x;
	} else {
		double b1 = pb.y + pb.x / de.a;
		p.x = (b1 - de.b) / (de.a + 1 / de.a);
		p.y = -p.x / de.a + b1;
	}
}

static int SetConnRLins(line *l1, line *l2, undone *und, HDC hDC)
{
	if (l1->IsSq() || l2->IsSq() || l1->GetLy() != l2->GetLy()) {
		return -1;
	}
	deltaLine_t de[2];
	de[0].li = l1;
	de[1].li = l2;
	pos p1, p2, pb, px;
	if (GetConnCirPos(de, px, pb)) {
		return -1;
	}
	GetNormPos(de[0], pb, p1);
	GetNormPos(de[1], pb, p2);
	double ang1 = atan2(p1.y - pb.y, p1.x - pb.x);
	double ang2 = atan2(p2.y - pb.y, p2.x - pb.x);
	if (ang1 > ang2) {
		Swap(ang1, ang2);
	}
	if (ang2 - ang1 > M_PI) {
		ang1 += M_PI * 2.0;
	}
	line *l11 = line::New(l1->GetLy(), l1->GetAp(), 0, p1, *de[0].pn);
	line *l21 = line::New(l2->GetLy(), l2->GetAp(), 0, p2, *de[1].pn);
	arc * ar = arc::New(l1->GetLy(), l1->GetAp(), pb, gConnR, ang1, ang2);
	if (IsOutSize(l11) || IsOutSize(l21) || IsOutSize(ar)) {
		l11->Del();
		l21->Del();
		ar->Del();
		return -1;
	}
	und->AddData(l1->Cut(), l1->Ref());
	und->AddData(l2->Cut(), l2->Ref());
	und->AddData(l11);
	und->AddData(l21);
	und->AddData(ar);
	gPcb->Add(l11);
	gPcb->Add(l21);
	gPcb->Add(ar);
	SetROP2(hDC, R2_BLACK);
	Draw(hDC, l1);
	Draw(hDC, l2);
	SetROP2(hDC, R2_MERGEPEN);
	SelectObject(hDC, gPcb->GetPen(l11));
	Draw(hDC, l11);
	Draw(hDC, l21);
	Draw(hDC, ar);
	return 0;
}

static void FnConnLu4()
{
	pos p, pb = GetScrHandle()->GetSeleBase();
	GetMouseHitPos(p);
	EndConnSeleLine();
	conns_t *cn = (conns_t *)ResetMem();
	gPcb->ClearWork();
	int n1, n2;
	if ((n1 = GetCrossLines(gSeleP1, gSeleP2, &cn, gPcb->GetElems())) <= 0) {
		MessageBeep(MB_OK);
	} else if ((n2 = GetCrossLines(pb, p, &cn, gPcb->GetElems())) <= 0) {
		MessageBeep(MB_OK);
	} else {
		int n = n1 < n2 ? n1 : n2;
		cn = (conns_t *)ResetMem();
		conns_t *cn1 = cn + n1;
		undone *und = undone::New();
		HDC hDC = GetScrHandle()->GetScrDC();
		if (IsFConnR()) {
			ClrFConnR();
			if ((GetScrHandle()->GetMouseKey() & MK_SHIFT)
					|| CallDialogBoxParam(gApp->Handle(), IDD_CONN_R, ConnRDlg, (LPARAM)&gConnR)) {
				for (int i = 0; i < n; ++i) {
					if (SetConnRLins((line*)cn[i].el, (line*)cn1[i].el, und, hDC)) {
						MessageBeep(MB_OK);
						break;
					}
				}
			}
		} else {
			for (int i = 0; i < n; ++i) {
				pos px;
				line *li1 = (line*)cn[i].el;
				line *li2 = (line*)cn1[i].el;
				if (GetCrossLineLinePos(li1, li2, px)) {
					MessageBeep(MB_OK);
					break;
				}
				line *li11 = line::New(li1->GetLy(), li1->GetAp(), li1->IsSq(), li1->GetFarPos(cn[i].px), px);
				line *li21 = line::New(li2->GetLy(), li2->GetAp(), li2->IsSq(), li2->GetFarPos(cn1[i].px), px);
				if (IsOutSize(li11) || IsOutSize(li21)) {
					li11->Del();
					li21->Del();
					MessageBeep(MB_OK);
					break;
				}
				und->AddData(li1->Cut(), li1->Ref());
				und->AddData(li2->Cut(), li2->Ref());
				gPcb->Add(li11);
				gPcb->Add(li21);
				und->AddData(li11);
				und->AddData(li21);
				SetROP2(hDC, R2_BLACK);
				Draw(hDC, li1);
				Draw(hDC, li2);
				SetROP2(hDC, R2_MERGEPEN);
				SelectObject(hDC, gPcb->GetPen(li11));
				Draw(hDC, li11);
				SelectObject(hDC, gPcb->GetPen(li21));
				Draw(hDC, li21);
			}
		}
		AddOwnUndone(und);
		SetFFileChenge();
	}
	GetScrHandle()->EndSeleLine();
	GetScrHandle()->OnLdMove();
	PopStack(3);
}

static void FnConnEsc4()
{
	EndConnSeleLine();
	GetScrHandle()->EndSeleLine();
	GetScrHandle()->OnLdMove();
	PopStack(3);
}

static void FnConnLu5()
{
	pos p;
	GetMouseHitPos(p);
	element *elem = gPcb->IsHit(p);
	if (elem != NULL && elem->IsLine() && !elem->IsSq()
			&& elem->GetOwner() == NULL && elem->GetLy() == gEditCross->GetLy()) {
		undone *und = undone::New();
		line *li = (line*)gEditCross;
		gEditCross->Prev()->Add(gEditCross);
		gEditCross = NULL;
		HDC hDC = GetScrHandle()->GetScrDC();
		SetROP2(hDC, R2_COPYPEN);
		SelectObject(hDC, gPcb->GetPen(li));
		Draw(hDC, li);
		if ((GetScrHandle()->GetMouseKey() & MK_SHIFT)
				|| CallDialogBoxParam(gApp->Handle(), IDD_CONN_R, ConnRDlg, (LPARAM)&gConnR)) {
			if (SetConnRLins((line*)li, (line*)elem, und, GetScrHandle()->GetScrDC())) {
				MessageBeep(MB_OK);
			} else {
				AddOwnUndone(und);
				SetFFileChenge();
			}
		}
		PopStack();
		GetScrHandle()->OnLdMove();
	}
}

static int __fastcall GetChengeMode(element **ppe, int f)
{
	element *pe = *ppe;
	if (pe->GetOwner() == NULL) {
		if ((f & (MK_SHIFT | MK_CONTROL)) == (MK_SHIFT | MK_CONTROL)) {
			return 5;
		}
		return 0;
	} else if ((f & (MK_SHIFT | MK_CONTROL)) == (MK_SHIFT | MK_CONTROL)) {
		return -1;
	}
	if ((f & MK_SHIFT) || IsFChengeUnit()) {
		if (*((group *)pe->GetTopOwner())->GetName() != ' ') {
			*ppe = (element *)pe->GetTopOwner();
			return 3;
		}
	} else if (f & MK_CONTROL) {
		if (*((group *)pe->GetTopOwner())->GetName() != ' '
				|| (pe->GetOwner()->GetName() == gPadIndex && !pe->GetOwner()->GetOwner())
				|| (pe->GetOwner()->GetName() != gTextIndex
					&& pe->GetOwner()->GetName() != gPadIndex
					&& pe->GetOwner()->GetName() != gPaintedLoop
					&& pe->GetOwner()->GetName() != gPaintedMado
					&& pe->GetOwner()->GetName() != gPaintedLines)) {
			return 4;
		}
	}
	if (pe->GetOwner()->GetName() == gTextIndex) {
		*ppe = (element *)pe->GetOwner();
		return 1;
	}
	if (pe->GetOwner()->GetOwner() && pe->GetOwner()->GetOwner()->GetName() == gPaintedIndex) {
		*ppe = (element *)pe->GetOwner()->GetOwner();
		return 2;
	}
	return 0;
}

static inline int FnChengeLuDlg(chenge_t &chng)
{
	if (chng.f & chng.B_RCT) {
		return CallDialogBoxParam(gApp->Handle(), IDD_CHENGERECT, ChengeRectDlg, (LPARAM)&chng);
	}
	if (chng.f & chng.B_ARC) {
		return CallDialogBoxParam(gApp->Handle(), IDD_CHENGEARC, ChengeArcDlg, (LPARAM)&chng);
	}
	if (chng.f & chng.B_TEXT) {
		return CallDialogBoxParam(gApp->Handle(), IDD_CHENGETEXT, ChengeTextDlg, (LPARAM)&chng);
	}
	if (chng.f & chng.B_PARTS) {
		return CallDialogBoxParam(gApp->Handle(), IDD_CHENGEUNIT, ChengeUnitDlg, (LPARAM)&chng);
	}
	if (chng.f & chng.B_PIN) {
		return CallDialogBoxParam(gApp->Handle(), IDD_CHENGEPIN, ChengePinDlg, (LPARAM)&chng);
	}
	if (chng.f & chng.B_DOT) {
		return CallDialogBoxParam(gApp->Handle(), IDD_CHENDOTLINE, ChengeDotLineDlg, (LPARAM)&chng);
	}
	return CallDialogBoxParam(gApp->Handle(), IDD_CHENGE, ChengeDlg, (LPARAM)&chng);
}

static void __fastcall ChengePaintLayer(const node &bas, int ly)
{
	for (group *pg = (group *)bas.Ref(); pg != &bas; pg = pg->Ref()) {
		const node &b = pg->GetCont();
		for (element *pe = (element *)b.Ref(); pe != &b; pe = pe->Ref()) {
			pe->ChengeLy(ly);
		}
	}
}

static void __fastcall ChengeDraw(HDC hDC, const node &bas, int no)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			ChengeDraw(hDC, ((group *)pc)->GetCont(), no);
		} else if (pc->IsElem() && gPcb->IsFDisp((element *)pc)) {
			if (no >= 0) {
				SelectObject(hDC, gPcb->GetPen((element *)pc));
			}
			Draw(hDC, (element *)pc);
		}
	}
}

static inline void ChengeDraw(HDC hDC, element *pe, int no)
{
	if (pe->IsGroup()) {
		ChengeDraw(hDC, ((group *)pe)->GetCont(), no);
	} else if (gPcb->IsFDisp(pe)) {
		Draw(hDC, pe);
	}
}

static int __fastcall GetEndPinNo(group *pg)
{
	int no = 0;
	const node &bas = pg->GetCont();
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			int n  = GetEndPinNo((group *)pc);
			if (n > no) {
				no = n;
			}
		} else if (pc->IsElem() && ((element *)pc)->GetPinNo() > no) {
			no = ((element *)pc)->GetPinNo();
		}
	}
	return no;
}

static int __fastcall CheckPulralPin(group *pg, int no)
{
	if (pg == NULL || no == 0) {
		return 0;
	}
	no = abs(no);
	const node &bas = pg->GetCont();
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			if (CheckPulralPin((group *)pc, no)) {
				return 1;
			}
		} else if (abs(((element *)pc)->GetPinNo()) == no) {
			if (DispMessage(IDS_PLURAL_PIN, MB_YESNO) == IDYES) {
				return 0;
			} else {
				return 1;
			}
		}
	}
	return 0;
}

static int __fastcall IsPartsPulral(group *pg, const unsigned char *unit)
{
	const node &bas = gPcb->GetElems();
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup() && pc != pg
				&& *((group *)pc)->GetUnit() && strcmp((char *)unit, ((group *)pc)->GetUnit()) == 0) {
			return 1;
		}
	}
	return 0;
}

static void DispPinNo(element *el, int no)
{
	if (no == 0) {
		return;
	}
	HDC hDC = GetScrHandle()->GetScrDC();
	SetTextAlign(hDC, TA_TOP | TA_LEFT | TA_NOUPDATECP);
	SetBkMode(hDC, TRANSPARENT);
	SetTextColor(hDC, RGB(255, 255, 255));
	SIZE sz;
	GetTextExtentPoint32(hDC, "A", 1, &sz);
	double x, y;
	if (el->IsArc()) {
		double ang = (((arc *)el)->GetSt() + ((arc *)el)->GetEnd()) / 2.0;
		Pos2NoSnap(pos(el->GetPos().x + ((arc *)el)->GetR() * cos(ang),
					   el->GetPos().y + ((arc *)el)->GetR() * sin(ang)), x, y);
	} else if (el->IsLine()) {
		Pos2NoSnap(pos((el->GetPos().x + ((line *)el)->GetPe().x) / 2,
					   (el->GetPos().y + ((line *)el)->GetPe().y) / 2), x, y);
	} else {
		Pos2NoSnap(el->GetPos(), x, y);
	}
	char str[16];
	*(int *)str = '0';
	itoa(abs(no), str, 10);
	TextOut(hDC, (int)x - sz.cx / 2, (int)y - sz.cy / 2, str, strlen(str));
}

char *__fastcall MakeTmpName(const char *tmp)
{
	static char name[64];
	strncpy(name, tmp, 50);
	name[50] = 0;
	char *ptr = name + strlen(name);
	SYSTEMTIME tim;
	GetLocalTime(&tim);
	sprintf(ptr, "%02d%02d%02d%02d%02d%02d", tim.wYear % 100,
			tim.wMonth % 100, tim.wDay % 100, tim.wHour % 100, tim.wMinute % 100, tim.wSecond % 100);
	return name;
}

static element* __fastcall SetDotLine(int ly, int ap, int s, const pos& pb, const pos& pe, dotLine_t& dot)
{
	double x = pb.x;
	double y = pb.y;
	double dx = pe.x - x;
	double dy = pe.y - y;
	if (fabs(dx) < ERR_MINI && fabs(dy) < ERR_MINI) {
		return land::New(ly, ap, s, pb);
	}
	double l = sqrt(dx * dx + dy * dy);
	double dxx = dx / l * dot.dl;
	double dyy = dy / l * dot.dl;
	double dl = dot.dl + dot.di;
	dx = dx / l * dl;
	dy = dy / l * dl;
	double l0 = 0.0;
	group *pg = group::New(gDottedLineIndex);
	char buff[32];
	sprintf(buff, "L%.12g", l);
	pg->SetAttrib(buff);
	l -= ERR_MINI;
	do {
		if (l0 + dot.dl > l) {
			if (sqrt(pe.Distance2(pos(x, y))) < ERR_MINI) {
				pg->AddGroup(land::New(ly, ap, s, pb));
			} else {
				pg->AddGroup(line::New(ly, ap, s, pos(x, y), pe));
			}
			break;
		}
		pg->AddGroup(line::New(ly, ap, s, pos(x, y), pos(x + dxx, y + dyy)));
		x += dx;
		y += dy;
	} while ((l0 += dl) < l);
	return (element*)pg;
}

static void FnChengeLu()
{
	pos p;
	GetMouseHitPos(p);
	element *el;
	if ((el = gPcb->IsHit(p)) != NULL) {
		double hi, wi;
		int ap;
		gCHENGE.ly = el->GetLy();
		gCHENGE.ap = el->GetAp();
		*(int *)gCHENGE.name = 0;
		int mod;
		switch (mod = GetChengeMode(&el, GetScrHandle()->GetMouseKey())) {
		case 0:
			if (el->IsRect()) {
				gCHENGE.f = gCHENGE.B_RCT;
				gCHENGE.x = ((rect *)el)->GetDx();
				gCHENGE.y = ((rect *)el)->GetDy();
				gCHENGE.ang = ((rect *)el)->GetAng() * 180.0 / M_PI;
			} else if (el->IsArc()) {
				gCHENGE.f = gCHENGE.B_ARC;
				gCHENGE.x = ((arc *)el)->GetPos().x;
				gCHENGE.y = ((arc *)el)->GetPos().y;
				gCHENGE.st = ((arc *)el)->GetSt() * 180.0 / M_PI;
				gCHENGE.end = ((arc *)el)->GetEnd() * 180.0 / M_PI;
				gCHENGE.r = ((arc *)el)->GetR();
				if (gCHENGE.end > 360.0) {
					gCHENGE.end -= 360.0;
					gCHENGE.st -= 360.0;
				}
			} else if (el->IsLand() || el->IsLine()) {
				gCHENGE.f = el->IsSq() ? gCHENGE.B_SQ : 0;
			} else {
				gCHENGE.f = 0;
			}
			break;
		case 1:
			gCHENGE.f = gCHENGE.B_TEXT;
			ap = gCHENGE.text.ap;
			hi = gCHENGE.text.hi;
			wi = gCHENGE.text.wi;
			gCHENGE.text.ap = gCHENGE.ap;
			strncpy(gCHENGE.text.buff, ((group *)el)->GetAttrib(), MAX_TEXTIN);
			gCHENGE.text.buff[MAX_TEXTIN] = 0;
			gCHENGE.text.wi = fabs(((group::ext_text *)((group *)el)->GetExt())->GetWi());
			gCHENGE.text.hi = ((group::ext_text *)((group *)el)->GetExt())->GetHi();
			gCHENGE.text.ang = ((group::ext_text *)((group *)el)->GetExt())->GetAng() * 180.0 / M_PI;
			gCHENGE.text.f = ((group::ext_text *)((group *)el)->GetExt())->GetWi() < 0;
			break;
		case 2:
			gCHENGE.f = gCHENGE.B_PAINT;
			break;
		case 3:
			//			if (((group*)el)->IsExtParts() && *((group*)el)->GetName()) {
			gCHENGE.f = gCHENGE.B_PARTS;
			strcpy((char *)gCHENGE.name, ((group *)el)->GetUnit());
			*(int *)gCHENGE.text.buff = 0;
			//			} else {
			//				MessageBeep(MB_OK);
			//				return;
			//			}
			break;
		case 4:
			if (!gPcb->IsFPtn(el)) {
				MessageBeep(MB_OK);
				return;
			}
			gCHENGE.f = gCHENGE.B_PIN;
			if ((gCHENGE.pin = el->GetPinNo()) == 0) {
				gCHENGE.pin = GetEndPinNo((group *)el->GetTopOwner()) + 1;
				gCHENGE.f |= gCHENGE.B_MODIFY;
			}
			if (el->GetPin()) {
				strcpy((char *)gCHENGE.name, el->GetPin());
			}
			break;
		case 5:
			if (!el->IsLine()) {
				MessageBeep(MB_OK);
				return;
			}
			gCHENGE.f = gCHENGE.B_DOT;
			break;
		default:
			MessageBeep(MB_OK);
			return;
		}
		gEditCross = el->Cut();
		HDC hDC = GetScrHandle()->GetScrDC();
		SelectObject(hDC, gSelePen);
		SetROP2(hDC, R2_COPYPEN);
		ChengeDraw(hDC, el, -1);
		FlushScr();
		if (FnChengeLuDlg(gCHENGE) == 0 && (gCHENGE.f & gCHENGE.B_MODIFY)) {
			SetROP2(hDC, R2_BLACK);
			ChengeDraw(hDC, el, -1);
			undone *und = undone::New(el, el->Ref());
			pos pb;
			switch (mod) {
			case 0:
				if (el->IsRect()) {
					el = el->Copy();
					el->ChengeLy(gCHENGE.ly);
					el->ChengeAp(gCHENGE.ap);
					((rect *)el)->ChengeSize(gCHENGE.x, gCHENGE.y, gCHENGE.ang * M_PI / 180.0);
				} else if (el->IsArc()) {
					if (GetCirAp(gCHENGE.ap) + ERR_MINI < gCHENGE.r) {
						el = arc::New(gCHENGE.ly, gCHENGE.ap, pos(gCHENGE.x, gCHENGE.y),
									  gCHENGE.r, gCHENGE.st * M_PI / 180.0, gCHENGE.end * M_PI / 180.0);
						el->CopyOwner(gEditCross);
					} else {
						goto noChenge;
					}
				} else if (gCHENGE.ly != el->GetLy() || gCHENGE.ap != el->GetAp()) {
					el = el->Copy();
					el->ChengeLy(gCHENGE.ly);
					el->ChengeAp(gCHENGE.ap);
				}
				break;
			case 1:
				el = (element *)CreateText(((group *)el)->GetExt()->GetPb(), gCHENGE.text, gCHENGE.ly, gCHENGE.ap);
				el->CopyOwner(gEditCross);
				break;
			case 2:
				ChengePaintLayer(((group *)(el = el->Copy()))->GetCont(), gCHENGE.ly);
				break;
			case 3:
				SetROP2(hDC, R2_COPYPEN);
				SelectObject(hDC, gPcb->GetPen(gCHENGE.ly));
				ChengeDraw(hDC, el, gCHENGE.ly);
				if (*gCHENGE.name && IsPartsPulral((group *)el, gCHENGE.name)
						&& DispMessage(IDS_IS_PARTS_PULRAL, MB_YESNO) == IDNO) {
					goto noChenge;
				}
				((group *)(el = el->Copy()))->SetUnit((char *)gCHENGE.name);
				if (((group *)el)->GetExt()) {
					pb = ((group *)el)->GetExt()->GetPb();
				} else {
					pb = GetGroupOrigPos(((group *)el)->GetCont());
					if (!*((group *)el)->GetName()) {
						((group *)el)->SetExt(group::ext_parts::New(GetGroupOrigPos(((group *)el)->GetCont())));
						((group *)el)->SetName(MakeTmpName(gTmpPartsNameBase));
					} else if (*((group *)el)->GetName() != ' ') {
						((group *)el)->SetExt(group::ext_parts::New(GetGroupOrigPos(((group *)el)->GetCont())));
					}
				}
				if (gCHENGE.f & gCHENGE.B_SWAPLY) {
					SwapParts(((group *)el)->GetCont(), pb);
				}
				if (gCHENGE.f & gCHENGE.B_TEXT) {
					((group *)el)->AddGroup(CreateText(pb, gCHENGE.text, gEditLayer, gEditText));
				}
				break;
			case 4:
				SetROP2(hDC, R2_COPYPEN);
				Draw(hDC, el);
				if (CheckPulralPin((group *)el->GetTopOwner(), gCHENGE.pin)
						|| (gCHENGE.pin == el->GetPinNo()
							&& (el->GetPin() ? strcmp((char *)gCHENGE.name, el->GetPin()) == 0 : *gCHENGE.name == 0))) {
					goto noChenge;
				}
				(el = el->Copy())->SetPinNo(gCHENGE.pin);
				el->SetPin((char *)gCHENGE.name);
				DispPinNo(el, gCHENGE.pin);
				break;
			case 5:
				el = SetDotLine(el->GetLy(), el->GetAp(), el->IsSq(), el->GetPos(), ((line*)el)->GetPe(), gCHENGE.dot);
				break;
			}
			gEditCross->Ref()->Add(el);
			und->AddData(el);
			AddOwnUndone(und);
			SetFFileChenge();
		} else {
noChenge:
			mod = 0;
			el->Ref()->Add(el);
		}
		if (mod == 1) {
			gCHENGE.text.ap = ap;
			gCHENGE.text.hi = hi;
			gCHENGE.text.wi = wi;
		}
		if (mod != 4) {
			SetROP2(hDC, R2_COPYPEN);
			SelectObject(hDC, gPcb->GetPen(gCHENGE.ly));
			ChengeDraw(hDC, el, gCHENGE.ly);
		}
		gEditCross = NULL;
	} else {
		MessageBeep(MB_OK);
	}
}

static void FnChengeKey()
{
	if (gEditKey == EDKEY_LY) {
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 0)) {
			GetScrHandle()->Invalidate();
		}
		SetLyApMenu();
	} else if (gEditKey == EDKEY_SEL) {
		CallDialogBox(gApp->Handle(), IDD_SELCHENGE, SeleChengeDlg);
	}
}

static void FnChengeEsc()
{
	FnSetEsc();
	ResetStack(MOD_CHENG, IDST_CHENGE);
	gApp->PushButton(IDB_CHENGE, TRUE);
	gApp->EnableButton(IDB_VALUEIN, FALSE);
	gApp->EnableButton(IDB_CROSSIN, FALSE);
	gApp->UpdateButton();
	gEditMode = EDMOD_NON;
	SetLyApMenu();
}

static inline double GetPaintCrossLine(paint_t *p, const double &y)
{
	if (p->flag & paint_t::B_LINEV) {
		return p->pl.x;
	}
	return (y - p->pl.y) / p->li.a + p->pl.x;
}

static int __fastcall IsPaintAngInAng(double a, paint_t *p)
{
	while (a < p->ar.st) {
		a += M_PI * 2.0;
	}
	return p->ar.st <= a && p->ar.en >= a;
}

static int __fastcall GetPaintCrossArc(paint_t *p, double *px, const double &y)
{
	double tm = y - p->ar.pb.y;
	if (fabs(tm) >= p->ar.r - AP_SIZE_MINI / 32.0) {
		return 0;
	}
	double x = p->ar.r * p->ar.r - tm * tm;
	if (x < ERR_MINI) {
		x = 0.0;
	} else {
		x = sqrt(x);
	}
	double ang = atan2(tm, x);
	int n = 0;
	if (IsPaintAngInAng(ang, p)) {
		*px = p->ar.pb.x + x;
		++px;
		++n;
	}
	ang = atan2(tm, -x);
	if (IsPaintAngInAng(ang, p)) {
		*px = p->ar.pb.x - x;
		++n;
	}
	return n;
}

static int __fastcall Insert(double *tab, double &x, int max)
{
	int m = 0, lo = 0, hi = max;
	while (lo < hi) {
		m = (lo + hi) / 2;
		if (tab[m] < x) {
			lo = ++m;
		} else {
			hi = m;
		}
	}
	if (m < max) {
		memmove(tab + m + 1, tab + m, (max - m) * sizeof(x));
		tab[m] = x;
	} else {
		tab[max] = x;
	}
	return max + 1;
}

#define MAX_CROSS_POINT	0x2000

static int __fastcall GoPaint(void *lim, node &bas, int ly, int ap)
{
	paint_t *mem;
	if ((mem = (paint_t *)ResetMem()) >= lim) {
		return 1;
	}
	node get(0);
	double maxY = 0.0;
	double minY = 1e200;
	for (; mem < lim; ++mem) {
		double y = mem->ph.y;
		if (maxY < y) {
			maxY = y;
		}
		if (mem->pl.y < minY) {
			minY = mem->pl.y;
		}
		paint_t *ptr;
		for (ptr = (paint_t *)get.Ref(); ptr != &get && ptr->ph.y < y; ptr = ptr->Ref())
			;
		ptr->Add(mem);
	}
	double dy = floor(GetCirDiam(ap) * PAINT_OVER / STEP_MINI) * STEP_MINI;
	if (dy < 0.001) {
		dy = 0.001;
	}
	double y = minY - AdjustStep(GetCirAp(ap) * PAINT_OVER / 2.0);
	maxY -= AdjustStep(GetCirAp(ap) * (1.0 - PAINT_OVER) / 2.0);
	while ((y += dy) <= maxY) {
		double cross[MAX_CROSS_POINT];
		int max = 0;
		while (((paint_t *)get.Ref())->ph.y < y) {
			get.Ref()->Cut();
		}
		for (paint_t *ptr = (paint_t *)get.Ref(); ptr != &get; ptr = ptr->Ref()) {
			if (ptr->pl.y <= y) {
				if (ptr->flag & paint_t::B_LINE) {
					double x = GetPaintCrossLine(ptr, y);
					max = Insert(cross, x, max);
				} else {
					double tmp[2];
					int n;
					if ((n = GetPaintCrossArc(ptr, tmp,  y)) != 0) {
						max = Insert(cross, tmp[0], max);
						if (n > 1 && max < MAX_CROSS_POINT) {
							max = Insert(cross, tmp[1], max);
						}
					}
				}
				if (max >= MAX_CROSS_POINT) {
					return 1;
				}
			}
		}
		if (max & 1) {
			return 1;
		}
		for (int i = 0; i < max; i += 2) {
			bas.Add(line::New(ly, ap, 0, pos(cross[i], y), pos(cross[i + 1], y)));
		}
	}
	return 0;
}

static paint_t *__fastcall SetPaintMem(paint_t *ptr, const node &bas)
{
	for (element *el = (element *)bas.Ref(); el != &bas; el = el->Ref()) {
		if (IsTempMemOver(ptr)) {
			return NULL;
		}
		if (el->IsLine()) {
			if (el->GetPos().y < ((line *)el)->GetPe().y) {
				ptr->pl = el->GetPos();
				ptr->ph = ((line *)el)->GetPe();
			} else {
				ptr->pl = ((line *)el)->GetPe();
				ptr->ph = el->GetPos();
			}
			ptr->li.dy = ptr->ph.y - ptr->pl.y;
			if (fabs(ptr->li.dy) >= ERR_MINI) {
				ptr->li.dx = ptr->ph.x - ptr->pl.x;
				if (fabs(ptr->li.dx) < ERR_MINI) {
					ptr->flag = paint_t::B_LINE | paint_t::B_LINEV;
				} else {
					ptr->flag = paint_t::B_LINE;
					ptr->li.a = ptr->li.dy / ptr->li.dx;
				}
				++ptr;
			}
		} else if (el->IsArc()) {
			ptr->flag = paint_t::B_ARC;
			ptr->ar.r = ((arc *)el)->GetR();
			ptr->pl = el->GetLo();
			ptr->ph = el->GetHi();
			ptr->ar.pb = *(posXY *)&el->GetPos();
			if (((arc *)el)->GetSt() < ((arc *)el)->GetEnd()) {
				ptr->ar.st = ((arc *)el)->GetSt();
				ptr->ar.en = ((arc *)el)->GetEnd();
			} else {
				ptr->ar.st = ((arc *)el)->GetEnd();
				ptr->ar.en = ((arc *)el)->GetSt();
			}
			while (ptr->ar.en > M_PI) {
				ptr->ar.st -= M_PI * 2.0;
				ptr->ar.en -= M_PI * 2.0;
			}
			while (ptr->ar.st < 0.0) {
				ptr->ar.st += M_PI * 2.0;
				ptr->ar.en += M_PI * 2.0;
			}
			++ptr;
		}
	}
	return ptr;
}

static element *__fastcall GetHitPaintElem(const node &bas, const pos &p, int ly, int f)
{
	for (element *el = (element *)bas.Ref(); el != &bas; el = el->Ref()) {
		if (el->IsGroup()) {
			element *elem;
			if ((elem = GetHitPaintElem(((group *)el)->GetCont(), p, ly, f)) != NULL) {
				return elem;
			}
		} else if (el->GetLy() == ly && ((el->IsLine() && ! el->IsSq()) || el->IsArc()) && el->IsEqPos(p)) {
			if (f || el->GetOwner() == NULL) {
				return el;
			}
		}
	}
	return NULL;
}

static int __fastcall GetPaintLoopFow(undone *und, HDC hDC, int f = 0)
{
	for (;;) {
		if (gSeleP1.IsEq(gSeleP2)) {
			return 0;
		}
		element *el;
		if ((el = GetHitPaintElem(gPcb->GetElems(), gSeleP1, ((element *)gEditSeles.Prev())->GetLy(), f)) == NULL) {
			return 1;
		}
		el->Cut();
		if (GetHitPaintElem(gPcb->GetElems(), gSeleP1, ((element *)gEditSeles.Prev())->GetLy(), f)) {
			el->Ref()->Add(el);
			return -1;
		}
		Draw(hDC, el);
		und->AddData(el, el->Ref());
		gEditSeles.Add(el);
		gSeleP1 = el->GetFarPos(gSeleP1);
	}
}

static int __fastcall GetPaintLoopRev(undone *und, HDC hDC, int f = 0)
{
	for (;;) {
		if (gSeleP1.IsEq(gSeleP2)) {
			return 0;
		}
		element *el;
		if ((el = GetHitPaintElem(gPcb->GetElems(), gSeleP2, ((element *)gEditSeles.Prev())->GetLy(), f)) == NULL) {
			return 1;
		}
		el->Cut();
		if (GetHitPaintElem(gPcb->GetElems(), gSeleP2, ((element *)gEditSeles.Prev())->GetLy(), f)) {
			el->Ref()->Add(el);
			return -1;
		}
		Draw(hDC, el);
		und->AddData(el, el->Ref());
		gEditSeles.Add(el);
		gSeleP2 = el->GetFarPos(gSeleP2);
	}
}

static int PaintPaint()
{
	paint_t *ptr;
	int rt = 1;
	if ((ptr = SetPaintMem((paint_t *)ResetMem(), gEditSeles)) != NULL) {
		int ap = ((element *)gEditSeles.Ref())->GetAp();
		double apl = GetCirAp(ap);
		for (element *el = (element *)gEditSeles.Ref()->Ref(); el != &gEditSeles; el = el->Ref()) {
			if (apl > GetCirAp(el->GetAp())) {
				apl = GetCirAp(ap = el->GetAp());
			}
		}
		node lines(0);
		rt = GoPaint(ptr, lines, ((element *)gEditSeles.Ref())->GetLy(), ap);
		if (!lines.IsNull()) {
			undone *und = undone::New();
			while (!gTmpUndone.IsNull()) {
				und->GetAllData((undone *)gTmpUndone.Ref()->Cut())->Del();
			}
			group *loop = group::New(gPaintedLoop);
			while (!gEditSeles.IsNull()) {
				loop->AddGroup(((element *)gEditSeles.Ref())->Cut()->Copy());
			}
			gEditSeles.Init();
			group *lis = group::New(gPaintedLines);
			while (!lines.IsNull()) {
				lis->AddGroup((line *)lines.Ref()->Cut());
			}
			group *pain = group::New(gPaintedIndex);
			pain->AddGroup(lis);
			pain->AddGroup(loop);
			HDC hDC = GetScrHandle()->GetScrDC();
			SetROP2(hDC, R2_COPYPEN);
			DrawGrp(hDC, pain->GetCont());
			FlushScr();
			gPcb->Add(pain);
			und->AddData(pain);
			AddOwnUndone(und);
			SetFFileChenge();
		} else {
			while (!gTmpUndone.IsNull()) {
				((undone *)gTmpUndone.Prev())->Undone()->Del();
			}
		}
	} else {
		while (!gTmpUndone.IsNull()) {
			((undone *)gTmpUndone.Prev())->Undone()->Del();
		}
	}
	return rt;
}

static element *__fastcall CheckHitPaint(container *el)
{
	el = el->GetTopOwner();
	if (!el->IsGroup() || ((group *)el)->GetName() != gPaintedIndex) {
		return NULL;
	}
	gEditCross = ((element *)el)->Cut();
	HDC hDC = GetScrHandle()->GetScrDC();
	SelectObject(hDC, gSelePen);
	SetROP2(hDC, R2_COPYPEN);
	DrawEditCross(hDC, el);
	FlushScr();
	return (element *)el;
}

static int __fastcall IsPaintLines(group *pg)
{
	const node &bas = pg->GetCont();
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup() && ((group *)pc)->GetName() == gPaintedLines) {
			return 1;
		}
	}
	pg->Ref()->Add(pg);
	gEditCross = NULL;
	return 0;
}

static void FnPaintLu()
{
	pos p;
	GetMouseHitPos(p);
	element *el = gPcb->IsHit(p);
	if (el == NULL) {
		MessageBeep(MB_OK);
		return;
	}
	if (IsFPaintBack() || (GetScrHandle()->GetMouseKey() & MK_SHIFT)) {
		if ((el = CheckHitPaint(el)) != NULL && IsPaintLines((group *)el)) {
			HDC hDC = GetScrHandle()->GetScrDC();
			SetROP2(hDC, R2_BLACK);
			if (DispMessage(IDS_PAINT_CUT, MB_YESNO) == IDYES) {
				const node *bas = &((group *)el)->GetCont();
				element *loop;
				for (loop = (element *)bas->Ref(); loop != bas; loop = loop->Ref()) {
					if (loop->IsGroup() && ((group *)loop)->GetName() == gPaintedLoop) {
						break;
					}
				}
				if (loop == bas || ((group *)loop)->IsNullGroup()) {
					SetROP2(hDC, R2_COPYPEN);
					DrawEditCrossNorm(hDC, gEditCross);
					gEditCross->Ref()->Add(gEditCross);
					MessageBeep(MB_OK);
				} else {
					undone *und = undone::New();
					und->AddData(el, el->Ref());
					bas = &((group *)loop)->GetCont();
					SelectObject(hDC, gPcb->GetPen((element *)bas->Ref()));
					for (el = (element *)bas->Ref(); el != bas; el = el->Ref()) {
						element *tm = el->Copy();
						tm->ClearOwner();
						gPcb->Add(tm);
						und->AddData(tm);
					}
					AddOwnUndone(und);
					SetFFileChenge();
					GetScrHandle()->Invalidate();
				}
			} else {
				DrawEditCross(hDC, gEditCross);
				SetROP2(hDC, R2_COPYPEN);
				DrawEditCrossNorm(hDC, gEditCross);
				gEditCross->Ref()->Add(gEditCross);
			}
			gEditCross = NULL;
		}
	} else if (IsFPaintMado() || (GetScrHandle()->GetMouseKey() & MK_CONTROL)) {
		if ((el = CheckHitPaint(el)) != NULL && IsPaintLines((group *)el)) {
			undone *und = undone::New();
			und->AddData(el, el->Ref());
			SetOwnTmpUndone(und);
			PushStack(2, IDST_PAINT2);
			SetButtonOnState();
			gApp->EnableButton(IDB_SELECT, FALSE);
			gApp->UpdateButton();
		}
	} else if (((el->IsLine() && ! el->IsSq()) || el->IsArc()) && el->GetOwner() == NULL) {
		HDC hDC = GetScrHandle()->GetScrDC();
		SelectObject(hDC, gSelePen);
		SetROP2(hDC, R2_COPYPEN);
		Draw(hDC, el);
		undone *und = undone::New();
		und->AddData(el, el->Ref());
		gEditSeles.Add(el->Cut());
		if (el->IsLine()) {
			gSeleP1 = el->GetPos();
			gSeleP2 = ((line *)el)->GetPe();
		} else {
			gSeleP1 = ((arc *)el)->GetStPos();
			gSeleP2 = ((arc *)el)->GetEndPos();
		}
		int rt;
		if ((rt = GetPaintLoopFow(und, hDC)) != 0) {
			rt = GetPaintLoopRev(und, hDC);
		}
		SetOwnTmpUndone(und);
		if (rt == 0) {
			if (DispMessage(IDS_PAINT_OK, MB_YESNO) == IDYES) {
				PaintPaint();
			} else {
				FnPaintEsc1();
			}
		} else {
			PushStack(1, IDST_PARA1);
			SetButtonOnState();
			gApp->EnableButton(IDB_SELECT, FALSE);
			gApp->EnableButton(IDB_LAYER, FALSE);
			EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_GRAYED);
			EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_AP, MF_GRAYED);
			DrawMenuBar(GetAppWnd());
			gApp->UpdateButton();
		}
	} else {
		MessageBeep(MB_OK);
	}
}

static void FnPaintKey()
{
	if (gEditKey == EDKEY_LY) {
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 0)) {
			GetScrHandle()->Invalidate();
		}
		SetLyApMenu();
	} else if (gEditKey == EDKEY_SEL) {
		CallDialogBox(gApp->Handle(), IDD_SELPAINT, PaintSeleDlg);
	}
}

static void FnPaintEsc()
{
	FnSetEsc();
	ResetStack(MOD_PAINT, IDST_PAINT);
	gApp->PushButton(IDB_PAINT, TRUE);
	gApp->EnableButton(IDB_VALUEIN, FALSE);
	gApp->EnableButton(IDB_CROSSIN, FALSE);
	gApp->UpdateButton();
	gEditMode = EDMOD_NON;
	SetLyApMenu();
}

static void FnPaintLu1()
{
	pos p;
	GetMouseHitPos(p);
	element *el;
	if ((el = gPcb->IsHit(p)) != NULL && ((element *)gEditSeles.Ref())->GetLy() == el->GetLy()
			&& ((el->IsLine() && ! el->IsSq()) || el->IsArc()) && el->GetOwner() == NULL) {
		int rt = 1;
		undone *und = undone::New();
		HDC hDC = GetScrHandle()->GetScrDC();
		SelectObject(hDC, gSelePen);
		SetROP2(hDC, R2_COPYPEN);
		if (el->IsEqPos(gSeleP1)) {
			Draw(hDC, el);
			und->AddData(el, el->Ref());
			gEditSeles.Add(el->Cut());
			gSeleP1 = el->GetFarPos(gSeleP1);
			rt = GetPaintLoopFow(und, hDC);
		} else if (el->IsEqPos(gSeleP2)) {
			Draw(hDC, el);
			und->AddData(el, el->Ref());
			gEditSeles.Add(el->Cut());
			gSeleP2 = el->GetFarPos(gSeleP2);
			rt = GetPaintLoopRev(und, hDC);
		}
		if (und->GetDataBase().IsNull()) {
			und->Del();
		} else {
			SetOwnTmpUndone(und);
		}
		if (rt == 0) {
			if (DispMessage(IDS_PAINT_OK, MB_YESNO) == IDYES) {
				PaintPaint();
				PopStack();
			} else {
				FnPaintEsc1();
			}
		}
	} else {
		MessageBeep(MB_OK);
	}
}

static void __fastcall SetPaintP1P2(element *el, pos &p1, pos &p2)
{
	if (el->IsArc()) {
		p1 = ((arc *)el)->GetStPos();
		p2 = ((arc *)el)->GetEndPos();
	} else {
		p1 = el->GetPos();
		p2 = ((line *)el)->GetPe();
	}
}

static void FnPaintUnd1()
{
	DrawOwnUndone((undone *)gTmpUndone.Prev());
	((undone *)gTmpUndone.Prev())->Undone()->Del();
	if (gTmpUndone.IsNull()) {
		gEditCross = NULL;
		PopStack();
	} else {
		pos p1, p2;
		SetPaintP1P2((element *)gEditSeles.Ref(), p1, p2);
		if (gEditSeles.Ref()->Ref() == &gEditSeles) {
			gSeleP1 = p1;
			gSeleP2 = p2;
		} else {
			gSeleP1 = ((element *)gEditSeles.Ref())->Ref()->IsEqPos(p1) ? p2 : p1;
			SetPaintP1P2((element *)gEditSeles.Prev(), p1, p2);
			gSeleP2 = ((element *)gEditSeles.Prev())->Prev()->IsEqPos(p1) ? p2 : p1;
		}
	}
}

static void FnPaintEsc1()
{
	DrawTmpUndoneElem();
	while (! gTmpUndone.IsNull()) {
		((undone *)gTmpUndone.Prev())->Undone()->Del();
	}
	gEditCross = NULL;
	PopStack();
}

static void FnPaintKey2()
{
	if (gEditKey == EDKEY_LY) {
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 0)) {
			GetScrHandle()->Invalidate();
		}
		SetLyApMenu();
	}
}

static void FnPaintUnd2()
{
	DrawTmpUndoneElem();
	while (! gTmpUndone.IsNull()) {
		((undone *)gTmpUndone.Prev())->Undone()->Del();
	}
	gEditCross = NULL;
	PopStack();
}

static paint_t *__fastcall SetPaintMadoLoop(int &ly, int &ap)
{
	ly = -1;
	paint_t *ptr;
	if ((ptr = SetPaintMem((paint_t *)ResetMem(), gEditSeles)) == NULL) {
		return NULL;
	}
	const node &base = ((group *)gEditCross)->GetCont();
	for (container *pc = (container *)base.Ref(); pc != &base; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			if (((group *)pc)->GetName() == gPaintedMado || ((group *)pc)->GetName() == gPaintedLoop) {
				if ((ptr = SetPaintMem(ptr, ((group *)pc)->GetCont())) == NULL) {
					return NULL;
				}
			} else if (ly < 0 && ((group *)pc)->GetName() == gPaintedLines) {
				ly = ((element *)((group *)pc)->GetCont().Ref())->GetLy();
				ap = ((element *)((group *)pc)->GetCont().Ref())->GetAp();
			}
		} else {
			return NULL;
		}
	}
	if (ly < 0) {
		return NULL;
	}
	double ap0 = GetCirAp(ap);
	for (container *pc = (container *)gEditSeles.Ref(); pc != &gEditSeles; pc = pc->Ref()) {
		if (pc->IsElem() && GetCirAp(((element *)pc)->GetAp()) < ap0) {
			ap = ((element *)pc)->GetAp();
			ap0 = GetCirAp(ap);
		}
	}
	return ptr;
}

/* p  p1 p2 ̊Ԃɂ邩 */
static inline int IsPointInLine(const double p, const double p1, const double p2)
{
	return (p - p1) * (p - p2) <= 0.0;
}

static int __fastcall IsCrossLineLine(line *li1, line *li2)
{
	double dx1 = li1->GetPe().x - li1->GetPos().x;
	double dy1 = li1->GetPe().y - li1->GetPos().y;
	double dx2 = li2->GetPe().x - li2->GetPos().x;
	double dy2 = li2->GetPe().y - li2->GetPos().y;
	pos px;
	if (fabs(dx1) < ERR_MINI) {
		if (fabs(dx2) < ERR_MINI)
			return fabs(li1->GetPos().x - li2->GetPos().x) < ERR_MINI
				   && (IsPointInLine(li1->GetPos().y, li2->GetPos().y, li2->GetPe().y)
					   || IsPointInLine(li1->GetPe().y, li2->GetPos().y, li2->GetPe().y)
					   || IsPointInLine(li2->GetPos().y, li1->GetPos().y, li1->GetPe().y)
					   || IsPointInLine(li2->GetPe().y, li1->GetPos().y, li1->GetPe().y));
		if (fabs(dy2) < ERR_MINI) {
			return IsPointInLine(li1->GetPos().x, li2->GetPos().x, li2->GetPe().x)
				   && IsPointInLine(li2->GetPos().y, li1->GetPos().y, li1->GetPe().y);
		}
		px.Set(li1->GetPos().x, dy2 / dx2 * (li1->GetPos().x - li2->GetPos().x) + li2->GetPos().y);
		return IsPointInLine(px.y, li1->GetPos().y, li1->GetPe().y)
			   && IsPointInLine(px.x, li2->GetPos().x, li2->GetPe().x)
			   && IsPointInLine(px.y, li2->GetPos().y, li2->GetPe().y);
	} else if (fabs(dx2) < ERR_MINI) {
		if (fabs(dy1) < ERR_MINI) {
			return IsPointInLine(li2->GetPos().x, li1->GetPos().x, li1->GetPe().x)
				   && IsPointInLine(li1->GetPos().y, li2->GetPos().y, li2->GetPe().y);
		}
		px.Set(li2->GetPos().x, dy1 / dx1 * (li2->GetPos().x - li1->GetPos().x) + li1->GetPos().y);
		return IsPointInLine(px.y, li2->GetPos().y, li2->GetPe().y)
			   && IsPointInLine(px.x, li1->GetPos().x, li1->GetPe().x)
			   && IsPointInLine(px.y, li1->GetPos().y, li1->GetPe().y);
	} else if (fabs(dy1) < ERR_MINI) {
		if (fabs(dy2) < ERR_MINI)
			return fabs(li1->GetPos().y - li2->GetPos().y) < ERR_MINI
				   && (IsPointInLine(li1->GetPos().x, li2->GetPos().x, li2->GetPe().x)
					   || IsPointInLine(li1->GetPe().x, li2->GetPos().x, li2->GetPe().x)
					   || IsPointInLine(li2->GetPos().x, li1->GetPos().x, li1->GetPe().x)
					   || IsPointInLine(li2->GetPe().x, li1->GetPos().x, li1->GetPe().x));
		px.Set((li1->GetPos().y - li2->GetPos().y) * dx2 / dy2 + li2->GetPos().x, li1->GetPos().y);
		return IsPointInLine(px.x, li1->GetPos().x, li1->GetPe().x)
			   && IsPointInLine(px.x, li2->GetPos().x, li2->GetPe().x)
			   && IsPointInLine(px.y, li2->GetPos().y, li2->GetPe().y);
	} else if (fabs(dy2) < ERR_MINI) {
		px.Set((li2->GetPos().y - li1->GetPos().y) * dx1 / dy1 + li1->GetPos().x, li2->GetPos().y);
		return IsPointInLine(px.x, li2->GetPos().x, li2->GetPe().x)
			   && IsPointInLine(px.x, li1->GetPos().x, li1->GetPe().x)
			   && IsPointInLine(px.y, li1->GetPos().y, li1->GetPe().y);
	}
	double a1 = dy1 / dx1;
	double a2 = dy2 / dx2;
	double b1 = li1->GetPos().y - a1 * li1->GetPos().x;
	double b2 = li2->GetPos().y - a2 * li2->GetPos().x;
	if (fabs(a1 - a2) < ERR_MINI) {
		if (fabs(b1 - b2) >= ERR_MINI) {
			return 0;
		}
		if (dx1 > dy1)
			return IsPointInLine(li2->GetPos().x, li1->GetPos().x, li1->GetPe().x)
				   || IsPointInLine(li2->GetPe().x, li1->GetPos().x, li1->GetPe().x)
				   || IsPointInLine(li1->GetPos().x, li2->GetPos().x, li2->GetPe().x)
				   || IsPointInLine(li1->GetPe().x, li2->GetPos().x, li2->GetPe().x);
		else
			return IsPointInLine(li2->GetPos().y, li1->GetPos().y, li1->GetPe().y)
				   || IsPointInLine(li2->GetPe().y, li1->GetPos().y, li1->GetPe().y)
				   || IsPointInLine(li1->GetPos().y, li2->GetPos().y, li2->GetPe().y)
				   || IsPointInLine(li1->GetPe().y, li2->GetPos().y, li2->GetPe().y);
	}
	px.x = (b2 - b1) / (a1 - a2);
	px.y = a1 * px.x + b1;
	return IsPointInLine(px.x, li1->GetPos().x, li1->GetPe().x)
		   && IsPointInLine(px.y, li1->GetPos().y, li1->GetPe().y)
		   && IsPointInLine(px.x, li2->GetPos().x, li2->GetPe().x)
		   && IsPointInLine(px.y, li2->GetPos().y, li2->GetPe().y);
}

static int inline IsInAng(const double &lo, const double &hi, double ang)
{
	return arc::IsInAng(ang, lo, hi);
}

static int __fastcall IsCrossLineArc(line *li, arc *ar)
{
	double dx = li->GetPe().x - li->GetPos().x;
	double r = ar->GetR();
	if (fabs(dx) < ERR_MINI) {
		if (fabs(li->GetPos().x - ar->GetPos().x) > r) {
			return 0;
		}
		dx = li->GetPos().x - ar->GetPos().x;
		double y = r * r - dx * dx;
		if (y < ERR_MINI) {
			y = 0.0;
		} else {
			y = sqrt(y);
		}
		if (IsPointInLine(ar->GetPos().y + y, li->GetPos().y, li->GetPe().y)
				&& IsInAng(ar->GetSt(), ar->GetEnd(), atan2(y, dx))) {
			return 1;
		}
		return IsPointInLine(ar->GetPos().y - y, li->GetPos().y, li->GetPe().y)
			   && IsInAng(ar->GetSt(), ar->GetEnd(), atan2(-y, dx));
	}
	double a = (li->GetPe().y - li->GetPos().y) / dx;
	if (fabs(a) < ERR_MINI) {
		if (fabs(li->GetPos().y - ar->GetPos().y) > r) {
			return 0;
		}
		dx = li->GetPos().y - ar->GetPos().y;
		double x = r * r - dx * dx;
		if (x < ERR_MINI) {
			x = 0.0;
		} else {
			x = sqrt(x);
		}
		if (IsPointInLine(ar->GetPos().x + x, li->GetPos().x, li->GetPe().x)
				&& IsInAng(ar->GetSt(), ar->GetEnd(), atan2(dx, x))) {
			return 1;
		}
		return IsPointInLine(ar->GetPos().x - x, li->GetPos().x, li->GetPe().x)
			   && IsInAng(ar->GetSt(), ar->GetEnd(), atan2(dx, -x));
	}
	double b = li->GetPos().y - ar->GetPos().y - a * (li->GetPos().x - ar->GetPos().x);
	double aa1 = a * a + 1;
	double tmp = aa1 * r * r - b * b;
	if (tmp < 0.0) {
		return 0;
	}
	if (tmp < ERR_MINI) {
		tmp = 0.0;
	} else {
		tmp = sqrt(tmp);
	}
	double x = (-a * b + tmp) / aa1;
	double y = a * x + b;
	if (IsInAng(ar->GetSt(), ar->GetEnd(), atan2(y, x))
			&& IsPointInLine(ar->GetPos().x + x, li->GetPos().x, li->GetPe().x)
			&& IsPointInLine(ar->GetPos().y + y, li->GetPos().y, li->GetPe().y)) {
		return 1;
	}
	x = (-a * b - tmp) / aa1;
	y = a * x + b;
	return IsInAng(ar->GetSt(), ar->GetEnd(), atan2(y, x))
		   && IsPointInLine(ar->GetPos().x + x, li->GetPos().x, li->GetPe().x)
		   && IsPointInLine(ar->GetPos().y + y, li->GetPos().y, li->GetPe().y);
}

static int __fastcall IsCrossArcArc(arc *ar1, arc *ar2)
{
	double dx = ar2->GetPos().x - ar1->GetPos().x;
	double dy = ar2->GetPos().y - ar1->GetPos().y;
	double ll = dx * dx + dy * dy;
	if (ll < ERR_MINI) {
		return 0;
	}
	double l = sqrt(ll);
	if (ar1->GetR() + ar2->GetR() <= l) {
		return 0;
	}
	double rr1 = ar1->GetR() * ar1->GetR();
	double dl = (ll + rr1 - ar2->GetR() * ar2->GetR()) / (2.0 * l);
	if ((ll = rr1 - dl * dl) < ERR_MINI) {
		return 0;
	}
	double h = sqrt(ll);
	double ang1 = atan2(dy, dx);
	double ang2 = atan2(h, dl);
	double ang3 = ang1 + M_PI;
	double ang4 = atan2(h, l - dl);
	return (IsInAng(ar1->GetSt(), ar1->GetEnd(), ang1 + ang2)  && IsInAng(ar2->GetSt(), ar2->GetEnd(), ang3 - ang4))
		   || (IsInAng(ar1->GetSt(), ar1->GetEnd(), ang1 - ang2) && IsInAng(ar2->GetSt(), ar2->GetEnd(), ang3 + ang4));
}

static int IsCrossLoopLine()
{
	const node &bas = ((group *)gEditCross)->GetCont();
	for (element *pe = (element *)bas.Ref(); pe != &bas; pe = pe->Ref()) {
		if (pe->IsGroup() && (((group *)pe)->GetName() == gPaintedMado || ((group *)pe)->GetName() == gPaintedLoop)) {
			const node &bl = ((group *)pe)->GetCont();
			for (element *loop = (element *)bl.Ref(); loop != &bl; loop = loop->Ref()) {
				for (element *el = (element *)gEditSeles.Ref(); el != &gEditSeles; el = el->Ref()) {
					if (loop->IsLine()) {
						if (el->IsLine()) {
							if (IsCrossLineLine((line *)loop, (line *)el)) {
								return 1;
							}
						} else if (IsCrossLineArc((line *)loop, (arc *)el)) {
							return 1;
						}
					} else {
						if (el->IsLine()) {
							if (IsCrossLineArc((line *)el, (arc *)loop)) {
								return 1;
							}
						} else if (IsCrossArcArc((arc *)loop, (arc *)el)) {
							return 1;
						}
					}
				}
			}
		}
	}
	return 0;
}

static int IsNotLoopInMado()
{
	const node &bas = ((group *)gEditCross)->GetCont();
	for (element *pe = (element *)bas.Ref(); pe != &bas; pe = pe->Ref()) {
		if (pe->IsGroup() && ((group *)pe)->GetName() == gPaintedLines) {
			const node &lb = ((group *)pe)->GetCont();
			for (line *pl = (line *)lb.Ref(); pl != &lb; pl = pl->Ref()) {
				for (element *el = (element *)gEditSeles.Ref(); el != &gEditSeles; el = el->Ref()) {
					if (el->IsLine()) {
						if (IsCrossLineLine(pl, (line *)el)) {
							return 0;
						}
					} else if (IsCrossLineArc(pl, (arc *)el)) {
						return 0;
					}
				}
			}
		}
	}
	return 1;
}

static int __fastcall PaintMado(HDC hDC)
{
	int rt = 0;
	if (DispMessage(IDS_PAINT_MADO, MB_YESNO) == IDYES) {
		int ly, ap;
		paint_t *ptr = SetPaintMadoLoop(ly, ap);
		if (ptr) {
			node base(0);
			rt = GoPaint(ptr, base, ly, ap);
			if (!base.IsNull()) {
				group *pain = group::New(gPaintedIndex);
				const node &nb = ((group *)gEditCross)->GetCont();
				for (container *pc = (container *)nb.Ref(); pc != &nb; pc = pc->Ref()) {
					if (pc->IsGroup() && (((group *)pc)->GetName() == gPaintedLoop || ((group *)pc)->GetName() == gPaintedMado)) {
						pain->AddGroup(pc->Copy());
					}
				}
				group *gtm = group::New(gPaintedMado);
				for (element *el = (element *)gEditSeles.Ref(); el != &gEditSeles; el = el->Ref()) {
					if (el->IsArc()) {
						gtm->AddGroup(arc::New(ly, ap, el->GetPos(), ((arc *)el)->GetR(), ((arc *)el)->GetSt(), ((arc *)el)->GetEnd()));
					} else {
						gtm->AddGroup(line::New(ly, ap, 0, el->GetPos(), ((line *)el)->GetPe()));
					}
				}
				while (gTmpUndone.Prev()->Prev() != &gTmpUndone) {
					((undone *)gTmpUndone.Pop()->Cut())->Undone()->Del();
				}
				pain->AddGroup(gtm);
				gtm = group::New(gPaintedLines);
				while (!base.IsNull()) {
					gtm->AddGroup((element *)base.Ref()->Cut());
				}
				pain->AddGroup(gtm);
				DrawEditSeleCrosSub(hDC, R2_BLACK);
				SetROP2(hDC, R2_COPYPEN);
				DrawEditSelesNorm(hDC, pain->GetCont());
				gPcb->Add(pain);
				undone *und = undone::New();
				und->GetAllData((undone *)gTmpUndone.Ref()->Cut())->Del();
				und->AddData(pain);
				AddOwnUndone(und);
				SetFFileChenge();
				gEditCross = NULL;
				return 0;
			}
		}
	}
	DrawEditSeleCrosSub(hDC, R2_BLACK);
	SetROP2(hDC, R2_COPYPEN);
	DrawEditSelesCrossNorm(hDC);
	while (!gTmpUndone.IsNull()) {
		((undone *)gTmpUndone.Pop())->Cut()->Undone()->Del();
	}
	gEditCross = NULL;
	return rt;
}

static int __fastcall GetPaintApLy(int &ap)
{
	const node &bas = ((group *)gEditCross)->GetCont();
	for (group *pg = (group *)bas.Ref(); pg != &bas; pg = pg->Ref()) {
		if (pg->GetName() == gPaintedLines) {
			ap = ((element *)pg->GetCont().Ref())->GetAp();
			return ((element *)pg->GetCont().Ref())->GetLy();
		}
	}
	return -1;
}

static void __fastcall PaintDist(HDC hDC)
{
	int ap, ly;
	paint_t *ptr = SetPaintMadoLoop(ly, ap);
	SetROP2(hDC, R2_BLACK);
	DrawEditSeles(hDC, ((group *)gEditCross)->GetCont());
	if (ptr) {
		node bas(0);
		GoPaint(ptr, bas, ly, ap);
		if (!bas.IsNull()) {
			group *pain = group::New(gPaintedIndex);
			const node &nb = ((group *)gEditCross)->GetCont();
			for (container *pc = (container *)nb.Ref(); pc != &nb; pc = pc->Ref()) {
				if (pc->IsGroup() && (((group *)pc)->GetName() == gPaintedLoop || ((group *)pc)->GetName() == gPaintedMado)) {
					pain->AddGroup(pc->Copy());
				}
			}
			group *gtm = group::New(gPaintedMado);
			while (!gEditSeles.IsNull()) {
				gtm->AddGroup((element *)gEditSeles.Ref()->Cut());
			}
			pain->AddGroup(gtm);
			gtm = group::New(gPaintedLines);
			while (!bas.IsNull()) {
				gtm->AddGroup((element *)bas.Ref()->Cut());
			}
			pain->AddGroup(gtm);
			SetROP2(hDC, R2_COPYPEN);
			DrawEditSelesNorm(hDC, pain->GetCont());
			gPcb->Add(pain);
			undone *und = undone::New();
			und->GetAllData((undone *)gTmpUndone.Ref()->Cut())->Del();
			und->AddData(pain);
			AddOwnUndone(und);
			SetFFileChenge();
			gEditCross = NULL;
			return;
		}
	}
	while (gEditSeles.Ref()->Ref() != &gEditSeles) {
		((container *)gEditSeles.Prev())->Cut()->Del();
	}
	DrawEditSelesCrossNorm(hDC);
	while (!gTmpUndone.IsNull()) {
		((undone *)gTmpUndone.Prev()->Cut())->Undone()->Del();
	}
	gEditCross = NULL;
}

static void FnPaintLu2()
{
	static double dl = 0.5;
	pos p;
	GetMouseHitPos(p);
	element *el = gPcb->IsHit(p);
	if (el) {
		HDC hDC = GetScrHandle()->GetScrDC();
		if ((el->IsLine() && ! el->IsSq()) || el->IsArc()) {
			SelectObject(hDC, gSelePen);
			SetROP2(hDC, R2_COPYPEN);
			Draw(hDC, el);
			undone *und = undone::New();
			und->AddData(el, el->Ref());
			gEditSeles.Add(el->Cut());
			if (el->IsLine()) {
				gSeleP1 = el->GetPos();
				gSeleP2 = ((line *)el)->GetPe();
			} else {
				gSeleP1 = ((arc *)el)->GetStPos();
				gSeleP2 = ((arc *)el)->GetEndPos();
			}
			int rt;
			if ((rt = GetPaintLoopFow(und, hDC, 1)) != 0) {
				rt = GetPaintLoopRev(und, hDC, 1);
			}
			if (IsCrossLoopLine() || IsNotLoopInMado()) {
				DrawUndone(hDC, und->GetDataBase());
				und->Undone()->Del();
				MessageBeep(MB_OK);
			} else if (rt) {
				SetOwnTmpUndone(und);
				PushStack(3, IDST_PAINT3);
			} else {
				FlushScr();
				SetOwnTmpUndone(und);
				if (PaintMado(hDC)) {
					MessageBeep(MB_OK);
				}
				PopStack();
			}
		} else if (el->IsLand()) {
			int ap, ly;
			ly = GetPaintApLy(ap);
			if ((unsigned)ly < MAX_LAYER && (unsigned)ap < MAX_AP
					&& CallDialogBoxParam(gApp->Handle(), IDD_PAINT_DIST, PaintDistDlg, (LPARAM)&dl) && dl >= 0.0 && dl < 100.0) {
				double apt = GetCirAp(ap);
				pos pb = el->GetPos();
				if (el->IsSq()) {
					double l = GetSqAp(el->GetAp()) + apt + dl;
					gEditSeles.Add(line::New(ly, ap, 0, pos(pb.x + l, pb.y - l), pos(pb.x + l, pb.y + l)));
					gEditSeles.Add(line::New(ly, ap, 0, pos(pb.x + l, pb.y + l), pos(pb.x - l, pb.y + l)));
					gEditSeles.Add(line::New(ly, ap, 0, pos(pb.x - l, pb.y + l), pos(pb.x - l, pb.y - l)));
					gEditSeles.Add(line::New(ly, ap, 0, pos(pb.x - l, pb.y - l), pos(pb.x + l, pb.y - l)));
				} else {
					double l = GetCirAp(el->GetAp()) + apt + dl;
					gEditSeles.Add(arc::New(ly, ap, pb, l, 0, M_PI));
					gEditSeles.Add(arc::New(ly, ap, pb, l, M_PI, 2.0 * M_PI));
				}
				if (IsCrossLoopLine() || IsNotLoopInMado()) {
					while (!gEditSeles.IsNull()) {
						((container *)gEditSeles.Ref()->Cut())->Del();
					}
					MessageBeep(MB_OK);
				} else {
					PaintDist(hDC);
					PopStack();
				}
			}
		} else if (el->IsRect()) {
			int ap, ly;
			ly = GetPaintApLy(ap);
			if ((unsigned)ly < MAX_LAYER && (unsigned)ap < MAX_AP
					&& CallDialogBoxParam(gApp->Handle(), IDD_PAINT_DIST, PaintDistDlg, (LPARAM)&dl) && dl >= 0.0 && dl < 100.0) {
				double apt = GetCirAp(ap);
				double dx = ((rect *)el)->GetDx() / 2.0 + apt + dl;
				double dy = ((rect *)el)->GetDy() / 2.0 + apt + dl;
				double dl = sqrt(dx * dx + dy * dy);
				double ang1 = atan2(dy, dx);
				double ang = ang1 + ((rect *)el)->GetAng();
				pos pb = el->GetPos();
				pos p1(pb.x + dl * cos(ang), pb.y + dl * sin(ang));
				ang = M_PI - ang1 + ((rect *)el)->GetAng();
				pos p2(pb.x + dl * cos(ang), pb.y + dl * sin(ang));
				ang = M_PI + ang1 + ((rect *)el)->GetAng();
				pos p3(pb.x + dl * cos(ang), pb.y + dl * sin(ang));
				ang = -ang1 + ((rect *)el)->GetAng();
				pos p4(pb.x + dl * cos(ang), pb.y + dl * sin(ang));
				gEditSeles.Add(line::New(ly, ap, 0, p1, p2));
				gEditSeles.Add(line::New(ly, ap, 0, p2, p3));
				gEditSeles.Add(line::New(ly, ap, 0, p3, p4));
				gEditSeles.Add(line::New(ly, ap, 0, p4, p1));
				if (IsCrossLoopLine() || IsNotLoopInMado()) {
					while (!gEditSeles.IsNull()) {
						((container *)gEditSeles.Ref()->Cut())->Del();
					}
					MessageBeep(MB_OK);
				} else {
					PaintDist(hDC);
					PopStack();
				}
			}
		} else {
			MessageBeep(MB_OK);
		}
	} else {
		MessageBeep(MB_OK);
	}
}

static void FnPaintLu3()
{
	pos p;
	GetMouseHitPos(p);
	element *el = gPcb->IsHit(p);
	if (el && ((el->IsLine() && !el->IsSq()) || el->IsArc()) && ((element *)gEditSeles.Prev())->GetLy() == el->GetLy()) {
		int rt;
		if (el->IsEqPos(gSeleP1)) {
			if (el->IsEqPos(gSeleP2)) {
				rt =  0;
			} else {
				gSeleP1 = el->GetFarPos(gSeleP1);
				rt = 1;
			}
		} else if (el->IsEqPos(gSeleP2)) {
			gSeleP2 = el->GetFarPos(gSeleP2);
			rt = 2;
		} else {
			MessageBeep(MB_OK);
			return;
		}
		HDC hDC = GetScrHandle()->GetScrDC();
		SelectObject(hDC, gSelePen);
		SetROP2(hDC, R2_COPYPEN);
		Draw(hDC, el);
		undone *und = undone::New();
		und->AddData(el, el->Ref());
		gEditSeles.Add(el->Cut());
		if (rt) {
			rt = GetPaintLoopFow(und, hDC, 1);
		}
		if (rt) {
			rt = GetPaintLoopRev(und, hDC, 1);
		}
		if (IsCrossLoopLine() || IsNotLoopInMado()) {
			DrawUndone(hDC, und->GetDataBase());
			und->Undone()->Del();
			MessageBeep(MB_OK);
		} else {
			SetOwnTmpUndone(und);
			FlushScr();
			if (rt == 0) {
				if (PaintMado(hDC)) {
					MessageBeep(MB_OK);
				}
				PopStack(2);
			}
		}
	} else {
		MessageBeep(MB_OK);
	}
}

static void FnPaintUnd3()
{
	undone *und = (undone *)gTmpUndone.Prev()->Cut();
	DrawOwnUndone(und);
	und->Undone()->Del();
	if (gEditSeles.IsNull()) {
		PopStack();
	} else {
		element *el = (element *)gEditSeles.Ref();
		if (el->IsLine()) {
			gSeleP1 = el->GetPos();
			gSeleP2 = ((line *)el)->GetPe();
		} else {
			gSeleP1 = ((arc *)el)->GetStPos();
			gSeleP2 = ((arc *)el)->GetEndPos();
		}
		while ((el = el->Ref()) != &gEditSeles) {
			if (el->IsEqPos(gSeleP1)) {
				gSeleP1 = el->GetFarPos(gSeleP1);
			} else if (el->IsEqPos(gSeleP2)) {
				gSeleP2 = el->GetFarPos(gSeleP2);
			}
		}
	}
}

static void FnPaintEsc3()
{
	DrawTmpUndoneElem();
	while (! gTmpUndone.IsNull()) {
		((undone *)gTmpUndone.Prev())->Undone()->Del();
	}
	gEditCross = NULL;
	PopStack(2);
}

static void FnGroupLu()
{
	pos p;
	GetMouseHitPos(p);
	if (GetScrHandle()->GetMouseKey() & MK_CONTROL) {
		SetFAdded();
	} else {
		ClrFAdded();
	}
	element *el;
	if (!(GetScrHandle()->GetMouseKey() & MK_SHIFT) && (el = gPcb->IsHit(p)) != NULL) {
		el = (element *)el->GetTopOwner();
		SetOwnTmpUndone(undone::New(el, el->Ref()));
		gEditSeles.Add(el->Cut());
		DrawEditSeleCros(GetScrHandle()->GetScrDC(), R2_COPYPEN);
		PushStack(1, IDST_PLURALSELE);
		SetButtonOnState();
//		gApp->EnableButton(IDB_LAYER, FALSE);
		gApp->EnableButton(IDB_LAYER, TRUE);
		gApp->EnableButton(IDB_SELECT, TRUE);
		gApp->EnableButton(IDB_PARTS, TRUE);
		gApp->EnableButton(IDB_MOVE, TRUE);
		gApp->EnableButton(IDB_COPY, TRUE);
		gApp->EnableButton(IDB_CUT, TRUE);
		gApp->UpdateButton();
	} else {
		MessageBeep(MB_OK);
	}
}

static void FnGroupLMv()
{
	GetScrHandle()->BeginSeleBox();
	GetScrHandle()->OffLdMove();
	PushStack(2, IDST_PLURALSEL1);
	SetButtonOnState();
	gApp->EnableButton(IDB_SELECT, FALSE);
	gApp->EnableButton(IDB_LAYER, FALSE);
	EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_GRAYED);
	EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_AP, MF_GRAYED);
	DrawMenuBar(GetAppWnd());
	gApp->UpdateButton();
}

static void FnGroupKey()
{
	switch (gEditKey) {
	case EDKEY_LY:
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 0)) {
			GetScrHandle()->Invalidate();
		}
		SetLyApMenu();
		break;
	case EDKEY_SEL:
		CallDialogBox(gApp->Handle(), IDD_SELPLURAL, PlupalSeleDlg);
		break;
	case EDKEY_PARTS:
		if (CallDialogBox(gApp->Handle(), IDD_PARTSIN, PartsInDlg)) {
			break;
		}
	/* no break */
	case EDKEY_PASTE:
		Scr2Snap(GetScrHandle()->GetMouseX(), GetScrHandle()->GetMouseY(), gEditBase);
		PopClipboard(gEditBase, gEditSeles);
		gClicBase.Set(GetOrigX(), GetOrigY());
		DrawEditSeleCros(GetScrHandle()->GetScrDC());
		FlushScr();
		GetScrHandle()->OffLdMove();
		PushStack(5, IDST_PLURALMOVE);
		SetButtonOnState();
		gApp->EnableButton(IDB_SELECT, FALSE);
		gApp->EnableButton(IDB_LAYER, FALSE);
		EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_GRAYED);
		EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_AP, MF_GRAYED);
		DrawMenuBar(GetAppWnd());
		gApp->EnableButton(IDB_ROTATE, TRUE);
		gApp->EnableButton(IDB_MIRROR, TRUE);
		gApp->EnableButton(IDB_CROSSIN, TRUE);
		gApp->EnableButton(IDB_VALUEIN, TRUE);
		gApp->UpdateButton();
		break;
	}
}

static void FnGroupEsc()
{
	FnSetEsc();
	ResetStack(MOD_GRP, IDST_PLURAL);
	GetScrHandle()->OnLdMove();
	gApp->PushButton(IDB_PLURAL, TRUE);
	gApp->EnableButton(IDB_VALUEIN, FALSE);
	gApp->EnableButton(IDB_CROSSIN, FALSE);
	gApp->EnableButton(IDB_SELECT, FALSE);
	gApp->UpdateButton();
	gEditMode = EDMOD_NON;
	SetLyApMenu();
}

static void FnGroupLu1()
{
	pos p;
	GetMouseHitPos(p);
	element *el;
	if (!(GetScrHandle()->GetMouseKey() & MK_SHIFT) && (el = gPcb->IsHit(p)) != NULL) {
		el = (element *)el->GetTopOwner();
		undone *und = undone::New();
		und->AddData(el, el->Ref());
		gEditSeles.Add(el->Cut());
		SetOwnTmpUndone(und);
		DrawEditSeleCros(GetScrHandle()->GetScrDC(), R2_COPYPEN);
	} else if (IsFPluralRev() || (GetScrHandle()->GetMouseKey() & MK_SHIFT)) {
		FnGroupUnd1();
	} else {
		MessageBeep(MB_OK);
	}
}

static void FnGroupLMv1()
{
	GetScrHandle()->OffLdMove();
	GetScrHandle()->BeginSeleBox();
	PushStack(3, IDST_PLURALSEL1);
	gApp->EnableButton(IDB_LAYER, FALSE);
	EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_GRAYED);
	DrawMenuBar(GetAppWnd());
	gApp->UpdateButton();
}

static int CheckSeleGroup()
{
	if (IsFAdded() && gEditSeles.Ref() != &gEditSeles
			&& gEditSeles.Ref()->Ref() != &gEditSeles && ((container *)gEditSeles.Ref())->IsGroup()) {
		return group_t::B_GROUP | group_t::B_ADDED;
	}
	int f = 0;
	for (container *pc = (container *)gEditSeles.Ref(); pc != &gEditSeles; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			f |= group_t::B_GROUP;
		}
	}
	if (gEditSeles.Ref()->Ref() == &gEditSeles) {
		f |= group_t::B_SINGLE;
	}
	return f;
}

static void __fastcall ClearPinNet(const node &bas)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			ClearPinNet(((group *)pc)->GetCont());
		} else if (pc->IsElem()) {
			((element *)pc)->SetNet(NULL);
			((element *)pc)->SetPin(NULL);
			((element *)pc)->SetPinNo(0);
		}
	}
}

static void SeleToGroup(const group_t &pg)
{
	HDC hDC = GetScrHandle()->GetScrDC();
	SetROP2(hDC, R2_COPYPEN);
	DrawEditSelesNorm(hDC, gEditSeles);
	FlushScr();
	undone *und = undone::New();
	group *grp;
	while (!gTmpUndone.IsNull()) {
		und->GetAllData((undone *)gTmpUndone.Ref()->Cut())->Del();
	}
	if (gEditSeles.Ref()->Ref() == &gEditSeles) {
		grp = (group *)((container *)gEditSeles.Ref()->Cut())->Copy();
	} else {
		if (pg.f & group_t::B_ADDED) {
			grp = (group *)((container *)gEditSeles.Ref()->Cut())->Copy();
		} else {
			grp = group::New();
		}
		while (!gEditSeles.IsNull()) {
			container *pc = ((container *)gEditSeles.Ref()->Cut())->Copy();
			grp->AddGroup(pc);
			if (pc->IsGroup()) {
				ClearPinNet(((group *)pc)->GetCont());
			} else if (pc->IsElem()) {
				((element *)pc)->SetNet(NULL);
				((element *)pc)->SetPin(NULL);
				((element *)pc)->SetPinNo(0);
			}
		}
	}
	und->AddData(grp);
	AddOwnUndone(und);
	gPcb->Add(grp);
	GetScrHandle()->OnLdMove();
	PopStack();
}

static void GroupToElem()
{
	HDC hDC = GetScrHandle()->GetScrDC();
	SetROP2(hDC, R2_COPYPEN);
	DrawEditSelesNorm(hDC, gEditSeles);
	FlushScr();
	undone *und = undone::New();
	while (!gTmpUndone.IsNull()) {
		und->GetAllData((undone *)gTmpUndone.Ref()->Cut())->Del();
	}
	while (!gEditSeles.IsNull()) {
		container *pc = (container *)gEditSeles.Ref()->Cut();
		if (pc->IsGroup()) {
			const node &bas = ((group *)pc)->GetCont();
			for (container *pc1 = (container *)bas.Ref(); pc1 != &bas; pc1 = pc1->Ref()) {
				container *pc2 = pc1->Copy();
				pc2->ClearOwner();
				if (pc2->IsGroup()) {
					ClearPinNet(((group *)pc2)->GetCont());
				} else {
					((element *)pc2)->SetNet(NULL);
					((element *)pc2)->SetPin(NULL);
					((element *)pc2)->SetPinNo(0);
				}
				gPcb->Add(pc2);
				und->AddData(pc2);
			}
		} else {
			pc = pc->Copy();
			pc->ClearOwner();
			((element *)pc)->SetNet(NULL);
			((element *)pc)->SetPin(NULL);
			((element *)pc)->SetPinNo(0);
			gPcb->Add(pc);
			und->AddData(pc);
		}
	}
	AddOwnUndone(und);
	GetScrHandle()->OnLdMove();
	PopStack();
}

static void __fastcall WriteLibrary(group_t &pg)
{
	pos p;
	if (gEditSeles.Ref()->Ref() == &gEditSeles && ((container *)gEditSeles.Ref())->IsGroup()) {
		p = GetGroupOrigPos(((group *)gEditSeles.Ref())->GetCont());
	} else {
		p = GetGroupOrigPos(gEditSeles);
	}
	group *pgrp;
	if (gEditSeles.Ref()->Ref() == &gEditSeles && ((container *)gEditSeles.Ref())->IsGroup()) {
		pgrp = ((group *)gEditSeles.Ref())->Copy();
	} else {
		pgrp = group::New();
		for (container *pc = (container *)gEditSeles.Ref(); pc != &gEditSeles; pc = pc->Ref()) {
			pgrp->AddGroup(pc->Copy());
		}
	}
	WritLibrary(p, pgrp->GetCont(), (char *)pg.path, (char *)pg.name, 0, (char *)pg.attrib);
	pgrp->SetExt(group::ext_parts::New(p));
	pgrp->SetName((char *)pg.name);
	pgrp->SetAttrib((char *)pg.attrib);
	pgrp->SetUnit(NULL);
	undone *und = undone::New();
	while (!gTmpUndone.IsNull()) {
		und->GetAllData((undone *)gTmpUndone.Ref()->Cut())->Del();
	}
	gPcb->Add(pgrp);
	und->AddData(pgrp);
	AddOwnUndone(und);
	HDC hDC = GetScrHandle()->GetScrDC();
	SetROP2(hDC, R2_COPYPEN);
	DrawEditSelesNorm(hDC, pgrp->GetCont());
	FlushScr();
	while (!gEditSeles.IsNull()) {
		gEditSeles.Ref()->Cut();
	}
	GetScrHandle()->OnLdMove();
	PopStack();
}

static void ChengePartsLy()
{
	HDC hDC = GetScrHandle()->GetScrDC();
	DrawEditSeleCros(hDC, R2_BLACK);
	FlushScr();
	pos p = GetGroupOrigPos(gEditSeles);
	SwapParts(gEditSeles, p);
	FnGroupEsc1();
}

static void FnGroupKey1()
{
	static group_t pg = { 0, 0, "", "", "" };
	undone *und;
	switch (gEditKey) {
	case EDKEY_LY:
		if (CallDialogBoxParam(gApp->Handle(), IDD_SELLAYER, LayerSelDlg, 0)) {
			GetScrHandle()->Invalidate();
		}
		SetLyApMenu();
		break;
	case EDKEY_SEL:
		CallDialogBox(gApp->Handle(), IDD_SELPLURAL, PlupalSeleDlg);
		break;
	case EDKEY_MOV:
		MoveStack(4, IDST_PLURALORG);
		gApp->EnableButton(IDB_SELECT, FALSE);
		gApp->EnableButton(IDB_LAYER, FALSE);
		EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_GRAYED);
		EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_AP, MF_GRAYED);
		DrawMenuBar(GetAppWnd());
		gApp->EnableButton(IDB_PARTS, FALSE);
		gApp->EnableButton(IDB_CUT, FALSE);
		gApp->EnableButton(IDB_COPY, FALSE);
		gApp->EnableButton(IDB_MOVE, FALSE);
		gApp->EnableButton(IDB_CROSSIN, TRUE);
		gApp->EnableButton(IDB_VALUEIN, TRUE);
		gApp->UpdateButton();
		break;
	case EDKEY_PARTS:
		pg.f = CheckSeleGroup();
		switch (CallDialogBoxParam(gApp->Handle(), IDD_SETGROUP, SetGroupDlg, (LPARAM)&pg)) {
		case 0:
			return;
		case 1:
			SeleToGroup(pg);
			break;
		case 2:
			GroupToElem();
			break;
		case 3:
			if (CallDialogBoxParam(gApp->Handle(), IDD_PARTSWRITE, PartsWriteDlg, (LPARAM)&pg)) {
				if (*pg.path && *pg.name) {
					WriteLibrary(pg);
				} else {
					MessageBeep(MB_OK);
				}
			}
			return;
		case 4:
			ChengePartsLy();
		}
		SetFFileChenge();
		break;
	case EDKEY_CUT:
		PushClipboard(GetPushPos(gEditSeles), gEditSeles);
	/* no break */
	case EDKEY_DELETE:
		und = undone::New();
		DrawEditSeles(1);
		while (!gEditSeles.IsNull()) {
			gEditSeles.Ref()->Cut();
		}
		while (!gTmpUndone.IsNull()) {
			und->GetAllData((undone *)gTmpUndone.Ref())->Del();
		}
		AddOwnUndone(und);
		GetScrHandle()->OnLdMove();
		FlushScr();
		PopStack();
		SetFFileChenge();
		break;
	case EDKEY_COPY:
		PushClipboard(GetPushPos(gEditSeles), gEditSeles);
		DrawTmpUndoneElem();
		while (!gTmpUndone.IsNull()) {
			((undone *)gTmpUndone.Prev())->Undone()->Del();
		}
		GetScrHandle()->OnLdMove();
		FlushScr();
		PopStack();
		break;
	case EDKEY_SETGRP:
		pg.f = CheckSeleGroup();
		SeleToGroup(pg);
		SetFFileChenge();
		break;
	case EDKEY_CANGRP:
		GroupToElem();
		SetFFileChenge();
		break;
	}
}

static void FnGroupUnd1()
{
	if (! gTmpUndone.IsNull()) {
		HDC hDC = GetScrHandle()->GetScrDC();
		SetROP2(hDC, R2_COPYPEN);
		undone *und = (undone *)gTmpUndone.Prev()->Cut();
		const node &bas = und->GetDataBase();
		for (undone::data *data = (undone::data *)bas.Ref(); data != &bas; data = data->Ref()) {
			DrawEditCrossNorm(hDC, data->GetAdd());
		}
		FlushScr();
		und->Undone()->Del();
	}
	if (gEditSeles.IsNull()) {
		GetScrHandle()->OnLdMove();
		PopStack();
	}
}

static void FnGroupEsc1()
{
	DrawTmpUndoneElem();
	while (! gTmpUndone.IsNull()) {
		((undone *)gTmpUndone.Prev())->Undone()->Del();
	}
	while (! gEditSeles.IsNull()) {
		gPcb->Add((element *)gEditSeles.Ref()->Cut());
	}
	GetScrHandle()->OnLdMove();
	PopStack();
}

static void FnGroupLu2()
{
	pos p, p1 = GetScrHandle()->GetSeleBase();
	GetMouseHitPos(p);
	GetScrHandle()->EndSeleBox();
	GetScrHandle()->OnLdMove();
	undone *und;
	if ((und = GetInBoxElems(p, p1, gEditSeles, gPcb->GetElems())) != NULL) {
		SetOwnTmpUndone(und);
		HDC hDC = GetScrHandle()->GetScrDC();
		SelectObject(hDC, gSelePen);
		SetROP2(hDC, R2_COPYPEN);
		DrawEditSeles(hDC, gEditSeles);
		MoveStack(1, IDST_PLURALSELE);
		gApp->EnableButton(IDB_MOVE, TRUE);
		gApp->EnableButton(IDB_PARTS, TRUE);
		gApp->EnableButton(IDB_LAYER, TRUE);
		gApp->EnableButton(IDB_SELECT, TRUE);
		EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_LY, MF_ENABLED);
		EnableMenuItem(GetMenu(GetAppWnd()), IDM_MENU_AP, MF_ENABLED);
		DrawMenuBar(GetAppWnd());
		gApp->EnableButton(IDB_COPY, TRUE);
		gApp->EnableButton(IDB_CUT, TRUE);
		gApp->UpdateButton();
	} else {
		MessageBeep(MB_OK);
		GetScrHandle()->OnLdMove();
		PopStack();
	}
}

static void FnGroupEsc2()
{
	GetScrHandle()->EndSeleBox();
	GetScrHandle()->OnLdMove();
	PopStack();
}

static void FnGroupLu3()
{
	pos p, p1 = GetScrHandle()->GetSeleBase();
	GetMouseHitPos(p);
	GetScrHandle()->EndSeleBox();
	GetScrHandle()->OnLdMove();
	undone *und;
	node bas(0);
	if (IsFPluralRev() || (GetScrHandle()->GetMouseKey() & MK_SHIFT)
			|| (und = GetInBoxElems(p, p1, bas, gPcb->GetElems())) == NULL) {
		MessageBeep(MB_OK);
	} else {
		HDC hDC = GetScrHandle()->GetScrDC();
		SetROP2(hDC, R2_COPYPEN);
		SelectObject(hDC, gSelePen);
		DrawEditSeles(hDC, bas);
		gEditSeles.GetAllNode(bas);
		SetOwnTmpUndone(und);
	}
	PopStack();
}

static void FnGroupEsc3()
{
	MessageBeep(MB_OK);
}

static void MoveGroupMove()
{
	gClicBase = gEditBase;
	node bas(0);
	undone *und = undone::New();
	bas.GetAllNode(gEditSeles);
	while (!bas.IsNull()) {
		container *pc = ((container *)bas.Ref())->Cut()->Copy();
		und->AddData(pc);
		gEditSeles.Add(pc);
	}
	AddOwnUndone(und);
	GetScrHandle()->OffLdMove();
	MoveStack(5, IDST_PLURALMOVE);
	gApp->EnableButton(IDB_ROTATE, TRUE);
	gApp->EnableButton(IDB_MIRROR, TRUE);
	gApp->UpdateButton();
}

static void FnGroupLu4()
{
	GetMouseSnapPos(gEditBase);
	MoveGroupMove();
}

static void FnGroupValue4()
{
	if (gValueX < 0.0 || gValueX > GetSizeX() || gValueY < 0.0 || gValueY > GetSizeY()) {
		MessageBeep(MB_OK);
	} else {
		gEditBase.Set(gValueX, gValueY);
		MoveGroupMove();
	}
}

static void FnGroupKey4()
{
	switch (gEditKey) {
	case EDKEY_VAL:
		EdKeyValSub(1);
		break;
	case EDKEY_MOVEGRP:
		if (gEditSeles.IsNull()) {
			break;
		}
		if (gEditSeles.Ref()->Ref() == &gEditSeles && ((container *)gEditSeles.Ref())->IsGroup()) {
			gEditBase = GetGroupOrigPos(((group *)gEditSeles.Ref())->GetCont());
		} else {
			gEditBase = GetGroupOrigPos(gEditSeles);
		}
		MoveGroupMove();
		break;
	}
}

static void FnGroupUnd4()
{
	DrawTmpUndoneElem();
	while (!gTmpUndone.IsNull()) {
		((undone *)gTmpUndone.Prev())->Undone()->Del();
	}
	GetScrHandle()->OnLdMove();
	FlushScr();
	PopStack();
}

static void __fastcall GroupMove5Sub(const pos &p)
{
	double dx = p.x - gEditBase.x;
	double dy = p.y - gEditBase.y;
	container *pc;
	for (pc = (container *)gEditSeles.Ref(); pc != &gEditSeles; pc = pc->Ref()) {
		pc->Move(dx, dy);
	}
	gEditBase = p;
}

static void FnGroupMv5()
{
	HDC hDC = GetScrHandle()->GetScrDC();
	DrawEditSeleCros(hDC);
	pos p;
	GetMouseSnapPos(p);
	GroupMove5Sub(p);
	DrawEditSeleCros(hDC);
	FlushScr();
}

static void GroupEnter5(const pos &p)
{
	HDC hDC = GetScrHandle()->GetScrDC();
	DrawEditSeleCros(hDC);
	GroupMove5Sub(p);
	if (IsOutSize(gEditSeles)) {
		DrawEditSeleCros(GetScrHandle()->GetScrDC());
		MessageBeep(MB_OK);
		return;
	}
	container *pc;
	undone *und = undone::New();
	SetROP2(hDC, R2_MERGEPEN);
	DrawEditSelesNorm(hDC, gEditSeles);
	while (!gEditSeles.IsNull()) {
		gPcb->Add(pc = (container *)gEditSeles.Ref()->Cut());
		und->AddData(pc);
	}
	while (!gTmpUndone.IsNull()) {
		und->GetAllData((undone *)gTmpUndone.Ref()->Cut())->Del();
	}
	AddOwnUndone(und);
	GetScrHandle()->OnLdMove();
	PopStack();
	SetFFileChenge();
}

static void FnGroupLu5()
{
	pos p;
	GetMouseSnapPos(p);
	GroupEnter5(p);
}

static void GroupMirror()
{
	HDC hDC = GetScrHandle()->GetScrDC();
	DrawEditSeleCros(hDC);
	for (container *pc = (container *)gEditSeles.Ref(); pc != &gEditSeles; pc = pc->Ref()) {
		pc->Mirror(gEditBase);
	}
	DrawEditSeleCros(hDC);
	FlushScr();
}

static void GroupRot()
{
	static double ang = M_PI / 2.0;
	if (CallDialogBoxParam(gApp->Handle(), IDD_GETROT, GetRotAngDlg, (LPARAM)&ang)) {
		return;
	}
	HDC hDC = GetScrHandle()->GetScrDC();
	DrawEditSeleCros(hDC);
	for (container *pc = (container *)gEditSeles.Ref(); pc != &gEditSeles; pc = pc->Ref()) {
		pc->Rotate(gEditBase, ang);
	}
	DrawEditSeleCros(hDC);
	FlushScr();
}

static void FnGroupValue5()
{
	GroupEnter5(pos(gValueX, gValueY));
}

static void FnGroupKey5()
{
	switch (gEditKey) {
	case EDKEY_MIRROR:
		GroupMirror();
		break;
	case EDKEY_ROT:
		GroupRot();
		break;
	case EDKEY_VAL:
		EdKeyValSub(0);
		break;
	}
}

static void FnGroupUnd5()
{
	DrawEditSeleCros(GetScrHandle()->GetScrDC());
	FlushScr();
	while (!gEditSeles.IsNull()) {
		((container *)gEditSeles.Ref()->Cut())->Del();
	}
	FnGroupUnd4();
}

void SavePadCross()
{
	gApp->GetScreen()->SaveDrawPad(gCrsPadSave);
}

static inline void RestorePad(node &bas)
{
	gApp->GetScreen()->RestoreDrawPad(bas);
}

static void FnCrsInLu()
{
	pos p;
	GetMouseHitPos(p);
	element *pe;
	if ((pe = gPcb->IsHit(p)) == NULL) {
		MessageBeep(MB_OK);
		return;
	} else if (GetScrHandle()->GetMouseKey() & MK_SHIFT) {
		if (pe->IsArc()) {
			p = ((arc*)pe)->GetPos();
		} else {
			MessageBeep(MB_OK);
			return;
		}
	} else if (pe->IsLine() || pe->IsArc()) {
		pos pn = pe->GetNearPos(p);
		if (p.Distance(pn) > GetHitInPoint(pe)) {
			SeleDrawAndSetEditSele(pe->Cut());
			PushStack(1, IDST_CROSIN2);
			return;
		}
		p = pn;
	} else {
		p = pe->GetNearPos(p);
	}
	PopStack();
	RestorePad(gCrsPadSave);
	RunEditValue(p.x, p.y);
}

static void FnCrsInUnd()
{
	PopStack();
	RestorePad(gCrsPadSave);
}

static void FnCrsInLu1()
{
	pos p;
	GetMouseHitPos(p);
	element *pe;
	if ((pe = gPcb->IsHit(p)) != NULL) {
		int f;
		pos cx[2];
		if (pe->IsLine() || pe->IsArc()) {
			if (gEditCross->IsLine()) {
				if (pe->IsArc()) {
					f = GetCrossLineArcPos((line *)gEditCross, (arc *)pe, cx);
				} else if ((f = GetCrossLineLinePos((line *)pe, (line *)gEditCross, cx[0])) == 0) {
					cx[1] = cx[0];
				}
			} else if (pe->IsLine()) {
				f = GetCrossLineArcPos((line *)pe, (arc *)gEditCross, cx);
			} else {
				f = GetCrossArcArcPos((arc *)pe, (arc *)gEditCross, cx);
			}
			RedrawAndBackEditCross();
			PopStack(2);
			RestorePad(gCrsPadSave);
			if (f) {
				MessageBeep(MB_OK);
			} else if (p.Distance(cx[0]) < p.Distance(cx[1])) {
				RunEditValue(cx[0].x, cx[0].y);
			} else {
				RunEditValue(cx[1].x, cx[1].y);
			}
		} else {
			MessageBeep(MB_OK);
		}
	} else {
		MessageBeep(MB_OK);
	}
}

static void FnCrsInUnd1()
{
	RedrawAndBackEditCross();
	PopStack();
}

static void FnCrsInEsc1()
{
	RedrawAndBackEditCross();
	PopStack(2);
	RestorePad(gCrsPadSave);
}

void SavePadZoom()
{
	gApp->GetScreen()->SaveDrawPad(gZoomPadSave);
}

static void FnZoomInLUp1()
{
	GetScrHandle()->BeginZoom();
	PushStack(1, IDST_ZOOM2);
}

static void FnZoomInUnd1()
{
	GetScrHandle()->EndZoom();
	PopStack();
	RestorePad(gZoomPadSave);
}

static void FnZoomInLUp2()
{
	SaveScreenOld();
	Screen *scr = GetScrHandle();
	scr->EndZoom();
	PopStack(2);
	RestorePad(gZoomPadSave);
	double zoom = GetZoom();
	double zmX = scr->GetZoomX() == scr->GetMouseX() ? 1000.0 :
				 zoom * scr->GetScrX() / abs(scr->GetMouseX() - scr->GetZoomX());
	double zmY = scr->GetZoomY() == scr->GetMouseY() ? 1000.0 :
				 zoom * scr->GetScrY() / abs(scr->GetMouseY() - scr->GetZoomY());
	double zoomn;
	if ((zoomn = zmX < zmY ? zmX : zmY) > 1000.0) {
		zoomn = 1000.0;
	}
	pos p;
	Scr2NoSnap((scr->GetMouseX() + scr->GetZoomX()) / 2, (scr->GetMouseY() + scr->GetZoomY()) / 2, p);
	SetOfsAndZoom(p.x - scr->GetScrX() / zoomn / 2.0, p.y - scr->GetScrY() / zoomn / 2.0, zoomn);
}

static void FnZoomInEsc2()
{
	GetScrHandle()->EndZoom();
	PopStack(2);
	RestorePad(gZoomPadSave);
}

void SavePadOrig()
{
	gApp->GetScreen()->SaveDrawPad(gOrgPadSave);
}

static void FnOrigLu()
{
	pos p;
	GetMouseHitPos(p);
	SetOrig(p.x, p.y);
	PopStack();
	RestorePad(gOrgPadSave);
	GetScrHandle()->Invalidate();
}

static void FnOrigValu()
{
	if (gValueX >= 0.0 && gValueX <= GetSizeX() && gValueY >= 0.0 && gValueY <= GetSizeY()) {
		SetOrig(gValueX, gValueY);
		PopStack();
		RestorePad(gOrgPadSave);
		GetScrHandle()->Invalidate();
	} else {
		MessageBeep(MB_OK);
	}
}

static void FnOrigKey()
{
	if (gEditKey == EDKEY_VAL) {
		EdKeyValSub(1);
	}
}

static void FnOrigUnd()
{
	PopStack();
	RestorePad(gOrgPadSave);
}

void CoorPadSave()
{
	gApp->GetScreen()->SaveDrawPad(gCoorPadSave);
}

static void __fastcall DispCoor(double x, double y)
{
	if (IsFCoorOrig()) {
		x -= GetOrigX();
		y -= GetOrigY();
	}
	double pp[4];
	pp[0] = x;
	pp[1] = y;
	pp[2] = x - GetOrigX();
	pp[3] = y - GetOrigY();
	CallDialogBoxParam(gApp->Handle(), IDD_DSPPOS, DispPosiDlg, (LPARAM)pp);
	PopStack();
	RestorePad(gCoorPadSave);
}

static void FnCoorLu()
{
	pos p;
	GetMouseSnapPos(p);
	element *pe;
	if ((pe = gPcb->IsHit(p)) != NULL) {
		HDC hDC;
		HPEN hPenOld = (HPEN)SelectObject(hDC = GetDC(GetScrHandle()->Handle()), gSelePen);
		HBRUSH hBrOld = (HBRUSH)SelectObject(hDC, gNullBrush);
		SetROP2(hDC, R2_COPYPEN);
		double ckap = GetHitInPoint(pe);
		if (pe->IsLand()) {
			Draw(hDC, pe);
			p = pe->GetPos();
		} else if (pe->IsArc()) {
			Draw(hDC, pe);
			if (p.Distance(pe->GetNearPos(p)) < ckap) {
				p = pe->GetNearPos(p);
			} else {
				p = pe->GetPos();
			}
		} else if (pe->IsLine()) {
			Draw(hDC, pe);
			if (p.Distance(pe->GetNearPos(p)) < ckap) {
				p =  pe->GetNearPos(p);
			} else
				p.Set((pe->GetPos().x + ((line *)pe)->GetPe().x) / 2.0,
					  (pe->GetPos().y + ((line *)pe)->GetPe().y) / 2.0);
		}
		SelectObject(hDC, hBrOld);
		SelectObject(hDC, hPenOld);
		ReleaseDC(GetScrHandle()->Handle(), hDC);
	}
	DispCoor(p.x, p.y);
}

static void FnCoorValu()
{
	DispCoor(gValueX, gValueY);
}

static void FnCoorKey()
{
	if (gEditKey == EDKEY_SEL) {
		CallDialogBox(gApp->Handle(), IDD_COORSELE, CoorSeleDlg);
	} else if (gEditKey == EDKEY_VAL) {
		EdKeyValSub(1);
	}
}

static void FnCoorUnd()
{
	PopStack();
	RestorePad(gCoorPadSave);
}

void DispPadSave()
{
	gApp->GetScreen()->SaveDrawPad(gDistPadSave);
}

static void FnDistLu()
{
	GetMouseHitPos(gDistClic);
	element *pe;
	if ((pe = gPcb->IsHit(gDistClic)) != NULL) {
		SeleDrawAndSetEditSele(pe->Cut());
		if (GetScrHandle()->GetMouseKey() & MK_SHIFT) {
			PushStack(2, IDST_DIST1);
		} else {
			PushStack(1, IDST_DIST1);
		}
	} else {
		MessageBeep(MB_OK);
	}
}

static void FnDistUnd()
{
	PopStack();
	RestorePad(gDistPadSave);
}

static void FnDistLu1()
{
	pos p;
	GetMouseHitPos(p);
	element *pe;
	if ((pe = gPcb->IsHit(p)) != NULL) {
		pos po[2];
		if (GetInterval(gEditCross, pe, po) || po[0].IsEq(po[1])) {
			FnDistEsc1();
			DispMessage(IDS_SHRT_MSG);
		} else {
			FnDistEsc1();
			double x, y;
			Pos2NoSnap(po[0], x, y);
			HDC hDC = GetScrHandle()->GetScrDC();
			SelectObject(hDC, gSelePen);
			SetROP2(hDC, R2_COPYPEN);
			MoveToScr(hDC, long(x), long(y));
			Pos2NoSnap(po[1], x, y);
			LineToScr(hDC, long(x), long(y));
			FlushScr();
			CallDialogBoxParam(gApp->Handle(), IDD_DSPINTERVAL, DispIntervalDlg, (LPARAM)po);
		}
	} else {
		MessageBeep(MB_OK);
	}
}

static void FnDistUnd1()
{
	RedrawAndBackEditCross();
	PopStack();
}

static void FnDistEsc1()
{
	RedrawAndBackEditCross();
	PopStack(2);
	RestorePad(gDistPadSave);
}

static void FnDistLu2()
{
	pos p;
	GetMouseHitPos(p);
	element *pe;
	if ((pe = gPcb->IsHit(p)) != NULL) {
		pos po[2];
		po[0] = gEditCross->GetNearPos(gDistClic);
		po[1] = pe->GetNearPos(p);
		FnDistEsc2();
		double x, y;
		Pos2NoSnap(po[0], x, y);
		HDC hDC = GetScrHandle()->GetScrDC();
		SelectObject(hDC, gSelePen);
		SetROP2(hDC, R2_COPYPEN);
		MoveToScr(hDC, long(x), long(y));
		Pos2NoSnap(po[1], x, y);
		LineToScr(hDC, long(x), long(y));
		FlushScr();
		CallDialogBoxParam(gApp->Handle(), IDD_DSPINTERVAL, DispIntervalDlg, (LPARAM)po);
	} else {
		MessageBeep(MB_OK);
	}
}

static inline element *Work2Element(void *ptr)
{
	return (element*)Work2Cont(ptr);
}

static inline group *Wrok2Grp(void *ptr)
{
	return (group*)Work2Cont(ptr);
}

static inline int GetPaintLy(const node& bas)
{
	return ((element*)((group*)bas.Ref())->GetCont().Ref())->GetLy();
}

static inline int GetPaintLy(group *pg)
{
	return GetPaintLy(pg->GetCont());
}

static void __fastcall SetNetBase(node& set, const node& bas)
{
	for (container *pc = (container*)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			if (((group*)pc)->GetName() == gPaintedIndex) {
				if (gPcb->IsFJmpPtn(GetPaintLy((group*)pc))) {
					set.Add((node*)&pc->tmp.nod);
				}
			} else {
				SetNetBase(set, ((group*)pc)->GetCont());
			}
		} else if (pc->IsLand() && !((land*)pc)->IsSq() && gPcb->IsFThJumper((element*)pc)) {
			set.Add((node*)&pc->tmp.nod);
		} else if (gPcb->IsFJmpPtn((element*)pc)) {
			set.Add((node*)&pc->tmp.nod);
		}
	}
}

static void __fastcall ClearAllNetNo(const node& bas)
{
	for (container *pc = (container*)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			ClearAllNetNo(((group*)pc)->GetCont());
		} else {
			((element*)pc)->SetNetNo(0);
		}
	}
}

static void __fastcall SetNetBase(node& bas)
{
	bas.Init();
	ClearAllWork(gPcb->GetElems());
	ClearAllNetNo(gPcb->GetElems());
	SetNetBase(bas, gPcb->GetElems());
}

static int __fastcall ChkEqNetLand(land *pl, container *chk)
{
	int f = gPcb->IsFThJumper(pl) && !pl->IsSq();
	switch (chk->GetType()) {
	case container::T_LAND:
		if (gPcb->IsFThJumper((land*)chk) && !((land*)chk)->IsSq()) {
			f = 1;
		}
		// no break
	case container::T_LINE:
	case container::T_ARC:
	case container::T_RECT:
		if ((f || pl->GetLy() == ((element*)chk)->GetLy()) && ((element*)chk)->IsEqPos(pl->GetPos())) {
			return 1;
		}
		break;
	case container::T_GROUP:
		if (((gPcb->IsFTh(pl) && !pl->IsSq()) || GetPaintLy((group*)chk) == pl->GetLy())
				&& IsPointInPaint(((group*)chk)->GetCont(), pl->GetPos())) {
			return 1;
		}
		break;
	}
	return 0;
}

static int __fastcall ChkEqNetLine(line *pl, container *chk)
{
	int f = 0;
	switch (chk->GetType()) {
	case container::T_LAND:
		if (gPcb->IsFThJumper((land*)chk) && !((land*)chk)->IsSq()) {
			f = 1;
		}
		// no break
	case container::T_LINE:
	case container::T_ARC:
	case container::T_RECT:
		if (f || pl->GetLy() == ((element*)chk)->GetLy()) {
			if (((element*)chk)->IsEqPos(pl->GetPos()) || ((element*)chk)->IsEqPos(pl->GetPe())) {
				return 1;
			}
		}
		break;
	case container::T_GROUP:
		if (GetPaintLy((group*)chk) == pl->GetLy()) {
			if (IsPointInPaint(((group*)chk)->GetCont(), pl->GetPos())
					|| IsPointInPaint(((group*)chk)->GetCont(), pl->GetPe())) {
				return 1;
			}
		}
		break;
	}
	return 0;
}

static int __fastcall ChkEqNetArc(arc *pa, container *chk)
{
	int f = 0;
	switch (chk->GetType()) {
	case container::T_LAND:
		if (gPcb->IsFThJumper((land*)chk) && !((land*)chk)->IsSq()) {
			f = 1;
		}
		// no break
	case container::T_LINE:
	case container::T_ARC:
	case container::T_RECT:
		if (f || pa->GetLy() == ((element*)chk)->GetLy()) {
			if (((element*)chk)->IsEqPos(pa->GetStPos()) || ((element*)chk)->IsEqPos(pa->GetEndPos())) {
				return 1;
			}
		}
		break;
	case container::T_GROUP:
		if (GetPaintLy((group*)chk) == pa->GetLy()) {
			if (IsPointInPaint(((group*)chk)->GetCont(), pa->GetStPos())
					|| IsPointInPaint(((group*)chk)->GetCont(), pa->GetEndPos())) {
				return 1;
			}
		}
		break;
	}
	return 0;
}

static int __fastcall ChkEqNetRect(rect *pr, container *chk)
{
	int f = 0;
	switch (chk->GetType()) {
	case container::T_LAND:
		if (gPcb->IsFThJumper((land*)chk) && !((land*)chk)->IsSq()) {
			f = 1;
		}
		// no break
	case container::T_LINE:
	case container::T_ARC:
	case container::T_RECT:
		if (f || pr->GetLy() == ((element*)chk)->GetLy()) {
			if (((element*)chk)->IsEqPos(pr->GetPos())
					|| ((element*)chk)->IsEqPos(pr->GetP1())
					|| ((element*)chk)->IsEqPos(pr->GetP2())) {
				return 1;
			}
		}
		break;
	case container::T_GROUP:
		if (GetPaintLy((group*)chk) == pr->GetLy()) {
			if (IsPointInPaint(((group*)chk)->GetCont(), pr->GetPos())
					|| IsPointInPaint(((group*)chk)->GetCont(), pr->GetP1())
					|| IsPointInPaint(((group*)chk)->GetCont(), pr->GetP2())) {
				return 1;
			}
		}
		break;
	}
	return 0;
}

static int __fastcall IsBetaInBeta(const node& b1, const node& b2)
{
	for (group *pg = (group*)b1.Ref(); pg != &b1; pg = pg->Ref()) {
		const node *pn = &pg->GetCont();
		for (element *pe = (element*)pn->Ref(); pe != pn; pe = pe->Ref()) {
			if (pe->IsLine()) {
				if (IsPointInPaint(b2, pe->GetPos()) || IsPointInPaint(b2, ((line*)pe)->GetPos())) {
					return 1;
				}
			} else if (pe->IsArc()) {
				if (IsPointInPaint(b2, ((arc*)pe)->GetStPos()) || IsPointInPaint(b2, ((arc*)pe)->GetEndPos())) {
					return 1;
				}
			}
		}
	}
	return 0;
}

static int __fastcall ChkEqNetBeta(const node& bas, container *chk)
{
	int ly = GetPaintLy(bas);
	switch (chk->GetType()) {
	case container::T_LAND:
		if (((gPcb->IsFTh((land*)chk) && !((land*)chk)->IsSq()) || ((land*)chk)->GetLy() == ly)
				&& IsPointInPaint(bas, ((land*)chk)->GetPos())) {
			return 1;
		}
		break;
	case container::T_LINE:
		if (((line*)chk)->GetLy() == ly) {
			if (IsPointInPaint(bas, ((line*)chk)->GetPos()) || IsPointInPaint(bas, ((line*)chk)->GetPe())) {
				return 1;
			}
		}
		break;
	case container::T_ARC:
		if (((arc*)chk)->GetLy() == ly) {
			if (IsPointInPaint(bas, ((arc*)chk)->GetStPos()) || IsPointInPaint(bas, ((arc*)chk)->GetEndPos())) {
				return 1;
			}
		}
		break;
	case container::T_RECT:
		if (((rect*)chk)->GetLy() == ly) {
			if (IsPointInPaint(bas, ((rect*)chk)->GetPos())
					|| IsPointInPaint(bas, ((rect*)chk)->GetP1())
					|| IsPointInPaint(bas, ((rect*)chk)->GetP2())) {
				return 1;
			}
		}
		break;
	case container::T_GROUP:
		if (GetPaintLy((group*)chk) == ly) {
			if (IsBetaInBeta(bas, ((group*)chk)->GetCont()) || IsBetaInBeta(((group*)chk)->GetCont(), bas)) {
				return 1;
			}
		}
		break;
	}
	return 0;
}

static void __fastcall GetOwnNet(node& in, node *pn, node& bas)
{
	node tmp(0);
	node *chk = bas.Ref();
	for (;;) {
		container *pc = Work2Cont(pn);
		in.Add(pn);
		while (chk != &bas) {
			container *pck = Work2Cont(chk);
			int f;
			switch (pc->GetType()) {
			case container::T_LAND:
				f = ChkEqNetLand((land*)pc, pck);
				break;
			case container::T_LINE:
				f = ChkEqNetLine((line*)pc, pck);
				break;
			case container::T_ARC:
				f = ChkEqNetArc((arc*)pc, pck);
				break;
			case container::T_RECT:
				f = ChkEqNetRect((rect*)pc, pck);
				break;
			case container::T_GROUP:
				f = ChkEqNetBeta(((group*)pc)->GetCont(), pck);
				break;
			default:
				f = 0;
				break;
			}
			if (f) {
				node *back = chk->Prev();
				tmp.Add(chk->Cut());
				chk = back->Ref();
			} else {
				chk = chk->Ref();
			}
		}
		if (tmp.IsNull()) {
			break;
		}
		pn = tmp.Ref()->Cut();
		chk = bas.Ref();
	}
}

static void __fastcall SetNetNo(const node& bas, int no)
{
	for (container *pc = (container*)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			SetNetNo(((group*)pc)->GetCont(), no);
		} else {
			((element*)pc)->SetNetNo(no);
		}
	}
}

static void __fastcall GetOwnNet(container *pc, node& net, int no = 1)
{
	node chkBase;
	SetNetBase(chkBase);
	if (pc->GetOwner()) {
		group *pg = (group*)pc->GetOwner();
		if (pg->GetName() == gPaintedLines || pg->GetName() == gPaintedLoop || pg->GetName() == gPaintedMado) {
			while ((pg = (group*)pg->GetOwner()) != NULL) {
				if (pg->GetName() == gPaintedIndex) {
					break;
				}
			}
			if (pg == NULL) {
				net.Add(((node*)&pc->tmp.nod)->Cut());
				return;
			}
			pc = pg;
		}
	}
	if ((pc->IsElem() && (gPcb->IsFJmpPtn((element*)pc)
			|| (pc->IsLand() && !((element*)pc)->IsSq() && gPcb->IsFThJumper((element*)pc))))
			|| (pc->IsGroup() && ((group*)pc)->GetName() == gPaintedIndex
			    && gPcb->IsFJmpPtn(GetPaintLy((group*)pc)))) {
		GetOwnNet(net, ((node*)&pc->tmp.nod)->Cut(), chkBase);
		for (node *pn = net.Ref(); pn != &net; pn = pn->Ref()) {
			container *pc = Work2Cont(pn);
			if (pc->IsGroup()) {
				SetNetNo(((group*)pc)->GetCont(), no);
			} else {
				((element*)pc)->SetNetNo(no);
			}
		}
	}
}

node* GetAllNet()
{
	node chk;
	SetNetBase(chk);
	node *pn = (node*)((char*)gTempMem + sizeof(node));
	ZeroMemory(gTempMem, sizeof(node));
	int chkCnt = CHECK_NUM;
	int no = 1;
	BeginCheckStop();
	while (!chk.IsNull()) {
		pn->Init();
		GetOwnNet(*pn, chk.Ref()->Cut(), chk);
		for (node *tmp = pn->Ref(); tmp != pn; tmp = tmp->Ref()) {
			container *pc = Work2Cont(tmp);
			if (pc->IsGroup()) {
				SetNetNo(((group*)pc)->GetCont(), no);
			} else {
				((element*)pc)->SetNetNo(no);
			}
		}
		++no;
		if (++pn > gTempMemMax) {
			break;
		}
		if (--chkCnt == 0) {
			if (CheckStop()) {
				EndCheckStop();
				return NULL;
			}
			chkCnt = CHECK_NUM;
		}
	}
	EndCheckStop();
	pn->Init();
	return (node*)((char*)gTempMem + sizeof(node));
}

static container* __fastcall GetHitNetElem(const pos& p, const node& bas)
{
	for (container *pc = (container*)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			container *hit = GetHitNetElem(p, ((group*)pc)->GetCont());
			if (hit) {
				return hit;
			}
		} else if (gPcb->IsFJmpPtn((element*)pc) && ((element*)pc)->IsHit(p)) {
			return pc;
		}
	}
	return NULL;
}

static void DispNet(node& bas)
{
	HDC hDC = GetScrHandle()->GetScrDC();
	SelectObject(hDC, gSelePen);
	SetROP2(hDC, R2_COPYPEN);
	for (node *pn = bas.Ref(); pn != &bas; pn = pn->Ref()) {
		container *pc = Work2Cont(pn);
		if (pc->IsGroup()) {
			const node &bas = ((group *)pc)->GetCont();
			for (group *pg1 = (group *)bas.Ref(); pg1 != &bas; pg1 = pg1->Ref()) {
				const node &b = pg1->GetCont();
				for (element *pe = (element *)b.Ref(); pe != &b; pe = pe->Ref()) {
					Draw(hDC, pe);
				}
			}
		} else {
			Draw(hDC, (element *)pc);
		}
	}
}

static void FnDispNetLu()
{
	pos p;
	GetMouseHitPos(p);
	container *pc = GetHitNetElem(p, gPcb->GetElems());
	if (pc) {
		node net(0);
		GetOwnNet(pc, net);
		DispNet(net);
	}
	PopStack();
	RestorePad(gDispNetPadSave);
}

static void FnDispNetEsc()
{
	PopStack();
	RestorePad(gDispNetPadSave);
}

void DispNetPadSave()
{
	gApp->GetScreen()->SaveDrawPad(gDispNetPadSave);
}

void DispPinPadSave()
{
	gApp->GetScreen()->SaveDrawPad(gDispPinPadSave);
}

#define MAX_DISP_PIN	999

static void GetDispPin(const node &bas, element *pin[], int &n)
{
	for (container *pc = (container *)bas.Ref(); pc != &bas; pc = pc->Ref()) {
		if (pc->IsGroup()) {
			GetDispPin(((group *)pc)->GetCont(), pin, n);
		} else if (((element *)pc)->GetPinNo() && gPcb->IsFDisp((element *)pc)) {
			pin[n] = (element *)pc;
			if (++n >= MAX_DISP_PIN) {
				break;
			}
		}
	}
}

static void FnDispPinLu()
{
	pos p;
	GetMouseHitPos(p);
	container *pc = gPcb->IsHit(p);
	if (pc && pc->GetOwner() && (pc = pc->GetTopOwner())->IsGroup() && *((group *)pc)->GetName() != ' ') {
		element *pin[MAX_DISP_PIN];
		int n = 0;
		GetDispPin(((group *)pc)->GetCont(), pin, n);
		HDC hDC = GetScrHandle()->GetScrDC();
		SetROP2(hDC, R2_COPYPEN);
		SetTextAlign(hDC, TA_TOP | TA_LEFT | TA_NOUPDATECP);
		SetBkMode(hDC, TRANSPARENT);
		SetTextColor(hDC, RGB(255, 255, 255));
		SIZE sz;
		GetTextExtentPoint32(hDC, "A", 1, &sz);
		int cx = sz.cx / 2;
		int cy = sz.cy / 2;
		if (n) {
			for (int i = 0; i < n; ++i) {
				double x, y;
				if (pin[i]->IsArc()) {
					double ang = (((arc *)pin[i])->GetSt() + ((arc *)pin[i])->GetEnd()) / 2.0;
					Pos2NoSnap(pos(pin[i]->GetPos().x + ((arc *)pin[i])->GetR() * cos(ang),
								   pin[i]->GetPos().y + ((arc *)pin[i])->GetR() * sin(ang)), x, y);
				} else if (pin[i]->IsLine()) {
					Pos2NoSnap(pos((pin[i]->GetPos().x + ((line *)pin[i])->GetPe().x) / 2,
								   (pin[i]->GetPos().y + ((line *)pin[i])->GetPe().y) / 2), x, y);
				} else {
					Pos2NoSnap(pin[i]->GetPos(), x, y);
				}
				if (x >= 0.0 && x < GetScrMaxX() && y >= 0.0 && y < GetScrMaxY()) {
					char str[16];
					itoa(abs(pin[i]->GetPinNo()), str, 10);
					TextOut(hDC, (int)x - cx, (int)y - cy, str, strlen(str));
				}
			}
			FlushScr();
			PopStack();
			RestorePad(gDispPinPadSave);
		} else {
			MessageBeep(MB_OK);
		}
	} else {
		MessageBeep(MB_OK);
	}
}

static void FnDispPinEsc()
{
	PopStack();
	RestorePad(gDispPinPadSave);
}
